//-----------------------------------------------------------------------
//	SU(3) Chiral Instabilities in 3+1 dimensions v1.0
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
//
//	This program provides an implementation of the effective theory
//	description of the chiral instabilities first introduced in [1].
//	It relies on a generalization of the effective field theory
//	of classically hard fermion particles and soft classical color
//	fields.
//
//	An effective equation of motion with dissipative ohmic and anomalous
//	non-dissipatice currents is solved in real-time.
//
//-----------------------------------------------------------------------

//*********************************************************************************
//
//     Copyright (C) 2015  Dr. Alexander Rothkopf
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//*********************************************************************************
//-----------------------------------------------------------------------
//	HERE GO THE NECESSARY SYSTEM INCLUDE FILES
//-----------------------------------------------------------------------

	#include <stdio.h>
	#include <string.h>
	#include <stdlib.h>
	#include <stdint.h>
	#include <math.h> 
	#include <unistd.h>
	#include <iostream>
	#include <fstream>
	#include <istream>
	
	#include "fftw3.h"
	
	#include "./EIGENFAST/zheevh3.c"
	#include "./EIGENFAST/zheevq3.c"
	#include "./EIGENFAST/zheevc3.c"
	#include "./EIGENFAST/slvsec3.c"
	
//-----------------------------------------------------------------------
//	DEFINE ANY PREPROCESSOR MACROS THAT ARE NEEDED IN THE 
//	AUXILIARY ROUTINES TO BE INCLUDED AFTERWARDS
//-----------------------------------------------------------------------

	#define PI (3.141592653589793)
	
	#define NDIM 3
	
	#define SUN (2)			// NUMBER OF COLORS NEEDS TO BE THREE SINCE OPTIMIZATION IS USED	
	#define SUNCOLORS (SUN*SUN)	// NUMBER OF ENTRIES IN THE GAUGE LINKS
	#define	SUNGENNUM (SUN*SUN-1)  	// NUMBER OF FUNDAMENTAL REP. GENERATORS
	
	#define Matrix ComplGenMat
	
	#define THREADS 1
	
//-----------------------------------------------------------------------
//	Include the complex datatype
//-----------------------------------------------------------------------
	
	#include "./AUXILIARY/1_COMPLEX_DATATYPE.c"
	
//-----------------------------------------------------------------------
//	For accelerated linear algebra routines use MKL or LAPACK
//-----------------------------------------------------------------------
		
	#define LAPACK 1
	
	#if (LAPACK==1)
	
	#define MKL_Complex16 Complex
	#define _lapack_complex_double Complex
	#include "mkl.h"
	
	#endif
	
	#if (LAPACK==2)
	
 	#define _lapack_complex_double Complex
	#include "lapacke.h"
	
	#endif

//-----------------------------------------------------------------------
//	Include the matrix datatype and the random number generator
//-----------------------------------------------------------------------

	#include "./AUXILIARY/3_RNG_IMPLEMENTATION.c"
	#include "./AUXILIARY/2_MATRIX_DATATYPES.c"
	
	
//-----------------------------------------------------------------------
//	Define the parameter datatype
//-----------------------------------------------------------------------
		
	struct Parameters
	{
		//GEOMETRY 
	  
		int NX;
		int NY;
		int NZ;
		
		int NT;
		
		double dx;
		double dt;
		
		//Physics Parameters
		
		double g;
		double BetaL;
		
		double ColorConductivity;
		double DiffusionConst;
		double ConstantChemicalPotential;
		int NFermions;
				
		// Cooling Parameters
		
		int WHICHCOOLPRESCR;
		int COARSENING;
		int COOLBEFORECOARSENING;

		int COOLDEPTH;
		int COOLEVERY;
		
		double dtau;

		double ConditioningCooling;
		
		// Initial Condition Parameters
		
		int INITCOND;
		double InitCondN5;

		int METRPRECONSTEPS;
		int MOOREMIXINGSTPS;
		int MOOREPROJMIXSTP;

		double SigmaX,SigmaY,SigmaZ,InitGGFStrength;	
		
		//LeviCivita Implementation
		
		int *mucomp, *nucomp, *sgcomp, *rocomp;
		double *LCval;

		//Measurement Parameters

		int MEASEACH;
		
		int WloopRMax;
		int WloopRMin;
		
		int SpecBins;
		
		int WHICHNCSDEF;
		
		// Time Evolution Parameters
		
		int WHICHREALTIME;
		
		int DYNAMICALN5;
		int N5SWITCHONAT;

		int SAVEEACH;
		
		int VERBOSITY;
		
		//Testing Parameters
		
		int DispersionSwitch;

	};


//-----------------------------------------------------------------------
//	FORWARD DECLARATIONS
//-----------------------------------------------------------------------
	
	void SUN_CLASS_GAUSS_CONSTR(double* FIELD, double* DeviationLocal, double* DeviationGlobal, void* param);	
	void SUN_CLASS_ENERGY(double* FIELDNOW, double* CENRG, double* CMNRG, Parameters* param);
	void SUN_CLASS_SPATIALPLAQ(double* FIELDNOW, Complex* SpatialPlaqTr, void* param);
	void FMUNU_SPATIAL( Matrix** Fmunu, double* FIELDNOW, int64_t* AXLEN, int64_t* ACNT, int64_t* STRIDES, double SUNCLDX, double SUNCLDT, double SUNCLCOUPL);
	void FMUNU_TMPORAL( Matrix** Fmunu, double* FIELDNOW, double* FIELDNEXT, double* FIELDBEF, int64_t* AXLEN, int64_t* ACNT, int64_t* STRIDES, double SUNCLDX, double SUNCLDT, double SUNCLCOUPL, int BACKWARDS);
	
//-----------------------------------------------------------------------
//	INITIAL CONDITION FOR THERMAL EQUILIBRIUM
//-----------------------------------------------------------------------

	void SUN_CLASS_DRAW_E_THERMAL(double* FIELDNOW, TLP_RANDGEN **LocGSLr, Parameters* param)
	{

	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF  
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNBETA=CurParam->BetaL;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double EFAC= 1.;
				double SIGMA=sqrt(1./(2.*SUNCLDX*SUNCLDX*SUNBETA));
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				for(int64_t mu=0; mu<NDIM; mu++){
				
					double ECurG[SUNGENNUM];
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
				
						ECurG[SUNGENCNT]=TLP_RAND_GAUSSIAN_ZIGGURAT (LocGSLr[0],SIGMA);
					}
									
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){						
						FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=ECurG[SUNGENCNT];
					}
					
				}
			}
		    }
		}
	}


	void SUN_CLASS_INITIALIZE_BOTH(double *FIELDNOW, TLP_RANDGEN *RndGen, Parameters* param)
	{

	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF  
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		int which=CurParam->INITCOND;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				for(int64_t mu=0; mu<NDIM; mu++){
					
					Matrix TMP;
				  
					if(which==2||which==3){
					  TMP.Rand1SUN(RndGen);
					}else{
					  TMP.Unit();  
					}
					
					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]= TMP[CMP];
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
						
					    if(which==2){
						FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=(1.-2.*TLP_RAND_GEN_DOUBLE_UNIFORM (RndGen));
					    }else{
						FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=0.;
					    }
					}
					
				}			
			}
		    }
		}	
	}

    void SUN_CL_WILSON_ACTION_STAPLE_SUM( double *FIELDNOW, int64_t* ACNT, int64_t* AXLEN, int64_t* STRIDES, int64_t mu, Matrix* AllStaple)
	{
					
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		int64_t USTRIDE=2*SUN*SUN;
		int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
		int64_t POS=0;	
		for (int k=0;k<NDIM;k++) POS+= ( ( ACNT[k] ) %AXLEN[k] )*STRIDES[k];
		POS*=(ESTRIDE+SUNGENNUM*NDIM);
					
		//-----------------------------------------------------------------------
		//	CALCULATE THE STAPLE SUM
		//-----------------------------------------------------------------------
		
		Matrix PSum;
		Matrix P1;
		Matrix Unit;
		
		PSum.Zero();

		for( int64_t nu=0; nu<NDIM;nu++){
			
			if(mu!=nu){
		
			//-----------------------------------------------------------------------
			//	FORWARD DIRECTION
			//-----------------------------------------------------------------------

			Matrix TMP;
			
			int64_t POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			int64_t POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
			POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1= &FIELDNOW[POS3 + (nu)*USTRIDE];
			TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
			P1*=TMP;
			TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
			P1*=TMP;
			
			PSum+=P1;		

			//-----------------------------------------------------------------------
			//	BACKWARD DIRECTION 
			//-----------------------------------------------------------------------

			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];	POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
			TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
			P1*=TMP;
			P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
			
			PSum+=P1;	
			
		}}

		AllStaple[0]=PSum;

		
	}	

    void SUN_CL_METR_KERNEL(Matrix& AllStaple, double* Link, double BetaL, int NHITS, TLP_RANDGEN **GSLrLoc)
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Matrix PlaqSum,TMP2,TMP1;

		Matrix Eps;

		double pf=BetaL;

		double sel,S_Old,S_New,accept;
		
		for(int HIT=0; HIT<NHITS; HIT++){


			//***************************************************//
			//	FIRST WE NEED A RAND MATRIX CLOSE TO ONE     //
			//***************************************************//

			Eps.Rand1SUN(GSLrLoc[0]);
			
			//***************************************************//
			//	CALCULATE THE SUM OF THE PLAQUETTES FOR S    //
			//	WITH THE CURRENT LINK U                      //
			//***************************************************//

			TMP1=Link; 
			PlaqSum.DotMEqual(AllStaple,TMP1);

			//***************************************************//
			//	CALCULATE THE CURRENT ACTION VALUE           //
			//***************************************************//

			S_Old=pf*( - PlaqSum.RealTrace() );

			//***************************************************//
			//	GENERATE A TRIAL RANDOM MATRIX FROM THE      //
			//	CURRENT LINK BY PERTURBING SLIGHTLY USING    //
			//	Eps                                          //
			//***************************************************//

			TMP1*=Eps;
			PlaqSum.DotMEqual(AllStaple,TMP1);
			
			//***************************************************//
			//	CALCULATE THE CURRENT ACTION VALUE           //
			//***************************************************//

			S_New=pf*( - PlaqSum.RealTrace() );
			
			accept=exp(S_Old-S_New);

			sel=TLP_RAND_GEN_DOUBLE_UNIFORM (GSLrLoc[0]);

			//***************************************************//
			//	AT EACH HIT WE TRY TO UPDATE THE LINK        //
			//	IF IT IS NOT UPDATED AFTER NHITS HITS        //
			//	WE LEAVE IT IN THE UNMODIFIED STATE          //
			//	NORMALLY ONE WOULD WAIT UNTIL THE LINK HAS   //
			//	BEEN UPDATED WHICH HOWEVER CAN INTRODUCE     //
			//	DELAYS WHEN PARALLELIZING THE SIMULATION     //
			//***************************************************//
			
			if(sel<accept){
				
				for(int i=0;i<(2*SUN*SUN);i++) Link[i]=TMP1[i];
				
			
			}

		}		
		
	}
	
	void SUN_CL_3D_METROPOLIS_PRECONDITION(double* FIELDNOW, TLP_RANDGEN** Rng, Parameters* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		int VERBOSITY=CurParam->VERBOSITY;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		int METRPRECONSTEPS=CurParam->METRPRECONSTEPS;
		double SUNCLBETA=CurParam->BetaL;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];

		for(int THCNT=0; THCNT<METRPRECONSTEPS; THCNT++){
		
			int CHKBRDSTEP=2;
			
			for(int CHKX=0; CHKX<CHKBRDSTEP; CHKX++)
			for(int CHKY=0; CHKY<CHKBRDSTEP; CHKY++)
			for(int CHKZ=0; CHKZ<CHKBRDSTEP; CHKZ++)
				
			for  (ACNT[0]=CHKX; ACNT[0]<AXLEN[0]; ACNT[0]+=CHKBRDSTEP){
				
			int64_t pos0=ACNT[0]*STRIDES[0];
			
			for (ACNT[1]=CHKY; ACNT[1]<AXLEN[1]; ACNT[1]+=CHKBRDSTEP){
				
				int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
				
				for (ACNT[2]=CHKZ; ACNT[2]<AXLEN[2]; ACNT[2]+=CHKBRDSTEP){
								
					int64_t USTRIDE=2*SUN*SUN;
					int64_t ESTRIDE=2*SUN*SUN*NDIM;

					int64_t POS=pos1 + ACNT[2]*STRIDES[2];
					POS*=(ESTRIDE+SUNGENNUM*NDIM);

					for(int64_t mu=0; mu<NDIM; mu++){
			
						Matrix AllStaple;
						
						//***************************************************//
						//	CALCULATE AND ADD THE SIX STAPLES AT X:MU    //
						//***************************************************//

						SUN_CL_WILSON_ACTION_STAPLE_SUM(FIELDNOW, ACNT, AXLEN, STRIDES, mu, &AllStaple);
						
						//***************************************************//
						//	     START MULTIHIT METROPOLIS               //
						//***************************************************//

						SUN_CL_METR_KERNEL(AllStaple,&FIELDNOW[POS+USTRIDE*mu], SUNCLBETA, 10, Rng);
					}
					
				}
			}
			}
			
			if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");	
			if(VERBOSITY==1) fprintf(stderr, "***  THERM: STEP:%3d    METROPOLIS PRECONDITIONING     ***\n",THCNT);
			
			if((VERBOSITY==1)){
				//---------------------------------------------------------------------------------------------
				//	EVALUATE THE INITIAL ENERGY DENSITY
				//---------------------------------------------------------------------------------------------
				
				double ClassENrg, ClassMNrg;
				
				SUN_CLASS_ENERGY(FIELDNOW, &ClassENrg, &ClassMNrg, param);

				//---------------------------------------------------------------------------------------------
				//	EVALUATE THE INITIAL PLAQUETTE
				//---------------------------------------------------------------------------------------------
				
				Complex SpPlaq;
				
				SUN_CLASS_SPATIALPLAQ(FIELDNOW, &SpPlaq, param);	
				
				fprintf(stderr, "***         NRG: (%.8e - %.8e)     ***\n",ClassENrg,ClassMNrg);
				fprintf(stderr, "***        PLAQ: (%+.8e - %+.8e)   ***\n",SpPlaq.real(), SpPlaq.imag() );
				fprintf(stderr, "***                                                    ***\n");
			}
		}
	}

	void SUN_CLASS_PROJECT_E_GC(double* FIELDNOW, double* FIELDNEXT, int ATWHICHSTEP, Parameters* param)
	{

	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		
		int VERBOSITY=CurParam->VERBOSITY;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];

		double CONSTRERR=1000;
		
		int thcnt=0;	
		while (CONSTRERR>1e-15){	
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				//-----------------------------------------------------------------------
				//	CALCULATE THE GAUSS CONSTRAINT AND ITS DEVIATION FROM ZERO
				//-----------------------------------------------------------------------

				Matrix E1,E2,GCATX,GCATXP,QUENCH,TMP;
				Matrix U1,U2;

				double *CurDeV;
				CurDeV= new double [SUNGENNUM];

				//-----------------------------------------------------------------------
				//	CALCULATE THE GAUSS CONSTRAINT AT POINT X
				//-----------------------------------------------------------------------
				
				GCATX.Zero();
				
				for(int64_t DIM=0; DIM<NDIM; DIM++){

					int64_t POS2=0;
					for (int k=0;k<NDIM;k++) POS2+= ( (ACNT[k] - ((k==DIM)?1:0) + AXLEN[k] )%AXLEN[k] )*STRIDES[k];
					POS2*=(ESTRIDE+SUNGENNUM*NDIM);

					E1.FROM_REAL_SUN_GENERATORS( &FIELDNOW[POS +ESTRIDE+SUNGENNUM*DIM] );
					E2.FROM_REAL_SUN_GENERATORS( &FIELDNOW[POS2+ESTRIDE+SUNGENNUM*DIM] );

					U2= &FIELDNOW[POS2+USTRIDE*DIM];

					TMP.Dagger(U2);
					TMP*=E2;
					TMP*=U2;
					
					TMP-=E1; 
						
					GCATX+=TMP; //In Moores definition this becomes -C(x)
				}

				//-----------------------------------------------------------------------
				//	FOR EACH E_mu CALCULATE THE GAUSS CONSTRAINT AT THE NEXT 
				//	SPATIAL POSITION IN MU DIRECTION
				//-----------------------------------------------------------------------
				
				for(int64_t mu=0;mu<NDIM;mu++){

					int64_t NEXTPOS1=0;
					for (int k=0;k<NDIM;k++) NEXTPOS1+= ( (ACNT[k] + ((k==mu)?1:0) )%AXLEN[k] )*STRIDES[k];
					NEXTPOS1*=(ESTRIDE+SUNGENNUM*NDIM);

					//-----------------------------------------------------------------------
					//	CALCULATE THE GAUSS CONSTRAINT AT POINT X+\mu
					//-----------------------------------------------------------------------
					GCATXP.Zero();
					
					for(int64_t DIM=0; DIM<NDIM; DIM++){
						
						int64_t POS2=0;
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==DIM)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						E1.FROM_REAL_SUN_GENERATORS( &FIELDNOW[NEXTPOS1 +ESTRIDE+SUNGENNUM*DIM] );
						E2.FROM_REAL_SUN_GENERATORS( &FIELDNOW[POS2     +ESTRIDE+SUNGENNUM*DIM] );
						
						U2= &FIELDNOW[POS2+USTRIDE*DIM];

						TMP.Dagger(U2);
						TMP*=E2;
						TMP*=U2;
						
						TMP-=E1; //In Moores definition this is -C(x+mu)

						GCATXP+=TMP;
					}
					
					Matrix NU;
					NU=&FIELDNOW[POS +USTRIDE*mu];

					QUENCH=NU;
					QUENCH*=GCATXP;
					TMP.Dagger(NU);
					QUENCH*=TMP;
					
					QUENCH-=GCATX;
					
					QUENCH*=0.12;
					
					E1.FROM_REAL_SUN_GENERATORS( &FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu] );
					
					TMP=E1;

					TMP-=QUENCH; //Since I use -C instead of C here we need a minus sign
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
						FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=2.*TMP.TrGen(SUNGENCNT).real();
					}

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNEXT[POS + (mu)*USTRIDE + CMP]= FIELDNOW[POS + (mu)*USTRIDE + CMP];
				}
				
				delete[] CurDeV;
				
			}
		    }
		}
		
		
		
		thcnt++;
		double* tmp=FIELDNOW;		
		FIELDNOW=FIELDNEXT;
		FIELDNEXT=tmp;

		double GaussDevLoc,GaussDevGlob;
		
		SUN_CLASS_GAUSS_CONSTR(FIELDNOW, &GaussDevLoc, &GaussDevGlob, param);

		if(VERBOSITY==1){
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  THM PHASE [%3d] PROJECTION [%5d] TOW GAUSS CONS ***\n",ATWHICHSTEP,thcnt);
		fprintf(stderr, "***        GCLOC:%+.5e     GCGLB:%+.5e   ***\n",GaussDevLoc,GaussDevGlob);
		fprintf(stderr, "***                                                    ***\n");
		}
		
		CONSTRERR=GaussDevLoc;
		
		}
	}

//-----------------------------------------------------------------------
//	INITIAL CONDITION FOR FREE FIELDS
//-----------------------------------------------------------------------

	void SUN_CLASS_INITIALIZE_GAUSSIAN_MOMENTA(double* FIELDNOW, TLP_RANDGEN** Rng, Parameters* param)
	{
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "*** Init Gauge fields as aniso Gaussian in mom space   ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		FILE* XOUT, *POUT;

		//-----------------------------------------------------------------------
		//	I am grateful to Sabolcs Borsanyi for providing me with the
		//	3D counting scheme for the negative momenta 
		//-----------------------------------------------------------------------
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		int64_t MCNT[NDIM];	//COUNTERS FOR THE NEGATIVE MOMENTA
		int64_t PCNT;		//SINGLE COUNTER FOR THE POSITIVE MOMENTA
			
		int AXLENFFT[NDIM];	//AXES LENGTH FOR THE FFT PLANS
		double KVAL[NDIM];	//POSITIVE MOMENTA VALUES
		double MKVL[NDIM];	//NEGATIVE MOMENTA VALUES

		fftw_plan FFT;		//FFTW PLAN FOR THE FORWARD TRANSFORMATION
		fftw_plan IFFT;		//FFTW PLAN FOR THE BACKWARD TRANSFORMATION
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		AXLENFFT[0]=CurParam->NX;
		AXLENFFT[1]=CurParam->NY;
		AXLENFFT[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		int64_t IVOL=AXLEN[0]*AXLEN[1]*AXLEN[2];	

		double Dx,Dy,Dz;
		Dx=CurParam->SigmaX; Dy=CurParam->SigmaY; Dz=CurParam->SigmaZ;
		
		double Strength=CurParam->InitGGFStrength;
		double prefac=sqrt(Strength/(pow(2.*PI,1.5)*Dx*Dy*Dz));
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "*** SgmX:%.2e SgmY:%.2e SgmZ:%3.2e          ***\n", Dx,Dy,Dz);
		fprintf(stderr, "***                                                    ***\n");
		
		//*********************************************************************************
		//  SETUP THE FOURIER SPACE MOMENTA, WHICH NEED TO BE SET SUCH THAT THE NYQUIST
		//  FREQUENCY IS CORRECTLY CAPTURED AT AXLEN/2
		//*********************************************************************************
		
		double **PPOS;
		PPOS=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) PPOS[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<=AXLEN[DIM]/2;PCNT++) PPOS[DIM][PCNT]=PCNT*((2.*PI)/(((double)AXLEN[DIM])*((double)SUNCLDX)));
			for(int PCNT=AXLEN[DIM]/2+1;PCNT<AXLEN[DIM];PCNT++) PPOS[DIM][PCNT]=-PPOS[DIM][AXLEN[DIM]-PCNT];
		}
			
		
		//*********************************************************************************
		// THE ARRAY FOR THE A FIELDS USED IN THE INITIALIZATION
		//*********************************************************************************
		
		fftw_complex *AFIELDS, *CURFIELD, *TSTFIELD;
		AFIELDS=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*NDIM*SUNGENNUM*IVOL);
		TSTFIELD=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*IVOL);
		
		for(int64_t k=0;k<VOL;k++) { TSTFIELD[k][0]=0.; TSTFIELD[k][1]=0.; }
		
		//*********************************************************************************
		// GENERATE GAUSSIAN DISTRIBUTED MOMENTA FOR EACH COLOR AND SPATIAL COMPONENT
		//*********************************************************************************
		
		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				//fprintf(stderr, "COLOR %d SPATIAL DIR %d\n", COLINDEX, SPINDEX);
				
				CURFIELD=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				FFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,-1, FFTW_ESTIMATE);
				IFFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,+1, FFTW_ESTIMATE);

				//-----------------------------------------------------------------------
				//	MCNT, the counter for the negative frequencies is always 
				//	incremented at the end of each loop step since,
				//	the zero position does not have a counterpart (the NYQUIST
				//	FREQUENCY takes this role) and +K <-> -K sorting starts from
				//	step 1 NOT 0.
				//-----------------------------------------------------------------------
				
				for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				  for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
						
					if(PCNT<=MCNT[2]){
							
						for(int i=0; i<NDIM; i++){				
							KVAL[i]=PPOS[i][ACNT[i]];						
							if(ACNT[i]!=AXLEN[i]/2) {MKVL[i]=-KVAL[i];}else{MKVL[i]=KVAL[i];}
						}
						
	// 					CURFIELD[PCNT][1]=-(cos(1.*KVAL[0])*sin(2.*KVAL[1])+sin(1.*KVAL[0])*cos(2.*KVAL[1]));
	// 					CURFIELD[PCNT][0]=  cos(1.*KVAL[0])*cos(2.*KVAL[1])-sin(1.*KVAL[0])*sin(2.*KVAL[1]);

						CURFIELD[PCNT][1]=0.;					
						CURFIELD[PCNT][0]=sqrt(prefac*exp(-(KVAL[0]*KVAL[0])/(4.*Dx*Dx)-(KVAL[1]*KVAL[1])/(4.*Dy*Dy)-(KVAL[2]*KVAL[2])/(4.*Dz*Dz) ));
				
						if(CURFIELD[PCNT][0]<1e-8){
							CURFIELD[PCNT][0]=1e-8*	TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],1.);
						}else{
							CURFIELD[PCNT][0]*=	TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],1.);
						}
							
						if(PCNT!=MCNT[2]){
							CURFIELD[MCNT[2]][1]=-CURFIELD[PCNT][1];
							CURFIELD[MCNT[2]][0]= CURFIELD[PCNT][0];
						}
					
					}
					
				      }
				  }
				}
		
				//*********************************************************************************
				//  DEBUGGING OUTPUT TO CHECK THE CORRECTNESS OF THE FOURIER TRANSFORM
				//*********************************************************************************
		
				#if (0)
				
				POUT=fopen("InitFLDPSPACE.dat","w");
				
				for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
					
				int64_t pos0=ACNT[0]*STRIDES[0];
				
				for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
					
					int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
					
					for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
					
					int64_t POS=pos1 + ACNT[2]*STRIDES[2];

					fprintf(POUT, "%ld %ld %ld %.10e %.10e\n", ACNT[0], ACNT[1], ACNT[2], CURFIELD[POS][0],CURFIELD[POS][1]);
				}}}

				fclose(POUT);
			
				
				fftw_execute(IFFT);
				
				XOUT=fopen("InitFLDXSPACE.dat","w");
				
				for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
					
				int64_t pos0=ACNT[0]*STRIDES[0];
				
				for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
					
					int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
					
					for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
					
					int64_t POS=pos1 + ACNT[2]*STRIDES[2];

					fprintf(XOUT, "%ld %ld %ld %.10e %.10e\n", ACNT[0], ACNT[1], ACNT[2], CURFIELD[POS][0],CURFIELD[POS][1]);
				}}}
				
				fclose(XOUT);
				exit(-1);
				#endif
				
				fftw_destroy_plan(FFT);
				fftw_destroy_plan(IFFT);
			
			}
		}
		
		//*********************************************************************************
		//  DEBUGGING THE CORRECT NOISE STRUCTURE IN AVERAGING OVER ALL COLOR AND
		//  SPATIAL DIRECTIONS
		//*********************************************************************************
		
		#if (1)
		
		for(int k=0;k<SUNGENNUM*NDIM;k++){
			
			CURFIELD=&AFIELDS[IVOL*k];
					
			for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
					
				if(PCNT<=MCNT[2]){
					
					for(int i=0; i<NDIM; i++){				
						KVAL[i]=PPOS[i][ACNT[i]];						
						if(ACNT[i]!=AXLEN[i]/2) {MKVL[i]=-KVAL[i];}else{MKVL[i]=KVAL[i];}
					}
					
					TSTFIELD[PCNT][0]+=(1./(SUNGENNUM*NDIM))*(CURFIELD[PCNT][0]*CURFIELD[MCNT[2]][0]-CURFIELD[PCNT][1]*CURFIELD[MCNT[2]][1]);
					TSTFIELD[PCNT][1]+=(1./(SUNGENNUM*NDIM))*(CURFIELD[PCNT][0]*CURFIELD[MCNT[2]][1]+CURFIELD[PCNT][1]*CURFIELD[MCNT[2]][0]);

					if(PCNT!=MCNT[2]){
					TSTFIELD[MCNT[2]][0]+=(1./(SUNGENNUM*NDIM))*(CURFIELD[PCNT][0]*CURFIELD[MCNT[2]][0]-CURFIELD[PCNT][1]*CURFIELD[MCNT[2]][1]);
					TSTFIELD[MCNT[2]][1]+=(1./(SUNGENNUM*NDIM))*(CURFIELD[PCNT][0]*CURFIELD[MCNT[2]][1]+CURFIELD[PCNT][1]*CURFIELD[MCNT[2]][0]);
					}
					
				
				}
				
				}
				}
			}		
		}
		
		POUT=fopen("AvgFLDPSPACE.dat","w");

			for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
				
			int64_t pos0=ACNT[0]*STRIDES[0];
			
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
				
				int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
				
				for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];

				fprintf(POUT, "%ld %ld %ld %.10e %.10e\n", ACNT[0], ACNT[1], ACNT[2], TSTFIELD[POS][0],TSTFIELD[POS][1]);
			}}}

		fclose(POUT);
		#endif

		//*********************************************************************************
		//	USE THE JUST INITIALIZED A FIELDS TO SET UP THE LATTICE LINK VARIABLES
		//*********************************************************************************
		
		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				//fprintf(stderr, "IFFT COLOR %d SPATIAL DIR %d\n", COLINDEX, SPINDEX);
				
				CURFIELD=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				IFFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,+1, FFTW_ESTIMATE);
				
				fftw_execute(IFFT);
				
				fftw_destroy_plan(IFFT);
				
			}
		}
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		int64_t pos0=ACNT[0]*STRIDES[0];
		
		for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			
			int64_t USTRIDE=2*SUN*SUN;
			int64_t ESTRIDE=2*SUN*SUN*NDIM;
			
			int64_t POS1=pos1 + ACNT[2]*STRIDES[2];
			int64_t POS2=POS1;
			POS2*=(ESTRIDE+SUNGENNUM*NDIM);
			
	// 		double *TMP, *TMP2;
	// 		TMP=new double [SUNGENNUM];
	// 		TMP2=new double [SUNGENNUM];
		
			double TMP[SUNGENNUM];
			//double TMP2[SUNGENNUM];
			
			for(int64_t mu=0;mu<NDIM;mu++){
				
			for(int64_t k=0; k<SUNGENNUM; k++) TMP[k]=AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]/VOL;
			
			Matrix U;
			U.FROM_REAL_SUN_GENERATORS(TMP);
			
			Complex Ufac(0.,SUNCLCOUPL*SUNCLDX);
			
			U*=Ufac;
			U.Exponential();
			
			for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS2 + (mu)*USTRIDE + CMP]=U[CMP];
			
			#if (0)
			U=&FIELDNOW[POS2 + (mu)*USTRIDE];
			for(int64_t k=0; k<SUNGENNUM; k++) TMP2[k]=U.ImTrGen(k)/(SUNCLCOUPL*SUNCLDX); 
			for(int64_t k=0; k<SUNGENNUM; k++) fprintf(stderr, "Comp: %e %e\n", TMP[k], TMP2[k]);
			exit(-1);
			#endif
			
			for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) FIELDNOW[POS2+ESTRIDE+SUNGENNUM*mu+CMP]=0.;
				
			}
			
	// 		delete[] TMP;
	// 		delete[] TMP2;
			
		}}}
		
		fftw_free(AFIELDS);
		fftw_free(TSTFIELD);
			
		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(PPOS[DIM]);
		fftw_free(PPOS);
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "*** Gaussian Initialization DONE!                      ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		
	}


//-----------------------------------------------------------------------
//	EVOLUTION EQUATIONS FOR U AND E FIELDS YANG MILLS ONLY
//-----------------------------------------------------------------------

	void SUN_CLASS_EVOLVE_EFIELD_FIRSTSTEP(double* FIELDNOW, Parameters* param )
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double EEVOLFAC=SUNCLDT/(2.*sqrt(SUN)*SUNCLDX*SUNCLDX);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
					
						double FLDNOWTMP=FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT];
					  
						FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=FLDNOWTMP-EEVOLFAC*PSum.ImTrGen(SUNGENCNT);
					
					}

				}			
				
			}
		    }
		}
	}


	void SUN_CLASS_EVOLVE_EFIELD(double* FIELDNOW, double* FIELDNEXT, Parameters* param )
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double EEVOLFAC=SUNCLDT/(sqrt(SUN)*SUNCLDX*SUNCLDX);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
					
						FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]-EEVOLFAC*PSum.ImTrGen(SUNGENCNT);
					
					}

				}			
				
			}
		    }
		}
	}

	void SUN_CLASS_EVOLVE_UFIELD(double* FIELDNOW, double* FIELDNEXT, Parameters* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				Complex UEvolFac(0.,2.*SUNCLDT*sqrt(SUN));
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				for(int64_t mu=0; mu<NDIM; mu++){

					Matrix E,U,UE;
					
					//-----------------------------------------------------------------------
					//	LEAPFROG STEP USING THE NEXT EVOLVED E-FIELD TO UPDATE
					//	THE CURRENT U LINK
					//-----------------------------------------------------------------------
					
					E.FROM_REAL_SUN_GENERATORS(&FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu]);
					U= &FIELDNOW[POS + (mu)*USTRIDE];
	#if (SUN==3)	
					
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=2.*SUNCLDT*sqrt(SUN);
					eval[1]*=2.*SUNCLDT*sqrt(SUN);
					eval[2]*=2.*SUNCLDT*sqrt(SUN);
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];		
					
	#elif (SUN==2)
					double rsqr=0;
					
					for(int64_t k=0;k<SUNGENNUM;k++) rsqr+=FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+k]*FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+k];

					double r=SUNCLDT*sqrt(SUN*rsqr);
					
					Complex SinFac(0.,2.*sin(r)/sqrt(rsqr)); //FACTOR 2 BECUASE OF GENERATORS DEFINITION WITH 1/2
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;
	#endif					
					UE=E;
					UE*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNEXT[POS + (mu)*USTRIDE + CMP]=UE[CMP];
					
				}			
				
			}
		    }
		}
	}

//-----------------------------------------------------------------------
//	EVOLUTION EQUATIONS FOR E,U AND N5 IN CHIRAL EFT
//-----------------------------------------------------------------------

    void NFIVE_DIFF_TRIDIAG_SOLVER(double x[], const size_t N, const double a[], const double b[], const double c[],double scratch[]) 
	{

	    size_t in;
	    
	    scratch[0] = c[0] / b[0];
	    x[0] = x[0] / b[0];
	
	    /* loop from 1 to N - 1 inclusive */
	    for (in = 1; in < N; in++) {
		double m = 1.0 / (b[in] - a[in] * scratch[in - 1]);
		scratch[in] = c[in] * m;
		x[in] = (x[in] - a[in] * x[in - 1]) * m;
	    }
	
	    /* loop from N - 2 to 0 inclusive, safely testing loop end condition */
	    for (in = N - 1; in-- > 0; )
		x[in] = x[in] - scratch[in] * x[in + 1];	
		
		
	}
	
	
    void SUN_CLASS_EVOLVE_UFIELD_LANGEVIN_RUNGE_KUTTA(double* FIELDNOW, double* FIELDTMP, TLP_RANDGEN** Rng, Parameters* param )
	{
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		double SUNCONDUCT=CurParam->ColorConductivity;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
			
		for(int RKSTEP=0;RKSTEP<3;RKSTEP++){
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}				
					
					double TMP[SUNGENNUM];
					double Hfac=-2./SUNCLCOUPL;
					double SIGMA=sqrt(4.*SUN*SUNCONDUCT/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLBETA*SUNCLDT));
					
					if(RKSTEP==0){
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++){
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]=0.25*( Hfac*PSum.ImTrGen(CMP)+TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],SIGMA) );
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==1) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-17./9.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(8./9.)*( Hfac*PSum.ImTrGen(CMP) + TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],SIGMA) ) ;
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==2) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-1.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(3./4.)*( Hfac*PSum.ImTrGen(CMP) + TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],SIGMA) );
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					}				
					
					E.FROM_REAL_SUN_GENERATORS(TMP);

	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					eval[1]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					eval[2]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=SUNCLDT*SUNCLDX*SUNCLCOUPL*sqrt(0.25*rsqr);
					
					Complex SinFac(0.,2.*sin(r)/sqrt(rsqr));
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;		
					
	#endif				
					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDTMP[POS + (mu)*USTRIDE + CMP]=E[CMP];

				}			
				
			}
		    }
		}
		

		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]=FIELDTMP[POS + (mu)*USTRIDE + CMP];
					
				}
				
			}
		    }
		}
		
		}
		
	}

    void SUN_CLASS_EVOLVE_UFIELD_LANGEVIN_FINITEMU(double* FIELDNOW, double* FIELDNEXT,double* FIELDBEF, double* N5NOW, double* N5NEXT, double* N5SCRATCH, TLP_RANDGEN** Rng, Parameters* param )
	{

		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		double SUNCOLORCONDUCT=CurParam->ColorConductivity;
		double SUNCLDIFFCONST=CurParam->DiffusionConst;
		double SUNCLMUCNST=CurParam->ConstantChemicalPotential;
		double SUNNUMBFERM=CurParam->NFermions;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;	
		
		//-----------------------------------------------------------------------
		//	AUXILIARY ARRAYS FOR THE ADI TRIDIAGONAL SOLVER
		//-----------------------------------------------------------------------
		
		int NX=AXLEN[0];
		
		if(AXLEN[0]!=AXLEN[1]||AXLEN[1]!=AXLEN[2]||AXLEN[2]!=AXLEN[0]){
			fprintf(stderr, "ASSUMING A LATTICE WITH EQUAL SPATIAL EXTEND IN EACH DIMENSION\n"); exit(-1);
		}
		
		double* SubDiag, *SupDiag, *Diag, *x1, *x2, *q, *qprime, *scratch;
		
		SubDiag =  new double [NX];
		SupDiag =  new double [NX];
		Diag =  new double [NX];
		x1 =  new double [NX];
		x2 =  new double [NX];
		q =  new double [NX];
		qprime =  new double [NX];
		scratch =  new double [NX];
		
		double nuConst=SUNCLDIFFCONST*SUNCLDT/(SUNCLDX*SUNCLDX);
		
		for(int k=0;k<NX;k++) qprime[k]=0.;
		qprime[0]=nuConst/2.;
		qprime[NX-2]=nuConst/2.;
		for(int k=0;k<NX;k++){
			SubDiag[k]=-nuConst/2.;
			SupDiag[k]=-nuConst/2.;
			Diag[k]=1.+nuConst;
			
		}
			
		//-----------------------------------------------------------------------
		//	THE LANGEVIN TIME EVOLUTION FOR THE GAUGE FIELD
		//-----------------------------------------------------------------------
		
		Matrix** Fmunu;
		
		Fmunu = new Matrix* [4];
		for (int k=0;k<4;k++) Fmunu[k]= new Matrix [4];

	#if (1)	
		
		int64_t USTRIDE=2*SUN*SUN;
		int64_t ESTRIDE=2*SUN*SUN*NDIM;
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
													
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				int64_t POSN5=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				FMUNU_SPATIAL( Fmunu, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);			
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}					
					
					double BCURR[SUNGENNUM];
					
					if(mu==0){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[1][2].ReTrGen(SUNGENCNT);
					}else if (mu==1){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[2][0].ReTrGen(SUNGENCNT);
					}else if(mu==2){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[0][1].ReTrGen(SUNGENCNT);
					}
					
					double TMP[SUNGENNUM];
					double Hfac=-2./SUNCLCOUPL;
					double Bfac=(SUNNUMBFERM*SUNCLCOUPL*SUNCLCOUPL/(4.*PI*PI));

					//-----------------------------------------------------------------------
					//	GET THE MU5 APPROXIMATELY FROM N5
					//
					//	CHECK WHETHER WE HAVE THE SINGLE REAL SOLUTION FOR MU5
					//	THEN CALCULATE THIS VALUE
					//-----------------------------------------------------------------------
					
					double Ttmp=2.*SUN/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLBETA);
					
					double c=(3.*SUNCLMUCNST*SUNCLMUCNST*SUNCLMUCNST + PI*PI*Ttmp*Ttmp);
					double d=-N5NOW[POSN5]*3.*PI*PI/(SUN*SUNNUMBFERM);
									
					double DSCR=-4.*c-27.*d*d;
					
					if(DSCR>0){
						fprintf(stderr, "PROBLEM WITH DETERMINING A SINGLE REAL ROOT FOR MU5\n"); exit(-1);
					}
					
					double CapC=exp((1./3.)*log( 0.5 * ( 27.*d + sqrt(729.*d*d+108.*c*c*c) ) ) );
					double D0=-3.*c;
					
					double MU5CUR=-(1./3.)*(CapC+D0/CapC);
					
					//-----------------------------------------------------------------------
					//	CARRY OUT THE LANGEVIN STEP FOR THE GAUGE FIELD
					//-----------------------------------------------------------------------
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){			
						// NOTE: THE DELTA FUNCTION HAS UNITS d_ij ~ 1/dx -> d_ij^3 ~ 1/dx^3	
						double SIGMA=sqrt(4.*SUN*SUNCOLORCONDUCT/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLBETA*SUNCLDT));
						
						
						//ACCORDING TO VERSION 13 OF THE MANUSCRIPT THE B TERM SHOULD COME WITH THE SAME SIGN AS THE NOISE
						TMP[SUNGENCNT]=Hfac*PSum.ImTrGen(SUNGENCNT)+Bfac*MU5CUR*BCURR[SUNGENCNT] + TLP_RAND_GAUSSIAN_ZIGGURAT (Rng[0],SIGMA);
					}
					
					E.FROM_REAL_SUN_GENERATORS(TMP);
					
	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					eval[1]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					eval[2]*=SUNCLDT*SUNCLDX*SUNCLCOUPL;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=SUNCLDT*SUNCLDX*SUNCLCOUPL*sqrt(0.25*rsqr);
					
					Complex SinFac(0.,2.*sin(r)/sqrt(rsqr));
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;		
					
	#endif				

					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNEXT[POS + (mu)*USTRIDE + CMP]=E[CMP];
					
				}
				
			}
		    }
		}
	#endif
		//-----------------------------------------------------------------------
		//	DIFFUSION EQUATION FOR THE CHIRAL NUMBER DENSITY
		//-----------------------------------------------------------------------
		
		//-----------------------------------------------------------------------
		//	FIRST STEP ADI: X-DIRECTION IMPLICIT + HETEROGENEOUS TERM
		//-----------------------------------------------------------------------
		
		for  (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
		
		  for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=0;
				int64_t POSMU=0;

				int64_t POSMUP1X=0;
				int64_t POSMUM1X=0;
				int64_t POSMUP1Y=0;
				int64_t POSMUM1Y=0;
				int64_t POSMUP1Z=0;
				int64_t POSMUM1Z=0;
				
				for(int k=0;k<NDIM;k++){
					
					POS+= ACNT[k]*STRIDES[k];
					POSMU+= ACNT[k]*STRIDES[k];
					
					POSMUP1X+=( ( ACNT[k] + ((k==0)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1X+=( ( ACNT[k] - ((k==0)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Y+=( ( ACNT[k] + ((k==1)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Y+=( ( ACNT[k] - ((k==1)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Z+=( ( ACNT[k] + ((k==2)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Z+=( ( ACNT[k] - ((k==2)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				//-----------------------------------------------------------------------
				//	SINCE WE HAVE EVOLVED THE UFIELDS ALREADY THE EVOLUTION EQUATION
				//	FOR N5 HAS TO CALCULATE THE TOPOLOGICAL TERM ExB FROM THE
				//	FIELDNEXT AND FIELDNOW ARRAY AND NOT FROM FIELDNOW AND FIELDBEF
				//-----------------------------------------------------------------------
				
				FMUNU_SPATIAL( Fmunu, FIELDNEXT, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);	
				
				FMUNU_TMPORAL( Fmunu, FIELDNEXT, NULL, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL,1);	
				
				Complex FFtildeTr;
				FFtildeTr=0.;
					
				for(int64_t k=0;k<24;k++){
					
					Matrix TMP;
					
					TMP=Fmunu[mucomp[k]][nucomp[k]];
					TMP*=Fmunu[sgcomp[k]][rocomp[k]];
					
					FFtildeTr+=-0.5*LCval[k]*TMP.Trace();
				}
				
				double EBFac=SUNNUMBFERM*SUNCLCOUPL*SUNCLCOUPL/(4.*PI*PI);
				
				//1-D
	//			q[ACNT[0]]= (1.-nu)*N5NOW[POSMU]+ (nu/2.)*( N5NOW[POSMUP1X] + N5NOW[POSMUM1X] );

				//2-D
	// 			q[ACNT[0]]= (1.-nu)*N5NOW[POSMU]+ (nu/2.)*( N5NOW[POSMUP1Y] + N5NOW[POSMUM1Y] );
				
				//3-D
				q[ACNT[0]]= 		(1.-5.*nuConst)*N5NOW[POSMU]
						+ 	(nuConst/2.)*( N5NOW[POSMUP1X]+ N5NOW[POSMUM1X] )
						+	nuConst*( N5NOW[POSMUP1Y]+N5NOW[POSMUP1Z] + N5NOW[POSMUM1Y]+N5NOW[POSMUM1Z] )
						+	SUNCLDT*EBFac*FFtildeTr.real();			

				x1[ACNT[0]]=q[ACNT[0]];
				x2[ACNT[0]]=qprime[ACNT[0]];
				
			}
			
			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);		
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]-1; ACNT[0]++){
				
				x1[ACNT[0]]+=x2[ACNT[0]]*x1[NX-1];
				
			}	
			
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
				
				int64_t POSMU=0;
				
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				//2-D
				//N5SCRATCH[POSMU]=x1[ACNT[0]];
				
				//3-D
				N5NEXT[POSMU]=x1[ACNT[0]];
			
			}
		    }
		}
		
		//-----------------------------------------------------------------------
		//	SECOND STEP ADI: Y-DIRECTION IMPLICIT 
		//-----------------------------------------------------------------------
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
		
		  for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSMU=0;

	// 			int64_t POSMUP1X=0;
	// 			int64_t POSMUM1X=0;
				
				int64_t POSMUP1Y=0;
				int64_t POSMUM1Y=0;
				
				for(int k=0;k<NDIM;k++){
					
					POSMU+= ACNT[k]*STRIDES[k];
					
	// 				POSMUP1X+=( ( ACNT[k] + ((k==0)?1:0) ) %AXLEN[k] )*STRIDES[k];
	// 				POSMUM1X+=( ( ACNT[k] - ((k==0)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Y+=( ( ACNT[k] + ((k==1)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Y+=( ( ACNT[k] - ((k==1)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				//2-D
				//q[ACNT[1]]= (1.-nuConst)*N5SCRATCH[POSMU]+ (nuConst/2.)*( N5SCRATCH[POSMUP1X] + N5SCRATCH[POSMUM1X] );
				
				//3-D
				q[ACNT[1]]= N5NEXT[POSMU]+ nuConst*N5NOW[POSMU]- (nuConst/2.)*( N5NOW[POSMUP1Y]+ N5NOW[POSMUM1Y] );
				
				x1[ACNT[1]]=q[ACNT[1]];
				x2[ACNT[1]]=qprime[ACNT[1]];
				
			}

			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]-1; ACNT[1]++){
				
				x1[ACNT[1]]+=x2[ACNT[1]]*x1[NX-1];
				
			}
			
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
				
				int64_t POSMU=0;
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				//2-D
	// 			N5NEXT[POSMU]=x1[ACNT[1]];

				//3-D
				N5SCRATCH[POSMU]=x1[ACNT[1]];
			
			}
		    }
		}

		//-----------------------------------------------------------------------
		//	THIRD STEP ADI: Z-DIRECTION IMPLICIT 
		//-----------------------------------------------------------------------
		
		for  (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
		
		  for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSMU=0;

				int64_t POSMUP1Z=0;
				int64_t POSMUM1Z=0;
				
				for(int k=0;k<NDIM;k++){
					
					POSMU+= ACNT[k]*STRIDES[k];

					POSMUP1Z+=( ( ACNT[k] + ((k==2)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Z+=( ( ACNT[k] - ((k==2)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				q[ACNT[2]]= N5SCRATCH[POSMU]+ nuConst*N5NOW[POSMU]- (nuConst/2.)*( N5NOW[POSMUP1Z]+ N5NOW[POSMUM1Z] );
				
				x1[ACNT[2]]=q[ACNT[2]];
				x2[ACNT[2]]=qprime[ACNT[2]];
			}

			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]-1; ACNT[2]++){
				
				x1[ACNT[2]]+=x2[ACNT[2]]*x1[NX-1];
				
			}
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
				
				int64_t POSMU=0;
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				N5NEXT[POSMU]=x1[ACNT[2]];
			
			}
		    }
		}

		delete[] SubDiag;
		delete[] SupDiag;
		delete[] Diag;
		delete[] x1;
		delete[] x2;
		delete[] q;
		delete[] qprime;
		delete[] scratch;

		for (int k=0;k<4;k++) delete[] Fmunu[k];
		delete[] Fmunu;
		
		
	}	
	
	void SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT(double* FIELDNOW, double* FIELDNEXT,double* N5NOW, TLP_RANDGEN** Rng, Parameters* param )
	{
	  
	//CHECKED ON MAY 22nd 2015 ==> OK	A. ROTHKOPF 
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		double SUNCONDUCT=CurParam->ColorConductivity;
		int DISPERISONONOFF=CurParam->DispersionSwitch;
		
		int DYNAMICALN5=CurParam->DYNAMICALN5;
		
		double DISPSWITCH=1.;
		
		if(DISPERISONONOFF==0) DISPSWITCH=0.;
		
		double SUNCLMUCNST=CurParam->ConstantChemicalPotential;
		double SUNNUMBFERM=CurParam->NFermions;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;	
					
		//-----------------------------------------------------------------------
		//	THE LANGEVIN TIME EVOLUTION FOR THE GAUGE FIELD
		//-----------------------------------------------------------------------
		
		Matrix** Fmunu;
		
		Fmunu = new Matrix* [4];
		for (int k=0;k<4;k++) Fmunu[k]= new Matrix [4];
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			  
				FMUNU_SPATIAL( Fmunu, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);			
			  
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double EEVOLFAC=SUNCLDT/(sqrt(SUN)*SUNCLDX*SUNCLDX);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				int64_t POSN5=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					double BCURR[SUNGENNUM];
					
					if(mu==0){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[1][2].ReTrGen(SUNGENCNT);
					}else if (mu==1){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[2][0].ReTrGen(SUNGENCNT);
					}else if(mu==2){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[0][1].ReTrGen(SUNGENCNT);
					}
					
					double TMP[SUNGENNUM];
					double Bfac=(SUNNUMBFERM*SUNCLCOUPL*SUNCLCOUPL/(4.*PI*PI));

					//-----------------------------------------------------------------------
					//	GET THE MU5 APPROXIMATELY FROM N5
					//
					//	CHECK WHETHER WE HAVE THE SINGLE REAL SOLUTION FOR MU5
					//	THEN CALCULATE THIS VALUE
					//-----------------------------------------------------------------------
					
					//fprintf(stderr, "DEBUG1\n");
					
					double CURN5;
					
					if(DYNAMICALN5==1) { 
					  
					  CURN5=N5NOW[0]; 
					  
					} else if (DYNAMICALN5==2) { 
					  
					  CURN5=N5NOW[POSN5]; 
					  
					} else {
					  
					  CURN5=0.;
					  
					}
					
					//if(ACNT[0]==0&&ACNT[1]==0&&ACNT[2]==0) fprintf(stderr, "DEBUG: DYNN5: %d\n", DYNAMICALN5);
					//fprintf(stderr, "DEBUG2\n");
					
					double Ttmp=2.*SUN/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLBETA);
					
					double c=(3.*SUNCLMUCNST*SUNCLMUCNST*SUNCLMUCNST + PI*PI*Ttmp*Ttmp);
					double d=-CURN5*3.*PI*PI/(SUN*SUNNUMBFERM);
									
					double DSCR=-4.*c-27.*d*d;
					
					if(DSCR>0){
						fprintf(stderr, "PROBLEM WITH DETERMINING A SINGLE REAL ROOT FOR MU5\n"); exit(-1);
					}
					
					double CapC=exp((1./3.)*log( 0.5 * ( 27.*d + sqrt(729.*d*d+108.*c*c*c) ) ) );
					double D0=-3.*c;
					
					double MU5CUR=-(1./3.)*(CapC+D0/CapC);				
					
					double F0iECOnv=(SUNCLCOUPL*SUNCLDX/(2.*sqrt(SUN)));
					//fprintf(stderr, "DEBUG4\n");
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){

						// NOTE: THE DELTA FUNCTION HAS UNITS d_ij ~ 1/dx -> d_ij^3 ~ 1/dx^3
						double SIGMA=sqrt(4.*SUN*SUNCONDUCT/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLDX*SUNCLBETA*SUNCLDT));
					
						FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=(1.-DISPSWITCH*SUNCONDUCT*SUNCLDT)*FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]-EEVOLFAC*PSum.ImTrGen(SUNGENCNT) - SUNCLDT*Bfac*MU5CUR*F0iECOnv*BCURR[SUNGENCNT]  + SUNCLDT*F0iECOnv* TLP_RAND_GAUSSIAN_ZIGGURAT  (Rng[0],SIGMA);
					
					}
					//fprintf(stderr, "DEBUG5\n");

				}			
				
			}
		    }
		}
	}

    void SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT_TASSLER(double* FIELDNOW, double* FIELDNEXT,double* N5NOW, Parameters* param )
	{
		
	//CHECKED ON MAY 22nd 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		double SUNCLMUCNST=CurParam->ConstantChemicalPotential;
		double SUNNUMBFERM=CurParam->NFermions;
		
		int DYNAMICALN5=CurParam->DYNAMICALN5;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;	
					
		//-----------------------------------------------------------------------
		//	THE LANGEVIN TIME EVOLUTION FOR THE GAUGE FIELD
		//-----------------------------------------------------------------------
		
		Matrix** Fmunu;
		
		Fmunu = new Matrix* [4];
		for (int k=0;k<4;k++) Fmunu[k]= new Matrix [4];
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			  
				FMUNU_SPATIAL( Fmunu, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);			
			  
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double EEVOLFAC=SUNCLDT/(sqrt(SUN)*SUNCLDX*SUNCLDX);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				int64_t POSN5=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					double BCURR[SUNGENNUM];
					
					if(mu==0){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[1][2].ReTrGen(SUNGENCNT);
					}else if (mu==1){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[2][0].ReTrGen(SUNGENCNT);
					}else if(mu==2){
						for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) BCURR[SUNGENCNT]=-2.*Fmunu[0][1].ReTrGen(SUNGENCNT);
					}
					
					double TMP[SUNGENNUM];
					double Bfac=(SUNNUMBFERM*SUNCLCOUPL*SUNCLCOUPL/(4.*PI*PI));

					//-----------------------------------------------------------------------
					//	GET THE MU5 APPROXIMATELY FROM N5
					//
					//	CHECK WHETHER WE HAVE THE SINGLE REAL SOLUTION FOR MU5
					//	THEN CALCULATE THIS VALUE
					//-----------------------------------------------------------------------

					double CURN5;
					
					if(DYNAMICALN5==1) { 
					  
					  CURN5=N5NOW[0]; 
					  
					} else if (DYNAMICALN5==2) { 
					  
					  CURN5=N5NOW[POSN5]; 
					  
					} else {
					  
					  CURN5=0.;
					  
					}
					
					double Ttmp=2.*SUN/(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX*SUNCLBETA);
					
					double c=(3.*SUNCLMUCNST*SUNCLMUCNST*SUNCLMUCNST + PI*PI*Ttmp*Ttmp);
					double d=-CURN5*3.*PI*PI/(SUN*SUNNUMBFERM);
									
					double DSCR=-4.*c-27.*d*d;
					
					if(DSCR>0){
						fprintf(stderr, "PROBLEM WITH DETERMINING A SINGLE REAL ROOT FOR MU5\n"); exit(-1);
					}
					
					double CapC=exp((1./3.)*log( 0.5 * ( 27.*d + sqrt(729.*d*d+108.*c*c*c) ) ) );
					double D0=-3.*c;
					
					double MU5CUR=-(1./3.)*(CapC+D0/CapC);				
					
					double F0iECOnv=(SUNCLCOUPL*SUNCLDX/(2.*sqrt(SUN)));
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++){
						
						//BEWARE NO DAMPING IN HERE!!!
					  
						FIELDNEXT[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]=FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+SUNGENCNT]-EEVOLFAC*PSum.ImTrGen(SUNGENCNT) - SUNCLDT*MU5CUR*F0iECOnv*Bfac*BCURR[SUNGENCNT];
					
					}

				}			
				
			}
		    }
		}
	}

	void SUN_CLASS_EVOLVE_N5_NOSPATIAL(double* N5NOW, double* N5NEXT, double CURDNCS, Parameters* param )
	{
	  
	//CHECKED ON MAY 27th 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		double SUNNUMBFERM=CurParam->NFermions;
				
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		double VOLFAC=AXLEN[0]*AXLEN[1]*AXLEN[2]*SUNCLDX*SUNCLDX*SUNCLDX;

		double EBFac=SUNNUMBFERM*2./VOLFAC;
		
		N5NEXT[0] = N5NOW[0] - SUNCLDT*EBFac*CURDNCS;			
			
		//IN MOORE's DEFINITION OF \partial_mu j^mu he has \propto - F F^tilde
		
	}
	


    void SUN_CLASS_EVOLVE_N5_SPATIAL(double* FIELDNOW, double* FIELDNEXT,double* FIELDBEF, double* N5NOW, double* N5NEXT, double* N5SCRATCH, TLP_RANDGEN** Rng, Parameters* param )
	{
	  
	//NEED TO CHECK FOR MINUS SIGN 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		double SUNCOLORCONDUCT=CurParam->ColorConductivity;
		double SUNCLDIFFCONST=CurParam->DiffusionConst;
		double SUNCLMUCNST=CurParam->ConstantChemicalPotential;
		double SUNNUMBFERM=CurParam->NFermions;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;	

		//-----------------------------------------------------------------------
		//	AUXILIARY ARRAYS FOR THE ADI TRIDIAGONAL SOLVER
		//-----------------------------------------------------------------------
		
		int NX=AXLEN[0];
		
		if(AXLEN[0]!=AXLEN[1]||AXLEN[1]!=AXLEN[2]||AXLEN[2]!=AXLEN[0]){
			fprintf(stderr, "ASSUMING A LATTICE WITH EQUAL SPATIAL EXTEND IN EACH DIMENSION\n"); exit(-1);
		}
		
		double* SubDiag, *SupDiag, *Diag, *x1, *x2, *q, *qprime, *scratch;
		
		SubDiag =  new double [NX];
		SupDiag =  new double [NX];
		Diag =  new double [NX];
		x1 =  new double [NX];
		x2 =  new double [NX];
		q =  new double [NX];
		qprime =  new double [NX];
		scratch =  new double [NX];
		
		double nuConst=SUNCLDIFFCONST*SUNCLDT/(SUNCLDX*SUNCLDX);
		
		for(int k=0;k<NX;k++) qprime[k]=0.;
		qprime[0]=nuConst/2.;
		qprime[NX-2]=nuConst/2.;
		for(int k=0;k<NX;k++){
			SubDiag[k]=-nuConst/2.;
			SupDiag[k]=-nuConst/2.;
			Diag[k]=1.+nuConst;
			
			//fprintf(stderr, "%d - %e %e %e %e\n", k, SubDiag[k],Diag[k],SupDiag[k],qprime[k]);
			
		}

		//-----------------------------------------------------------------------
		//	THE LANGEVIN TIME EVOLUTION FOR THE GAUGE FIELD
		//-----------------------------------------------------------------------
		
		Matrix** Fmunu;
		
		Fmunu = new Matrix* [4];
		for (int k=0;k<4;k++) Fmunu[k]= new Matrix [4];
		
		//-----------------------------------------------------------------------
		//	DIFFUSION EQUATION FOR THE CHIRAL NUMBER DENSITY
		//-----------------------------------------------------------------------
		
		//-----------------------------------------------------------------------
		//	FIRST STEP ADI: X-DIRECTION IMPLICIT + HETEROGENEOUS TERM
		//-----------------------------------------------------------------------
		
		for  (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
		
		  for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=0;
				int64_t POSMU=0;

				int64_t POSMUP1X=0;
				int64_t POSMUM1X=0;
				int64_t POSMUP1Y=0;
				int64_t POSMUM1Y=0;
				int64_t POSMUP1Z=0;
				int64_t POSMUM1Z=0;
				
				for(int k=0;k<NDIM;k++){
					
					POS+= ACNT[k]*STRIDES[k];
					POSMU+= ACNT[k]*STRIDES[k];
					
					POSMUP1X+=( ( ACNT[k] + ((k==0)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1X+=( ( ACNT[k] - ((k==0)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Y+=( ( ACNT[k] + ((k==1)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Y+=( ( ACNT[k] - ((k==1)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Z+=( ( ACNT[k] + ((k==2)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Z+=( ( ACNT[k] - ((k==2)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				//-----------------------------------------------------------------------
				//	SINCE WE HAVE EVOLVED THE UFIELDS ALREADY THE EVOLUTION EQUATION
				//	FOR N5 HAS TO CALCULATE THE TOPOLOGICAL TERM ExB FROM THE
				//	FIELDNEXT AND FIELDNOW ARRAY AND NOT FROM FIELDNOW AND FIELDBEF
				//-----------------------------------------------------------------------
				
				FMUNU_SPATIAL( Fmunu, FIELDNEXT, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);	

				FMUNU_TMPORAL( Fmunu, FIELDNEXT, NULL, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL,1);	

				Complex FFtildeTr;
				FFtildeTr=0.;
					
				for(int64_t k=0;k<24;k++){
					
					Matrix TMP;
					
					TMP=Fmunu[mucomp[k]][nucomp[k]];
					TMP*=Fmunu[sgcomp[k]][rocomp[k]];
					
					FFtildeTr+=-0.5*LCval[k]*TMP.Trace();
				}
				
				double EBFac=SUNNUMBFERM*SUNCLCOUPL*SUNCLCOUPL/(4.*PI*PI);
				
				//1-D
	//			q[ACNT[0]]= (1.-nuConst)*N5NOW[POSMU]+ (nuConst/2.)*( N5NOW[POSMUP1X] + N5NOW[POSMUM1X] );

				//2-D
	// 			q[ACNT[0]]= (1.-nuConst)*N5NOW[POSMU]+ (nuConst/2.)*( N5NOW[POSMUP1Y] + N5NOW[POSMUM1Y] );
				
				//3-D
				q[ACNT[0]]= 		(1.-5.*nuConst)*N5NOW[POSMU]
						+ 	(nuConst/2.)*( N5NOW[POSMUP1X]+ N5NOW[POSMUM1X] )
						+	nuConst*( N5NOW[POSMUP1Y]+N5NOW[POSMUP1Z] + N5NOW[POSMUM1Y]+N5NOW[POSMUM1Z] )
						+	SUNCLDT*EBFac*FFtildeTr.real();			

				x1[ACNT[0]]=q[ACNT[0]];
				x2[ACNT[0]]=qprime[ACNT[0]];
				
			}
			
			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);		
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]-1; ACNT[0]++){
				
				x1[ACNT[0]]+=x2[ACNT[0]]*x1[NX-1];
				
			}	
			
			for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
				
				int64_t POSMU=0;
				
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				//2-D
				//N5SCRATCH[POSMU]=x1[ACNT[0]];
				
				//3-D
				N5NEXT[POSMU]=x1[ACNT[0]];
			
			}
		    }
		}
		

		//-----------------------------------------------------------------------
		//	SECOND STEP ADI: Y-DIRECTION IMPLICIT 
		//-----------------------------------------------------------------------
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
		
		  for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSMU=0;

	// 			int64_t POSMUP1X=0;
	// 			int64_t POSMUM1X=0;
				
				int64_t POSMUP1Y=0;
				int64_t POSMUM1Y=0;
				
				for(int k=0;k<NDIM;k++){
					
					POSMU+= ACNT[k]*STRIDES[k];
					
	// 				POSMUP1X+=( ( ACNT[k] + ((k==0)?1:0) ) %AXLEN[k] )*STRIDES[k];
	// 				POSMUM1X+=( ( ACNT[k] - ((k==0)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
					POSMUP1Y+=( ( ACNT[k] + ((k==1)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Y+=( ( ACNT[k] - ((k==1)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				//2-D
				//q[ACNT[1]]= (1.-nuConst)*N5SCRATCH[POSMU]+ (nuConst/2.)*( N5SCRATCH[POSMUP1X] + N5SCRATCH[POSMUM1X] );
				
				//3-D
				q[ACNT[1]]= N5NEXT[POSMU]+ nuConst*N5NOW[POSMU]- (nuConst/2.)*( N5NOW[POSMUP1Y]+ N5NOW[POSMUM1Y] );
				
				x1[ACNT[1]]=q[ACNT[1]];
				x2[ACNT[1]]=qprime[ACNT[1]];
				
			}

			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]-1; ACNT[1]++){
				
				x1[ACNT[1]]+=x2[ACNT[1]]*x1[NX-1];
				
			}
			
			for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
				
				int64_t POSMU=0;
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				//2-D
	// 			N5NEXT[POSMU]=x1[ACNT[1]];

				//3-D
				N5SCRATCH[POSMU]=x1[ACNT[1]];
			
			}
		    }
		}

		//-----------------------------------------------------------------------
		//	THIRD STEP ADI: Z-DIRECTION IMPLICIT 
		//-----------------------------------------------------------------------
		
		for  (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
		
		  for (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
		
				//-----------------------------------------------------------------------
				//	AUXILIARY POSITIONS FOR THE SECOND DERIVATIVES
				//-----------------------------------------------------------------------
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSMU=0;

				int64_t POSMUP1Z=0;
				int64_t POSMUM1Z=0;
				
				for(int k=0;k<NDIM;k++){
					
					POSMU+= ACNT[k]*STRIDES[k];

					POSMUP1Z+=( ( ACNT[k] + ((k==2)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POSMUM1Z+=( ( ACNT[k] - ((k==2)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					
				}
				
				q[ACNT[2]]= N5SCRATCH[POSMU]+ nuConst*N5NOW[POSMU]- (nuConst/2.)*( N5NOW[POSMUP1Z]+ N5NOW[POSMUM1Z] );
				
				x1[ACNT[2]]=q[ACNT[2]];
				x2[ACNT[2]]=qprime[ACNT[2]];
			}

			//-----------------------------------------------------------------------
			//	TWO STEPS OF THE THOMAS ALGORITHM
			//-----------------------------------------------------------------------

			NFIVE_DIFF_TRIDIAG_SOLVER(x1, NX-1, SubDiag, Diag, SupDiag , scratch);
			NFIVE_DIFF_TRIDIAG_SOLVER(x2, NX-1, SubDiag, Diag, SupDiag, scratch);
			
			x1[NX-1]=(q[NX-1]-(-nuConst/2.)*x1[0]-(-nuConst/2.)*x1[NX-2])/( Diag[NX-1]+(-nuConst/2.)*x2[0]+(-nuConst/2.)*x2[NX-2] );
				
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]-1; ACNT[2]++){
				
				x1[ACNT[2]]+=x2[ACNT[2]]*x1[NX-1];
				
			}
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
				
				int64_t POSMU=0;
				for(int k=0;k<NDIM;k++){
					POSMU+= ACNT[k]*STRIDES[k];
				}
				
				N5NEXT[POSMU]=x1[ACNT[2]];
			
			}
		    }
		}

		delete[] SubDiag;
		delete[] SupDiag;
		delete[] Diag;
		delete[] x1;
		delete[] x2;
		delete[] q;
		delete[] qprime;
		delete[] scratch;
		
		for (int k=0;k<4;k++) delete[] Fmunu[k];
		delete[] Fmunu;
		
		
		
	}	

//-----------------------------------------------------------------------
//	COOLING AND COARSENING
//-----------------------------------------------------------------------

    void LOAD_UFIELD_COARSENDED(double* FIELDCOARSE, double* FIELDIN, Parameters* param)
	{
	
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		int64_t AXLENCOARSE[NDIM];
		int64_t STRIDESCOARSE[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		AXLENCOARSE[0]=AXLEN[0]/2;
		AXLENCOARSE[1]=AXLEN[1]/2;
		AXLENCOARSE[2]=AXLEN[2]/2;
		
		double SUNCLDTAU=CurParam->dtau;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
	  
		STRIDESCOARSE[0]=AXLENCOARSE[2]*AXLENCOARSE[1];
		STRIDESCOARSE[1]=AXLENCOARSE[2];
		STRIDESCOARSE[2]=1;
	  
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]+=2){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]+=2){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]+=2){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				Matrix TMP1,TMP2;
				
				for(int64_t mu=0; mu<NDIM; mu++){
		      
				      int64_t POS2=0;	
				      for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
				      POS2*=ESTRIDE+SUNGENNUM*NDIM;			  

				      int64_t POS3=0;	
				      for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k]/2 ) )*STRIDESCOARSE[k];
				      POS3*=ESTRIDE+SUNGENNUM*NDIM;			  
				      
				      TMP1=&FIELDIN[POS + (mu)*USTRIDE];
				      TMP2=&FIELDIN[POS2 + (mu)*USTRIDE];
				      
				      TMP1*=TMP2;
				      
				      for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDCOARSE[POS3 + (mu)*USTRIDE + CMP]=TMP1[CMP];  
				  
				}
				
			}
		    }
		}
	  
	}



    void SUN_CLASS_UFIELD_COOLING_INPLACE(double* FIELDNOW, double* FIELDTMP, Parameters* param )
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDTAU=CurParam->dtau;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					double TMP[SUNGENNUM];
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) TMP[SUNGENCNT]=PSum.ImTrGen(SUNGENCNT);
					
					E.FROM_REAL_SUN_GENERATORS(TMP);

	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[1]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[2]*=-2.*SUNCLDTAU*SUNCLDX;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
				
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=-SUNCLDTAU*SUNCLDX*sqrt(rsqr);
					
					Complex SinFac(0.,0.);
					
					if(rsqr>5e-15){				
					  Complex SinFacSafe(0.,2.*sin(r)/sqrt(rsqr));  
					  SinFac+=SinFacSafe;
					}	
					
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;				
				
	#endif				
					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDTMP[POS + (mu)*USTRIDE + CMP]=E[CMP];

				}			
				
			}
		    }
		}
		
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]=FIELDTMP[POS + (mu)*USTRIDE + CMP];
					
				}
				
			}
		    }
		}
		
	}

    void SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(double* FIELDNOW, double* FIELDTMP, Parameters* param )
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDTAU=CurParam->dtau;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		double SUNCLCOOLEPS=CurParam->ConditioningCooling;
		
		const double PREFAC1=(4.-SUNCLCOOLEPS)/3.;
		const double PREFAC2=(SUNCLCOOLEPS-1.)/48.;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,PSum22,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					PSum22.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION 
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?2:0) ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==mu)?2:0) + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] + ((k==mu)?1:0) + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						int64_t POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1 = &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS3 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS4 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS7 + (nu)*USTRIDE]);
						P1*=TMP;					
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum22+=P1;

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						
						P1 = &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS3 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS4 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (nu)*USTRIDE]);
						P1*=TMP;					
						TMP.Dagger(&FIELDNOW[POS7 + (nu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS7 + (mu)*USTRIDE];
						
						PSum22+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION 
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?2:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==mu)?2:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS4 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS6 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS7 + (nu)*USTRIDE];
						
						PSum22+=P1;

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] - ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS4 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS5 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS6 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS7 + (mu)*USTRIDE];
						
						PSum22+=P1;
						
					}}
					
					double TMP[SUNGENNUM];
					
					for(int64_t SUNGENCNT=0; SUNGENCNT<SUNGENNUM; SUNGENCNT++) TMP[SUNGENCNT]= PREFAC1*PSum.ImTrGen(SUNGENCNT)+PREFAC2*PSum22.ImTrGen(SUNGENCNT);
					
					E.FROM_REAL_SUN_GENERATORS(TMP);

	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[1]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[2]*=-2.*SUNCLDTAU*SUNCLDX;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=-SUNCLDTAU*SUNCLDX*sqrt(rsqr);
					
					Complex SinFac(0.,0.);
					
					if(rsqr>5e-15){				
					  Complex SinFacSafe(0.,2.*sin(r)/sqrt(rsqr));  
					  SinFac+=SinFacSafe;
					}	

					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;		
					
	#endif				
					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDTMP[POS + (mu)*USTRIDE + CMP]=E[CMP];

				}			
				
			}
		    }
		}
		
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]=FIELDTMP[POS + (mu)*USTRIDE + CMP];
					
				}
				
			}
		    }
		}
		
	}


    void SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(double* FIELDNOW, double* FIELDTMP, Parameters* param )
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDTAU=CurParam->dtau;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;

		double SUNCLCOOLEPS=CurParam->ConditioningCooling;
		
		const double PREFAC1=(4.-SUNCLCOOLEPS)/3.;
		const double PREFAC2=(SUNCLCOOLEPS-1.)/48.;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		for(int RKSTEP=0;RKSTEP<3;RKSTEP++){
		
		//fprintf(stderr, "DEBUG RUNGE KUTTA STEP: %d\n",RKSTEP);
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,PSum22,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					PSum22.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?2:0) ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==mu)?2:0) + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] + ((k==mu)?1:0) + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						int64_t POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1 = &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS3 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS4 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS7 + (nu)*USTRIDE]);
						P1*=TMP;					
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum22+=P1;

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						
						P1 = &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS3 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS4 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (nu)*USTRIDE]);
						P1*=TMP;					
						TMP.Dagger(&FIELDNOW[POS7 + (nu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS7 + (mu)*USTRIDE];
						
						PSum22+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?2:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] + ((k==mu)?2:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*= &FIELDNOW[POS2 + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS4 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS6 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS6 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS7 + (nu)*USTRIDE];
						
						PSum22+=P1;

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;	
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS4=0;
						for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS5=0;
						for (int k=0;k<NDIM;k++) POS5+= ( ( ACNT[k] - ((k==mu)?1:0) - ((k==nu)?2:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS5*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS6=0;
						for (int k=0;k<NDIM;k++) POS6+= ( ( ACNT[k] - ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS6*=ESTRIDE+SUNGENNUM*NDIM;
						
						POS7=0;
						for (int k=0;k<NDIM;k++) POS7+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS7*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS4 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS5 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*= &FIELDNOW[POS5 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS6 + (nu)*USTRIDE];
						P1*= &FIELDNOW[POS7 + (mu)*USTRIDE];
						
						PSum22+=P1;
						
					}}
					
					double TMP[SUNGENNUM];
					
					if(RKSTEP==0){
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++){
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]=0.25*( PREFAC1*PSum.ImTrGen(CMP)+PREFAC2*PSum22.ImTrGen(CMP)   );
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==1) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-17./9.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(8./9.)*( PREFAC1*PSum.ImTrGen(CMP)+PREFAC2*PSum22.ImTrGen(CMP)   );
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==2) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-1.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(3./4.)*( PREFAC1*PSum.ImTrGen(CMP)+PREFAC2*PSum22.ImTrGen(CMP)   );
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					}				
					
					E.FROM_REAL_SUN_GENERATORS(TMP);

	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[1]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[2]*=-2.*SUNCLDTAU*SUNCLDX;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=-SUNCLDTAU*SUNCLDX*sqrt(rsqr);
					
					Complex SinFac(0.,0.);
					
					if(rsqr>5e-15){				
					  Complex SinFacSafe(0.,2.*sin(r)/sqrt(rsqr));  
					  SinFac+=SinFacSafe;
					}	
					
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;		
					
	#endif				
					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDTMP[POS + (mu)*USTRIDE + CMP]=E[CMP];

				}			
				
			}
		    }
		}
		
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]=FIELDTMP[POS + (mu)*USTRIDE + CMP];
					
				}
				
			}
		    }
		}
		
		}
		
	}

    void SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(double* FIELDNOW, double* FIELDTMP, Parameters* param )
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDTAU=CurParam->dtau;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		double SUNCLBETA=CurParam->BetaL;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		for(int RKSTEP=0;RKSTEP<3;RKSTEP++){
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){
				
					//-----------------------------------------------------------------------
					//	CALCULATE THE PLAQUETTE SUM
					//-----------------------------------------------------------------------
					
					Matrix PSum,U,E;
					Matrix P1;
					Matrix Unit;
					
					PSum.Zero();
					
					for(int64_t cnt=1; cnt<NDIM; cnt++){
						
						{
						int64_t nu=(mu+cnt)%NDIM;
							
						//-----------------------------------------------------------------------
						//	FORWARD DIRECTION U(x)_mu U(x+mu)_nu U^\Dagg(x+nu)_mu U^\Dagg(x)_nu
						//-----------------------------------------------------------------------

						Matrix TMP;
						
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						PSum+=P1;
						
						//-----------------------------------------------------------------------
						//	BACKWARD DIRECTION U(x)_mu U^Dagg(x-nu+mu)_nu U^\Dagg(x-nu)_mu U(x-nu)_mu
						//-----------------------------------------------------------------------

						POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS + (mu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
						P1*=TMP;
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						
						PSum+=P1;
						
					}}

					double TMP[SUNGENNUM];
					
					if(RKSTEP==0){
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++){
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]=0.25*PSum.ImTrGen(CMP);
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==1) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-17./9.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(8./9.)*PSum.ImTrGen(CMP);
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					} else if (RKSTEP==2) {
						for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) {
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]*=-1.;
						FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP]+=(3./4.)*PSum.ImTrGen(CMP);
						TMP[CMP]=FIELDTMP[POS+ESTRIDE+SUNGENNUM*mu+CMP];
						}
					}				
					
					E.FROM_REAL_SUN_GENERATORS(TMP);

	#if (SUN==3)					
						
					double eval[3];				
					double __complex__ EMat[3][3];
					double __complex__ EVec[3][3];
					double __complex__ EVec2[3][3];
					double __complex__ Final[3][3];
					double __complex__ ExpEvals[3];
					
					EMat[0][0]=* ( (double __complex__ *) &E(0 ,0) );
					EMat[1][0]=* ( (double __complex__ *) &E(1 ,0) );
					EMat[2][0]=* ( (double __complex__ *) &E(2 ,0) );
					EMat[0][1]=* ( (double __complex__ *) &E(0 ,1) );
					EMat[1][1]=* ( (double __complex__ *) &E(1 ,1) );
					EMat[2][1]=* ( (double __complex__ *) &E(2 ,1) );
					EMat[0][2]=* ( (double __complex__ *) &E(0 ,2) );
					EMat[1][2]=* ( (double __complex__ *) &E(1 ,2) );
					EMat[2][2]=* ( (double __complex__ *) &E(2 ,2) );
					
					zheevh3( EMat, EVec, eval);
										
					eval[0]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[1]*=-2.*SUNCLDTAU*SUNCLDX;
					eval[2]*=-2.*SUNCLDTAU*SUNCLDX;
					
					ExpEvals[0]=cos(eval[0])+I*sin(eval[0]);
					ExpEvals[1]=cos(eval[1])+I*sin(eval[1]);
					ExpEvals[2]=cos(eval[2])+I*sin(eval[2]);
					
					EVec2[0][0]=conj(EVec[0][0])*ExpEvals[0];
					EVec2[1][0]=conj(EVec[0][1])*ExpEvals[1];
					EVec2[2][0]=conj(EVec[0][2])*ExpEvals[2];
					EVec2[0][1]=conj(EVec[1][0])*ExpEvals[0];
					EVec2[1][1]=conj(EVec[1][1])*ExpEvals[1];
					EVec2[2][1]=conj(EVec[1][2])*ExpEvals[2];
					EVec2[0][2]=conj(EVec[2][0])*ExpEvals[0];
					EVec2[1][2]=conj(EVec[2][1])*ExpEvals[1];
					EVec2[2][2]=conj(EVec[2][2])*ExpEvals[2];
					
					*((double __complex__*) &E(0,0)) = EVec2[0][0]*EVec[0][0]+ EVec2[1][0]*EVec[0][1] + EVec2[2][0]*EVec[0][2];
					*((double __complex__*) &E(1,0))= EVec2[0][0]*EVec[1][0]+ EVec2[1][0]*EVec[1][1] + EVec2[2][0]*EVec[1][2];
					*((double __complex__*) &E(2,0))= EVec2[0][0]*EVec[2][0]+ EVec2[1][0]*EVec[2][1] + EVec2[2][0]*EVec[2][2];
					
					*((double __complex__*) &E(0,1))= EVec2[0][1]*EVec[0][0]+ EVec2[1][1]*EVec[0][1] + EVec2[2][1]*EVec[0][2];
					*((double __complex__*) &E(1,1))= EVec2[0][1]*EVec[1][0]+ EVec2[1][1]*EVec[1][1] + EVec2[2][1]*EVec[1][2];
					*((double __complex__*) &E(2,1))= EVec2[0][1]*EVec[2][0]+ EVec2[1][1]*EVec[2][1] + EVec2[2][1]*EVec[2][2];
					
					*((double __complex__*) &E(0,2))= EVec2[0][2]*EVec[0][0]+ EVec2[1][2]*EVec[0][1] + EVec2[2][2]*EVec[0][2];
					*((double __complex__*) &E(1,2))= EVec2[0][2]*EVec[1][0]+ EVec2[1][2]*EVec[1][1] + EVec2[2][2]*EVec[1][2];
					*((double __complex__*) &E(2,2))= EVec2[0][2]*EVec[2][0]+ EVec2[1][2]*EVec[2][1] + EVec2[2][2]*EVec[2][2];

	#elif (SUN==2)
					double rsqr=0;
					
					for(int k=0;k<SUNGENNUM;k++) rsqr+=TMP[k]*TMP[k];
					
					double r=-SUNCLDTAU*SUNCLDX*sqrt(rsqr);
					
					Complex SinFac(0.,0.);
					
					if(rsqr>5e-15){				
					  Complex SinFacSafe(0.,2.*sin(r)/sqrt(rsqr));  
					  SinFac+=SinFacSafe;
					}	
					
					double  CosFac=cos(r);
					
					Matrix CosMat;					
					CosMat.Unit();
					CosMat*=CosFac;
					E*=SinFac;
					E+=CosMat;		
					
	#endif				
					U= &FIELDNOW[POS + (mu)*USTRIDE];
					E*=U;

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDTMP[POS + (mu)*USTRIDE + CMP]=E[CMP];

				}			
				
			}
		    }
		}
		
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);
				
				for(int64_t mu=0; mu<NDIM; mu++){

					for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS + (mu)*USTRIDE + CMP]=FIELDTMP[POS + (mu)*USTRIDE + CMP];
					
				}
				
			}
		    }
		}
		
		}
		
	}

    void SUN_CLASS_FOURIER_COOL_INPLACE(double* FIELDNOW, double MAXMOM, Parameters* param)
	{

		//-----------------------------------------------------------------------
		//	I am grateful to Sabolcs Borsanyi for providing me with the
		//	original 3D counting scheme for the negative momenta 
		//-----------------------------------------------------------------------
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		int64_t MCNT[NDIM];	//COUNTERS FOR THE NEGATIVE MOMENTA
		int64_t PCNT;		//SINGLE COUNTER FOR THE POSITIVE MOMENTA
			
		int AXLENFFT[NDIM];	//AXES LENGTH FOR THE FFT PLANS
		double KVAL[NDIM];	//POSITIVE MOMENTA VALUES
		double MKVL[NDIM];	//NEGATIVE MOMENTA VALUES

		fftw_plan FFT;		//FFTW PLAN FOR THE FORWARD TRANSFORMATION
		fftw_plan IFFT;		//FFTW PLAN FOR THE BACKWARD TRANSFORMATION
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		AXLENFFT[0]=CurParam->NX;
		AXLENFFT[1]=CurParam->NY;
		AXLENFFT[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		int64_t IVOL=AXLEN[0]*AXLEN[1]*AXLEN[2];		
		
		//*********************************************************************************
		//  SETUP THE FOURIER SPACE MOMENTA, WHICH NEED TO BE SET SUCH THAT THE NYQUIST
		//  FREQUENCY IS CORRECTLY CAPTURED AT AXLEN/2
		//*********************************************************************************
		
		double **PPOS;
		PPOS=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) PPOS[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<=AXLEN[DIM]/2;PCNT++) PPOS[DIM][PCNT]=PCNT*((2.*PI)/(((double)AXLEN[DIM])*((double)SUNCLDX)));
			for(int PCNT=AXLEN[DIM]/2+1;PCNT<AXLEN[DIM];PCNT++) PPOS[DIM][PCNT]=-PPOS[DIM][AXLEN[DIM]-PCNT];
		}

		double **MOM;
		MOM=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) MOM[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<AXLEN[DIM];PCNT++) MOM[DIM][PCNT]=sqrt(2.5-(8./3.)*cos(2.*PCNT*PI/((double)AXLEN[DIM]))+(1./6.)*cos(4.*PCNT*PI/((double)AXLEN[DIM])))/SUNCLDX;
		}
		
		//*********************************************************************************
		// THE ARRAY FOR THE A FIELDS
		//*********************************************************************************
		
		int64_t MEMTOT=sizeof(fftw_complex)*NDIM*SUNGENNUM*IVOL;
		
		fftw_complex *AFIELDS,*EFIELDS, *CURFIELD;
		
		AFIELDS= new  fftw_complex [NDIM*SUNGENNUM*IVOL];
		EFIELDS= new  fftw_complex [NDIM*SUNGENNUM*IVOL];
		
		//*********************************************************************************
		// EXTRACT FROM THE LINK VARIABLES THE A FIELD
		//
		// BE AWARE THAT ONCE THE FIELD VALUES ARE LARGER THAN UNITY, THE SIMPLE
		// FORMULA U=1+iagA IS NOT APPLICABLE ANYMORE TO PROJECT OUT A. INDEED ONE HAS
		// TO CALCULATE THE COMPLEX LOGARITHM OF THE LINK U!
		//
		//*********************************************************************************

		char dsjobl='N';
		char dsjobr='V';
		
		int size=SUN;
		
		int *piv;
		piv = new int [SUN];
		
		Complex *dsw;
		dsw = new Complex [SUN];
		
		Complex *dsvr;
		dsvr = new Complex [SUN*SUN];
		
		Complex *dsvl;
		dsvl = new Complex [SUN*SUN];
		
		double *dsrwork;
		dsrwork= new double [2*SUN];
		
		int dsinfo;
		
		int dslda=SUN;
		int dslvr=SUN;
		int dslvl=SUN;
		int dslw=2*SUN;
					
		Complex *dswork;
		dswork= new Complex [2*SUN*SUN];	

		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		int64_t pos0=ACNT[0]*STRIDES[0];
		
		for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
				
			int64_t USTRIDE=2*SUN*SUN;
			int64_t ESTRIDE=2*SUN*SUN*NDIM;
			
			int64_t POS1=pos1 + ACNT[2]*STRIDES[2];
			
			int64_t POS2=POS1;
			POS2*=(ESTRIDE+SUNGENNUM*NDIM);
			
			double TMP[SUNGENNUM];
			
			for(int64_t mu=0;mu<NDIM;mu++){
				
				Matrix U,EigenVec;
			
				U=&FIELDNOW[POS2 + (mu)*USTRIDE];
				
				SetTranspose(&U[0] ,&FIELDNOW[POS2 + (mu)*USTRIDE]);		
				
				zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
				
				SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
		
				for(int i=0;i<SUN;i++){	

					double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
					double logabs=log(aval);
					double angle=atan2(dsw[i].imag(),dsw[i].real());
					
					dsw[i].real(logabs);
					dsw[i].imag(angle);
				}
				
				
				U=EigenVec;
				
				for(int i=0;i<SUN;i++){
				for(int j=0;j<SUN;j++){
				
					U(i,j)*=dsw[j];
				
				}
				}
				
				EigenVec.Dagger();
				U*=EigenVec;
				
				for(int64_t k=0; k<SUNGENNUM; k++) TMP[k]=2.*U.ImTrGen(k)/(SUNCLCOUPL*SUNCLDX); 

				for(int64_t k=0; k<SUNGENNUM; k++){
					
					if(TMP[k]!=TMP[k]) {fprintf(stderr, "NAN in A encountered\n"); exit(-1);}
					if(FIELDNOW[POS2 + ESTRIDE + (mu)*SUNGENNUM+k]!=FIELDNOW[POS2 + ESTRIDE + (mu)*SUNGENNUM+k]) {fprintf(stderr, "NAN in E encountered\n"); exit(-1);}
					
					AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]=TMP[k];
					AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][1]=0.;
									
					EFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]=FIELDNOW[POS2 + ESTRIDE + (mu)*SUNGENNUM+k];
					EFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][1]=0.;

				}

			}
			
		}}}

		delete[] piv;
		delete[] dsw;
		delete[] dsvr;
		delete[] dsvl;
		delete[] dsrwork;
		delete[] dswork;
		
		//*********************************************************************************
		//	CUT THE FREQUENCY COMPONENTS ABOVE THE MAXIMUM ALLOWED MOMENTUM
		//*********************************************************************************
		
		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				CURFIELD=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				FFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,-1, FFTW_ESTIMATE);
				
				fftw_execute(FFT);

				//-----------------------------------------------------------------------
				//	MCNT, the counter for the negative frequencies is always 
				//	incremented at the end of each loop step since,
				//	the zero position does not have a counterpart (the NYQUIST
				//	FREQUENCY takes this role) and +K <-> -K sorting starts from
				//	step 1 NOT 0.
				//-----------------------------------------------------------------------
				
				for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				  for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
						
					if(PCNT<=MCNT[2]){
						
						for(int i=0; i<NDIM; i++){				
							KVAL[i]=PPOS[i][ACNT[i]];						
							if(ACNT[i]!=AXLEN[i]/2) {MKVL[i]=-KVAL[i];}else{MKVL[i]=KVAL[i];}
						}
						
						double KABSS=0.;
						for(int i=0;i<NDIM;i++) KABSS+=pow(PPOS[i][ACNT[i]],2);
						double KABS=sqrt(KABSS);
						
						if(KABS>MAXMOM){
							
							CURFIELD[PCNT][1]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));					
							CURFIELD[PCNT][0]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
							
							if(PCNT!=MCNT[2]){
								CURFIELD[MCNT[2]][1]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
								CURFIELD[MCNT[2]][0]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
							}
						
						}
					}
				      }
				  }
				}

				fftw_destroy_plan(FFT);
				
				CURFIELD=&EFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				FFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,-1, FFTW_ESTIMATE);
				
				fftw_execute(FFT);

				//-----------------------------------------------------------------------
				//	MCNT, the counter for the negative frequencies is always 
				//	incremented at the end of each loop step since,
				//	the zero position does not have a counterpart (the NYQUIST
				//	FREQUENCY takes this role) and +K <-> -K sorting starts from
				//	step 1 NOT 0.
				//-----------------------------------------------------------------------
				
				for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				  for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
						
					if(PCNT<=MCNT[2]){
						
						for(int i=0; i<NDIM; i++){				
							KVAL[i]=PPOS[i][ACNT[i]];						
							if(ACNT[i]!=AXLEN[i]/2) {MKVL[i]=-KVAL[i];}else{MKVL[i]=KVAL[i];}
						}
						
						double KABSS=0.;
						for(int i=0;i<NDIM;i++) KABSS+=pow(PPOS[i][ACNT[i]],2);
						double KABS=sqrt(KABSS);

						if(KABS>MAXMOM){
							
							CURFIELD[PCNT][1]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));					
							CURFIELD[PCNT][0]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
							
							if(PCNT!=MCNT[2]){
								CURFIELD[MCNT[2]][1]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
								CURFIELD[MCNT[2]][0]*=exp(-(KABS-MAXMOM)*(AXLEN[0]/2));
							}
						
						}
					}
				      }
				  }
				}
				
				fftw_destroy_plan(FFT);

			}
		}

		//*********************************************************************************
		//	USE THE FOURIER SPACE CUTTED A FIELDS TO SET UP THE LATTICE LINK VARIABLES
		//*********************************************************************************

		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				CURFIELD=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				IFFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,+1, FFTW_ESTIMATE);
				
				fftw_execute(IFFT);
				
				fftw_destroy_plan(IFFT);
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				CURFIELD=&EFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];			
				
				IFFT=fftw_plan_dft(3, AXLENFFT, CURFIELD, CURFIELD,+1, FFTW_ESTIMATE);
				
				fftw_execute(IFFT);
				
				fftw_destroy_plan(IFFT);
				
			}
		}

		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		int64_t pos0=ACNT[0]*STRIDES[0];
		
		for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			
			int64_t USTRIDE=2*SUN*SUN;
			int64_t ESTRIDE=2*SUN*SUN*NDIM;
			
			int64_t POS1=pos1 + ACNT[2]*STRIDES[2];
			int64_t POS2=POS1;
			POS2*=(ESTRIDE+SUNGENNUM*NDIM);
			
			double TMP[SUNGENNUM];
			
			for(int64_t mu=0;mu<NDIM;mu++){
				
				for(int64_t k=0; k<SUNGENNUM; k++) TMP[k]=AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]/VOL;
				
				Matrix U;
				U.FROM_REAL_SUN_GENERATORS(TMP);
				
				Complex Ufac(0.,SUNCLCOUPL*SUNCLDX);
				
				U*=Ufac;
				U.Exponential();
				
				for(int64_t CMP=0; CMP<2*SUN*SUN; CMP++) FIELDNOW[POS2 + (mu)*USTRIDE + CMP]=U[CMP];
				
				for(int64_t CMP=0; CMP<SUNGENNUM; CMP++) FIELDNOW[POS2+ESTRIDE+SUNGENNUM*mu+CMP]=EFIELDS[IVOL*(mu*SUNGENNUM+CMP)+POS1][0]/VOL;
				
			}
			
		}}}	

		delete[] AFIELDS;
		delete[] EFIELDS;
		
		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(PPOS[DIM]);
		fftw_free(PPOS);
		
		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(MOM[DIM]);
		fftw_free(MOM);	
		
	}	
	
//-----------------------------------------------------------------------
//	MEASURING dNCS, CHERN SIMONS NUMBER CHANGES
//-----------------------------------------------------------------------

    void SetLeviCivita(Parameters* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		CurParam->mucomp= new int [24];
		CurParam->nucomp= new int [24];
		CurParam->sgcomp= new int [24];
		CurParam->rocomp= new int [24];
		CurParam->LCval= new double [24];
		
		CurParam->mucomp[0]=0;
		CurParam->nucomp[0]=1;
		CurParam->sgcomp[0]=2;
		CurParam->rocomp[0]=3;
		CurParam->mucomp[1]=0;
		CurParam->nucomp[1]=1;
		CurParam->sgcomp[1]=3;
		CurParam->rocomp[1]=2;
		CurParam->mucomp[2]=0;
		CurParam->nucomp[2]=2;
		CurParam->sgcomp[2]=1;
		CurParam->rocomp[2]=3;
		CurParam->mucomp[3]=0;
		CurParam->nucomp[3]=2;
		CurParam->sgcomp[3]=3;
		CurParam->rocomp[3]=1;
		CurParam->mucomp[4]=0;
		CurParam->nucomp[4]=3;
		CurParam->sgcomp[4]=1;
		CurParam->rocomp[4]=2;
		CurParam->mucomp[5]=0;
		CurParam->nucomp[5]=3;
		CurParam->sgcomp[5]=2;
		CurParam->rocomp[5]=1;
		CurParam->mucomp[6]=1;
		CurParam->nucomp[6]=0;
		CurParam->sgcomp[6]=2;
		CurParam->rocomp[6]=3;
		CurParam->mucomp[7]=1;
		CurParam->nucomp[7]=0;
		CurParam->sgcomp[7]=3;
		CurParam->rocomp[7]=2;
		CurParam->mucomp[8]=1;
		CurParam->nucomp[8]=2;
		CurParam->sgcomp[8]=0;
		CurParam->rocomp[8]=3;
		CurParam->mucomp[9]=1;
		CurParam->nucomp[9]=2;
		CurParam->sgcomp[9]=3;
		CurParam->rocomp[9]=0;
		CurParam->mucomp[10]=1;
		CurParam->nucomp[10]=3;
		CurParam->sgcomp[10]=0;
		CurParam->rocomp[10]=2;
		CurParam->mucomp[11]=1;
		CurParam->nucomp[11]=3;
		CurParam->sgcomp[11]=2;
		CurParam->rocomp[11]=0;
		CurParam->mucomp[12]=2;
		CurParam->nucomp[12]=0;
		CurParam->sgcomp[12]=1;
		CurParam->rocomp[12]=3;
		CurParam->mucomp[13]=2;
		CurParam->nucomp[13]=0;
		CurParam->sgcomp[13]=3;
		CurParam->rocomp[13]=1;
		CurParam->mucomp[14]=2;
		CurParam->nucomp[14]=1;
		CurParam->sgcomp[14]=0;
		CurParam->rocomp[14]=3;
		CurParam->mucomp[15]=2;
		CurParam->nucomp[15]=1;
		CurParam->sgcomp[15]=3;
		CurParam->rocomp[15]=0;
		CurParam->mucomp[16]=2;
		CurParam->nucomp[16]=3;
		CurParam->sgcomp[16]=0;
		CurParam->rocomp[16]=1;
		CurParam->mucomp[17]=2;
		CurParam->nucomp[17]=3;
		CurParam->sgcomp[17]=1;
		CurParam->rocomp[17]=0;
		CurParam->mucomp[18]=3;
		CurParam->nucomp[18]=0;
		CurParam->sgcomp[18]=1;
		CurParam->rocomp[18]=2;
		CurParam->mucomp[19]=3;
		CurParam->nucomp[19]=0;
		CurParam->sgcomp[19]=2;
		CurParam->rocomp[19]=1;
		CurParam->mucomp[20]=3;
		CurParam->nucomp[20]=1;
		CurParam->sgcomp[20]=0;
		CurParam->rocomp[20]=2;
		CurParam->mucomp[21]=3;
		CurParam->nucomp[21]=1;
		CurParam->sgcomp[21]=2;
		CurParam->rocomp[21]=0;
		CurParam->mucomp[22]=3;
		CurParam->nucomp[22]=2;
		CurParam->sgcomp[22]=0;
		CurParam->rocomp[22]=1;
		CurParam->mucomp[23]=3;
		CurParam->nucomp[23]=2;
		CurParam->sgcomp[23]=1;
		CurParam->rocomp[23]=0;	
		
		CurParam->LCval[0]=1;
		CurParam->LCval[1]=-1;
		CurParam->LCval[2]=-1;
		CurParam->LCval[3]=1;
		CurParam->LCval[4]=1;
		CurParam->LCval[5]=-1;
		CurParam->LCval[6]=-1;
		CurParam->LCval[7]=1;
		CurParam->LCval[8]=1;
		CurParam->LCval[9]=-1;
		CurParam->LCval[10]=-1;
		CurParam->LCval[11]=1;
		CurParam->LCval[12]=1;
		CurParam->LCval[13]=-1;
		CurParam->LCval[14]=-1;
		CurParam->LCval[15]=1;
		CurParam->LCval[16]=1;
		CurParam->LCval[17]=-1;
		CurParam->LCval[18]=-1;
		CurParam->LCval[19]=1;
		CurParam->LCval[20]=1;
		CurParam->LCval[21]=-1;
		CurParam->LCval[22]=-1;
		CurParam->LCval[23]=1;	
		
	}

    void FMUNU_SPATIAL( Matrix** Fmunu, double* FIELDNOW, int64_t* AXLEN, int64_t* ACNT, int64_t* STRIDES, double SUNCLDX, double SUNCLDT, double SUNCLCOUPL)
	{		
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		//-----------------------------------------------------------------------
		//	CALCULATE THE SYMMETRIZED PLAQUETTE
		//-----------------------------------------------------------------------
		
		Matrix PSumSTART;

		int64_t STEP=1;
		
		int64_t USTRIDE=2*SUN*SUN;
		int64_t ESTRIDE=2*SUN*SUN*NDIM;

		int64_t POSUSTART=0;	
		for (int k=0;k<NDIM;k++) POSUSTART+= ( ACNT[k] )*STRIDES[k];
		POSUSTART*=ESTRIDE+SUNGENNUM*NDIM;

		double SUNFAC=SUN;
		
		for(int64_t mu=0; mu<NDIM; mu++)			
			for(int64_t nu=mu+1; nu<NDIM; nu++){

			Fmunu[mu][nu].Zero();
			
			Matrix TMP,U,P1;	
			PSumSTART.Zero();
			
			
			int64_t POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?STEP:0) ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			int64_t POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?STEP:0) ) %AXLEN[k] )*STRIDES[k];					
			POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1 =&FIELDNOW[POSUSTART + (mu)*USTRIDE];
			P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
			TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
			P1*=TMP;
			TMP.Dagger(&FIELDNOW[POSUSTART + (nu)*USTRIDE]);
			P1*=TMP;

			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			Matrix TMP1, TMP2;
			
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			double ImTr;
			ImTr=P1.Trace().imag()/SUNFAC;

			TMP1-=ImTr;
			P1=TMP1;		
			
			PSumSTART+=P1;
				
			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==mu)?STEP:0) - ((k==nu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
			POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			int64_t POS4=0;
			for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] - ((k==nu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
			POS4*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
			TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
			P1*=TMP;
			P1*=&FIELDNOW[POS3 + (mu)*USTRIDE];
			P1*=&FIELDNOW[POS4 + (nu)*USTRIDE];
			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			TMP1=P1;
			TMP2.Dagger(P1);

			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;
				
			PSumSTART+=P1;
			
			
			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==nu)?STEP:0) + ((k==mu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
			POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
			P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
			P1*=&FIELDNOW[POS3 + (nu)*USTRIDE];
			TMP.Dagger(&FIELDNOW[POSUSTART + (mu)*USTRIDE]);
			P1*=TMP;
			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;				
			TMP1-=ImTr;

			P1=TMP1;
				
			
			PSumSTART+=P1;
			
			
			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
		
			POS3=0;
			for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==mu)?STEP:0) + ((k==nu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
			POS3*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1 =&FIELDNOW[POSUSTART + (nu)*USTRIDE];
			TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
			P1*=TMP;
			TMP.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
			P1*=TMP;
			P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;
				
			
			PSumSTART+=P1;

			//-----------------------------------------------------------------------
			//	Combine into the Clover approximation
			//-----------------------------------------------------------------------
			
			PSumSTART/=(SUNCLCOUPL*SUNCLDX*SUNCLDX);
			
			Fmunu[mu][nu]=PSumSTART;

			Fmunu[mu][nu]*=0.25;
				
			Fmunu[nu][mu]=Fmunu[mu][nu];
			Fmunu[nu][mu]*=-1.;
		}

	}	
	
    void FMUNU_TMPORAL( Matrix** Fmunu, double* FIELDNOW, double* FIELDNEXT, double* FIELDBEF, int64_t* AXLEN, int64_t* ACNT, int64_t* STRIDES, double SUNCLDX, double SUNCLDT, double SUNCLCOUPL, int BACKWARDSONLY)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF 
	  
		//-----------------------------------------------------------------------
		//	CALCULATE THE SYMMETRIZED PLAQUETTE
		//-----------------------------------------------------------------------
		
		Matrix PSumSTART;

		int64_t STEP=1;
		
		int64_t USTRIDE=2*SUN*SUN;
		int64_t ESTRIDE=2*SUN*SUN*NDIM;

		int64_t POSUSTART=0;	
		for (int k=0;k<NDIM;k++) POSUSTART+= ( ACNT[k] )*STRIDES[k];
		POSUSTART*=ESTRIDE+SUNGENNUM*NDIM;
		
		double SUNFAC=SUN;
		
		for(int64_t mu=0; mu<NDIM; mu++)			
			for(int64_t nu=3; nu<4; nu++){

			Fmunu[mu][nu].Zero();
			
			Matrix TMP,U,P1;	
			PSumSTART.Zero();
			
			int64_t POS2=0;
			int64_t POS3=0;
			Matrix TMP1, TMP2;
			double ImTr;
		
	if(BACKWARDSONLY==0){
			
			P1 =&FIELDNOW[POSUSTART + (mu)*USTRIDE];
			TMP.Dagger(&FIELDNEXT[POSUSTART + (mu)*USTRIDE]);
			P1*=TMP;

			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
		
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;
			
			
			PSumSTART+=P1;
	}
		
			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;
			
			P1.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
			P1*=&FIELDBEF[POS2 + (mu)*USTRIDE];

			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;

				
			PSumSTART+=P1;
			
			P1= &FIELDBEF[POSUSTART + (mu)*USTRIDE];
			TMP.Dagger(&FIELDNOW[POSUSTART + (mu)*USTRIDE]);
			P1*=TMP;
			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;
				
			PSumSTART+=P1;
			
	if(BACKWARDSONLY==0){		
			POS2=0;	
			for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?STEP:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
			POS2*=ESTRIDE+SUNGENNUM*NDIM;

			P1.Dagger(&FIELDNEXT[POS2 + (mu)*USTRIDE]);

			P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
			
			//-----------------------------------------------------------------------
			//	F is a member of the ALGEBRA, therefore it needs to be
			//	traceless!
			//-----------------------------------------------------------------------
			
			TMP1=P1;
			TMP2.Dagger(P1);
			
			TMP1-=TMP2;
			TMP1*=Complex(0.,-0.5);
			
			ImTr=P1.Trace().imag()/SUNFAC;
			TMP1-=ImTr;
		
			P1=TMP1;
				
			
			PSumSTART+=P1;	
	}
				
			//-----------------------------------------------------------------------
			//	Combine into the Clover approximation
			//-----------------------------------------------------------------------
			PSumSTART/=(SUNCLCOUPL*SUNCLDT*SUNCLDX);
			
			Fmunu[mu][nu]=PSumSTART;
			
				
			if(BACKWARDSONLY==0){
				Fmunu[mu][nu]*=0.25;
			}else{
				Fmunu[mu][nu]*=0.5;	
			}
			
			Fmunu[nu][mu]=Fmunu[mu][nu];
			Fmunu[nu][mu]*=-1.;
		}

	}


    void SUN_CLASS_MEAS_dNCS_ANTISYMONLY(double* FIELDNOW, double *FIELDNEXT, double* FIELDBEF, double* dNCS, Parameters* param)
	{
	  
	//CHECKED ON MAY 27th 2015 ==> OK	A. ROTHKOPF 
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		dNCS[0]=0.;
		int STEP=1;
		
		double SUNFAC=SUN;
		
		//-----------------------------------------------------------------------
		//	WE DIRECTLY CALCULATE F_\mu\nu\tilde{F^\mu\nu} USING 
		//	THE FIELD TENSOR FROM THE ANTISYMMETRIZED PLAQUETTE
		//	SIN(X)~X
		//-----------------------------------------------------------------------

		Matrix **Fmunu;
		Fmunu= new Matrix*[4];
		for(int j=0;j<4;j++) Fmunu[j]=new Matrix [4];
		
		
		Matrix PSumSTART;
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]+=STEP){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]+=STEP){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]+=STEP){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSUSTART=pos1 + ACNT[2]*STRIDES[2];
				POSUSTART*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				//-----------------------------------------------------------------------
				//	CALCULATE FMUNU
				//-----------------------------------------------------------------------
	
				FMUNU_TMPORAL( Fmunu, FIELDNOW, FIELDNEXT, FIELDBEF,AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL, 0);

				FMUNU_SPATIAL( Fmunu, FIELDNOW, AXLEN,  ACNT,  STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);
   
				Complex FFtildeTr;
				FFtildeTr=0.;
					
				for(int64_t k=0;k<24;k++){
					
					Matrix TMP;
					
					TMP=Fmunu[mucomp[k]][nucomp[k]];
					TMP*=Fmunu[sgcomp[k]][rocomp[k]];
					
					FFtildeTr+=-0.5*LCval[k]*TMP.Trace();
					
					//In Mathematica the definition of the LeviCivita is with lowered
					//indices, so one needs an additional minus sign to obtain
					//the levicivita with raised indices for e^abcd F_ab F_cd
				}
				
				dNCS[0]+=FFtildeTr.real();
				
				
			}
		    }
		}
		
		double dNCSfac=(SUNCLCOUPL*SUNCLCOUPL/(8.*PI*PI*2.));	// IN TOTAL THIS IS g^2/ ( 64 Pi^2) F \tilde(F)
		
		dNCS[0]*=dNCSfac;
		
		for(int j=0;j<4;j++) delete[] Fmunu[j];
		delete[] Fmunu;
	
	}


    void SUN_CLASS_MEAS_dNCS_ROTHKOPF(double* FIELDNOW, double *FIELDNEXT, double* FIELDBEF, double* dNCS, Parameters* param)
	{
	  
	//CHECKED ON MAY 27th 2015 ==> OK	A. ROTHKOPF
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		double *LCval=CurParam->LCval;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;

		char dsjobl='N';
		char dsjobr='V';
		
		int size=SUN;
		
		int *piv;
		piv = new int [SUN];
		
		Complex *dsw;
		dsw = new Complex [SUN];
		
		Complex *dsvr;
		dsvr = new Complex [SUN*SUN];
		
		Complex *dsvl;
		dsvl = new Complex [SUN*SUN];
		
		double *dsrwork;
		dsrwork= new double [2*SUN];
		
		int dsinfo;
		
		int dslda=SUN;
		int dslvr=SUN;
		int dslvl=SUN;
		int dslw=2*SUN;
					
		Complex *dswork;
		dswork= new Complex [2*SUN*SUN];		
		
		dNCS[0]=0.;
		
		//-----------------------------------------------------------------------
		//	WE DIRECTLY CALCULATE F_\mu\nu\tilde{F^\mu\nu} USING 
		//	THE FIELD TENSOR FROM THE MATRIX LOGARITHM OF THE PLAQUETTE
		//-----------------------------------------------------------------------

		Matrix PSumSTART;
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSUSTART=pos1 + ACNT[2]*STRIDES[2];
				POSUSTART*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				//-----------------------------------------------------------------------
				//	CALCULATE THE SYMMETRIZED PLAQUETTE
				//-----------------------------------------------------------------------
				Matrix Fmunu[4][4];

				for(int64_t mu=0; mu<NDIM; mu++)			
					for(int64_t nu=mu+1; nu<NDIM; nu++){

					Fmunu[mu][nu].Zero();
					
					Matrix TMP,U,P1,EigenVec;	
					PSumSTART.Zero();
					
					
					int64_t POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					int64_t POS3=0;
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					P1 =&FIELDNOW[POSUSTART + (mu)*USTRIDE];
					P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
					TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
					P1*=TMP;
					TMP.Dagger(&FIELDNOW[POSUSTART + (nu)*USTRIDE]);
					P1*=TMP;

					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
			
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDX*SUNCLDX));
					P1=U;				
					
					PSumSTART+=P1;
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					POS3=0;
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==mu)?1:0) - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					int64_t POS4=0;
					for (int k=0;k<NDIM;k++) POS4+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
					POS4*=ESTRIDE+SUNGENNUM*NDIM;
					
					P1.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
					TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
					P1*=TMP;
					P1*=&FIELDNOW[POS3 + (mu)*USTRIDE];
					P1*=&FIELDNOW[POS4 + (nu)*USTRIDE];
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDX*SUNCLDX));
					P1=U;				
						
					PSumSTART+=P1;
					
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					POS3=0;
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==nu)?1:0) + ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					P1.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
					P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
					P1*=&FIELDNOW[POS3 + (nu)*USTRIDE];
					TMP.Dagger(&FIELDNOW[POSUSTART + (mu)*USTRIDE]);
					P1*=TMP;
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDX*SUNCLDX));
					P1=U;									
					
					PSumSTART+=P1;
					
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					POS3=0;
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] - ((k==mu)?1:0) + ((k==nu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];					
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					P1 =&FIELDNOW[POSUSTART + (nu)*USTRIDE];
					TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
					P1*=TMP;
					TMP.Dagger(&FIELDNOW[POS2 + (nu)*USTRIDE]);
					P1*=TMP;
					P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDX*SUNCLDX));
					P1=U;				
					
					PSumSTART+=P1;
					
					//-----------------------------------------------------------------------
					//	Combine into the Clover approximation
					//-----------------------------------------------------------------------
					
					Fmunu[mu][nu]=PSumSTART;		
					Fmunu[mu][nu]*=0.25;

					Fmunu[nu][mu]=Fmunu[mu][nu];
					Fmunu[nu][mu]*=-1.;
				}
				
				
				for(int64_t mu=0; mu<NDIM; mu++)			
					for(int64_t nu=3; nu<4; nu++){

					Fmunu[mu][nu].Zero();
					
					Matrix TMP,U,P1,EigenVec;	
					PSumSTART.Zero();
										
					P1 =&FIELDNOW[POSUSTART + (mu)*USTRIDE];
					TMP.Dagger(&FIELDNEXT[POSUSTART + (mu)*USTRIDE]);
					P1*=TMP;

					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDT*SUNCLDX));
					P1=U;				
					
					PSumSTART+=P1;
					
					int64_t POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;

					P1.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
					P1*=&FIELDBEF[POS2 + (mu)*USTRIDE];
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDT*SUNCLDX));
					P1=U;				
						
					PSumSTART+=P1;
					
					
					P1= &FIELDBEF[POSUSTART + (mu)*USTRIDE];
					TMP.Dagger(&FIELDNOW[POSUSTART + (mu)*USTRIDE]);
					P1*=TMP;
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDT*SUNCLDX));
					P1=U;				
						
					PSumSTART+=P1;
					
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] - ((k==mu)?1:0) + AXLEN[k] ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;

					P1.Dagger(&FIELDNEXT[POS2 + (mu)*USTRIDE]);
					P1*=&FIELDNOW[POS2 + (mu)*USTRIDE];
					
					//-----------------------------------------------------------------------
					//	EXTRACT Fmunu FROM THE PLAQUETTE via MATRIX LOG
					//-----------------------------------------------------------------------
					SetTranspose(&U[0] ,&P1[0]);				
					zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
					
					SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
			
					for(int i=0;i<SUN;i++){	

						double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
						double logabs=log(aval);
						double angle=atan2(dsw[i].imag(),dsw[i].real());
						
						dsw[i].real(logabs);
						dsw[i].imag(angle);
					}
					
					U=EigenVec;
					
					for(int i=0;i<SUN;i++){
					for(int j=0;j<SUN;j++){
					
						U(i,j)*=dsw[j];
					
					}}
					
					EigenVec.Dagger();
					U*=EigenVec;
					U*=Complex(0,-1./(SUNCLCOUPL*SUNCLDT*SUNCLDX));
					P1=U;				
					
					PSumSTART+=P1;				
					
					//-----------------------------------------------------------------------
					//	Combine into the Clover approximation
					//-----------------------------------------------------------------------
					
					Fmunu[mu][nu]=PSumSTART;			
					Fmunu[mu][nu]*=0.25;

					Fmunu[nu][mu]=Fmunu[mu][nu];
					Fmunu[nu][mu]*=-1.;
				}
					
					
				Complex FFtildeTr;
				FFtildeTr=0.;
					
				for(int64_t k=0;k<24;k++){
					
					Matrix TMP;
					
					TMP=Fmunu[mucomp[k]][nucomp[k]];
					TMP*=Fmunu[sgcomp[k]][rocomp[k]];
					
					FFtildeTr+=-0.5*LCval[k]*TMP.Trace();
					
					//In Mathematica the definition of the LeviCivita is with lowered
					//indices, so one needs an additional minus sign to obtain
					//the levicivita with raised indices for e^abcd F_ab F_cd
				}
				
				dNCS[0]+=FFtildeTr.real();
				
			}
		    }
		}

		double dNCSfac=(SUNCLCOUPL*SUNCLCOUPL/(8.*PI*PI*2.));	
			
		dNCS[0]*=dNCSfac;
		
		delete[] piv;
		delete[] dsw;
		delete[] dsvr;
		delete[] dsvl;
		delete[] dsrwork;
		delete[] dswork;
	}

//-----------------------------------------------------------------------
//	MEASURING OBSERVABLES
//-----------------------------------------------------------------------

    void SUN_CLASS_MEASURE_ENERGY_FOURIER(double* FIELDNOW, int FILENUM, Parameters* param, int INDIV)
	{
		FILE *POUT;

		//-----------------------------------------------------------------------
		//	I am grateful to Sabolcs Borsanyi for providing me with the
		//	3D counting scheme for the negative momenta 
		//-----------------------------------------------------------------------
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		int64_t MCNT[NDIM];	//COUNTERS FOR THE NEGATIVE MOMENTA
		int64_t PCNT;		//SINGLE COUNTER FOR THE POSITIVE MOMENTA
			
		int AXLENFFT[NDIM];	//AXES LENGTH FOR THE FFT PLANS
		double KVAL[NDIM];	//POSITIVE MOMENTA VALUES
		double MKVL[NDIM];	//NEGATIVE MOMENTA VALUES

		fftw_plan FFTA,FFTE,FFTE2,FFTE3;		//FFTW PLAN FOR THE FORWARD TRANSFORMATION
		fftw_plan IFFTA,IFFTE,IFFTE2,IFFTE3;		//FFTW PLAN FOR THE BACKWARD TRANSFORMATION
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		AXLENFFT[0]=CurParam->NX;
		AXLENFFT[1]=CurParam->NY;
		AXLENFFT[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		int64_t BINNUM=CurParam->SpecBins;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		int64_t IVOL=AXLEN[0]*AXLEN[1]*AXLEN[2];		
		
		//*********************************************************************************
		//  SETUP THE FOURIER SPACE MOMENTA, WHICH NEED TO BE SET SUCH THAT THE NYQUIST
		//  FREQUENCY IS CORRECTLY CAPTURED AT AXLEN/2
		//*********************************************************************************
		
		double **PPOS;
		PPOS=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) PPOS[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<=AXLEN[DIM]/2;PCNT++) PPOS[DIM][PCNT]=PCNT*((2.*PI)/(((double)AXLEN[DIM])*((double)SUNCLDX)));
			for(int PCNT=AXLEN[DIM]/2+1;PCNT<AXLEN[DIM];PCNT++) PPOS[DIM][PCNT]=-PPOS[DIM][AXLEN[DIM]-PCNT];
		}

		double **MOM;
		MOM=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) MOM[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<AXLEN[DIM];PCNT++) MOM[DIM][PCNT]=sqrt(2.5-(8./3.)*cos(2.*PCNT*PI/((double)AXLEN[DIM]))+(1./6.)*cos(4.*PCNT*PI/((double)AXLEN[DIM])))/SUNCLDX;
		}
		
		//*********************************************************************************
		// THE ARRAY FOR THE A FIELDS USED IN THE INITIALIZATION
		//*********************************************************************************
		
		fftw_complex *ENERGY;
		ENERGY=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*IVOL);

		fftw_complex *MAGENERGY;
		MAGENERGY=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*IVOL);

		fftw_complex *ELENERGY;
		ELENERGY=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*IVOL);

		int64_t* INBIN;
		double* BINNAIVEK, *BINREALK, *BINEREAL, *BINEIMAG, *BINEABS, *BINEREAL2, *BINEIMAG2, *BINEABS2, *BINEREAL3, *BINEIMAG3, *BINEABS3;
		
		INBIN=new int64_t [BINNUM];
		
		BINNAIVEK= new double [BINNUM];
		BINREALK= new double [BINNUM];
		
		BINEREAL= new double [BINNUM];
		BINEIMAG= new double [BINNUM];
		BINEABS= new double [BINNUM];

		BINEREAL2= new double [BINNUM];
		BINEIMAG2= new double [BINNUM];
		BINEABS2= new double [BINNUM];

		BINEREAL3= new double [BINNUM];
		BINEIMAG3= new double [BINNUM];
		BINEABS3= new double [BINNUM];

		
		
		for(int k=0;k<BINNUM;k++){
			
			INBIN[k]=0;
			BINNAIVEK[k]=0;
			BINREALK[k]=0;
			
			BINEREAL[k]=0;
			BINEIMAG[k]=0;
			BINEABS[k]=0;
			
			BINEREAL2[k]=0;
			BINEIMAG2[k]=0;
			BINEABS2[k]=0;
			
			BINEREAL3[k]=0;
			BINEIMAG3[k]=0;
			BINEABS3[k]=0;
			
		}
		
		//fprintf(stderr, "\n\n DEBUG1 \n\n");
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){

				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double MAGFAC=2./(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX);
				double EFAC=2.*SUN*SUNCLDX/(SUNCLCOUPL*SUNCLCOUPL);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				int64_t POS1=POS; // POS1 FOR THE FFTW ARRAY
				
				POS*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				//-----------------------------------------------------------------------
				//	MAGNETIC ENERGY PART                        
				//-----------------------------------------------------------------------

				Matrix P1;
				Matrix Unit;
				
				ENERGY[POS1][0]=0.;
				ENERGY[POS1][1]=0.;
				
				ELENERGY[POS1][0]=0.;
				ELENERGY[POS1][1]=0.;
				
				MAGENERGY[POS1][0]=0.;
				MAGENERGY[POS1][1]=0.;
				
					for(int64_t mu=0; mu<NDIM-1; mu++){
					for(int64_t nu=mu+1; nu<NDIM; nu++){

						Matrix TMP;
		
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS  + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						//CMNRG[0]+=MAGFAC*( (SUN)-P1.RealTrace())/(VOL);
						
						ENERGY[POS1][0]+=MAGFAC*( (SUN)-P1.RealTrace());
						ENERGY[POS1][1]+=0.;
						
						MAGENERGY[POS1][0]+=MAGFAC*( (SUN)-P1.RealTrace());
						MAGENERGY[POS1][1]+=0.;
						
							
				}}
				
				//-----------------------------------------------------------------------
				//	ELECTRIC ENERGY PART                        
				//-----------------------------------------------------------------------
				
				for(int64_t mu=0; mu<NDIM; mu++){
				  for(int64_t gen=0; gen<SUNGENNUM; gen++){
					  
					//CENRG[0]+=EFAC*( FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen]*FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen] )/(VOL);
					
					ENERGY[POS1][0]+=EFAC*( FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen]*FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen] );
					ENERGY[POS1][1]+=0.;
					
					ELENERGY[POS1][0]+=EFAC*( FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen]*FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen] );
					ELENERGY[POS1][1]+=0.;
					
				  }
				}
				
			}
		    }
		}	

		//fprintf(stderr, "\n\n DEBUG2 \n\n");
		
		//*********************************************************************************
		//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
		//*********************************************************************************
		
		FFTE=fftw_plan_dft(3, AXLENFFT, ENERGY, ENERGY,-1, FFTW_ESTIMATE);
		IFFTE=fftw_plan_dft(3, AXLENFFT, ENERGY, ENERGY,+1, FFTW_ESTIMATE);
		
		FFTE2=fftw_plan_dft(3, AXLENFFT, ELENERGY, ELENERGY,-1, FFTW_ESTIMATE);
		IFFTE2=fftw_plan_dft(3, AXLENFFT, ELENERGY, ELENERGY,+1, FFTW_ESTIMATE);
		
		FFTE3=fftw_plan_dft(3, AXLENFFT, MAGENERGY, MAGENERGY,-1, FFTW_ESTIMATE);
		IFFTE3=fftw_plan_dft(3, AXLENFFT, MAGENERGY, MAGENERGY,+1, FFTW_ESTIMATE);
		
		fftw_execute(FFTE);
		fftw_execute(FFTE2);
		fftw_execute(FFTE3);
		
		//fprintf(stderr, "\n\n DEBUG3 \n\n");

		//-----------------------------------------------------------------------
		//	MCNT, the counter for the negative frequencies is always 
		//	incremented at the end of each loop step since,
		//	the zero position does not have a counterpart (the NYQUIST
		//	FREQUENCY takes this role) and +K <-> -K sorting starts from
		//	step 1 NOT 0.
		//-----------------------------------------------------------------------
		
		for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
		    for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
		      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
				
			register double KNAIVSQ=0.;
					
	// 		int INSIDEKBALL=1;
			
			for(int DIM=0; DIM<NDIM; DIM++){
			  
			  KNAIVSQ+=pow(PPOS[DIM][ACNT[DIM]],2);
			  
	// 		  if(ACNT[DIM]>NX/2) INSIDEKBALL=0;
			  
			}
			
			double KNAIV=sqrt(KNAIVSQ);      
			
			//double MAXMOM=sqrt(3.)*PI/SUNCLDX;
			double MAXMOM=PI/SUNCLDX;
			
				if( KNAIV < MAXMOM  /*&& INSIDEKBALL == 1*/){

					// USE A LOGARITHMIC SPACING FOR THE MOMENTA TO BETTER ACCOMMODATE THE LARGE
					// FLUCTUATIONS AT THE LOWER MOMENTA
						
					const int NBIN=(int)floor( BINNUM* ((log(KNAIVSQ/(2.*MAXMOM*2.*MAXMOM)) +log(AXLEN[0]*AXLEN[0]))/log(AXLEN[0]*AXLEN[0])) );

					if(NBIN<BINNUM&&NBIN>=0){			//ZERO MODE IS JUST THE AVERAGE SO WE DO NOT
											//INCLUDE IT. ALSO NUMBERS BECOME VERY LARGE
											//PROBLEMS WITH OVERFLOW MIGHT OCCUR
											
						double EREAL=(ENERGY[PCNT][0]);
						double EIMAG=(ENERGY[PCNT][1]);
						double EABS=sqrt(ENERGY[PCNT][0]*ENERGY[PCNT][0]+ENERGY[PCNT][1]*ENERGY[PCNT][1]);
						
						double EREAL2=(ELENERGY[PCNT][0]);
						double EIMAG2=(ELENERGY[PCNT][1]);
						double EABS2=sqrt(ELENERGY[PCNT][0]*ELENERGY[PCNT][0]+ELENERGY[PCNT][1]*ELENERGY[PCNT][1]);

						double EREAL3=(MAGENERGY[PCNT][0]);
						double EIMAG3=(MAGENERGY[PCNT][1]);
						double EABS3=sqrt(MAGENERGY[PCNT][0]*MAGENERGY[PCNT][0]+MAGENERGY[PCNT][1]*MAGENERGY[PCNT][1]);
						
						double KREALSQ=0.;					
						for(int DIM=0; DIM<NDIM; DIM++) KREALSQ+=pow(MOM[DIM][ACNT[DIM]],2);

						INBIN[NBIN]++;
						BINNAIVEK[NBIN]+=sqrt(KNAIVSQ);		//FIND THE AVERAGE MOMENTUM THAT CORRESPONDS TO THIS BIN
						BINREALK[NBIN]+=sqrt(KREALSQ);		//

						BINEREAL[NBIN]+=EREAL;
						BINEIMAG[NBIN]+=EIMAG;
						BINEABS[NBIN]+=EABS;
						
						BINEREAL2[NBIN]+=EREAL2;
						BINEIMAG2[NBIN]+=EIMAG2;
						BINEABS2[NBIN]+=EABS2;
						
						BINEREAL3[NBIN]+=EREAL3;
						BINEIMAG3[NBIN]+=EIMAG3;
						BINEABS3[NBIN]+=EABS3;
						
						
					}
				}
				
			}
		    }
		}
		
		//*********************************************************************************
		//  DEBUGGING THE CORRECT NOISE STRUCTURE IN AVERAGING OVER ALL COLOR AND
		//  SPATIAL DIRECTIONS
		//*********************************************************************************
		//fprintf(stderr, "\n\n DEBUG4 \n\n");
		
		char FNAME[256];
			
		if(INDIV==1){
		  
		    sprintf(FNAME,"EnergySpectrum.%d",FILENUM);
		    POUT=fopen(FNAME,"w");
		    for (int k=0; k<BINNUM; k++)  if(INBIN[k]!=0) fprintf(POUT, "%.10e %.10e %d %.10e %.10e %.10e %.10e %.10e %.10e %.10e %.10e %.10e\n", BINNAIVEK[k]/INBIN[k], BINREALK[k]/INBIN[k]  ,INBIN[k], BINEREAL[k]/INBIN[k], BINEIMAG[k]/INBIN[k], BINEABS[k]/INBIN[k], BINEREAL2[k]/INBIN[k], BINEIMAG2[k]/INBIN[k], BINEABS2[k]/INBIN[k], BINEREAL3[k]/INBIN[k], BINEIMAG3[k]/INBIN[k], BINEABS3[k]/INBIN[k]  );	
		    fclose(POUT);
		
		}
		
		//fprintf(stderr, "\n\n DEBUG5 \n\n");
		
		for(int k=0;k<BINNUM;k++) {
		  if(INBIN[k]==0) INBIN[k]==1;
		}
		
		sprintf(FNAME,"TotalEnergyMeasuredModes.dat");
		POUT=fopen(FNAME,"a");	
		fprintf(POUT, "%d ", FILENUM);
		for (int k=0; k<BINNUM; k++) fprintf(POUT, "%.10e %.10e %.10e ", BINEREAL[k]/INBIN[k], BINEIMAG[k]/INBIN[k], BINEABS[k]/INBIN[k] );
		fprintf(POUT, "\n");
		fclose(POUT);
		
		sprintf(FNAME,"ElectricEnergyMeasuredModes.dat");
		POUT=fopen(FNAME,"a");	
		fprintf(POUT, "%d ", FILENUM);
		for (int k=0; k<BINNUM; k++) fprintf(POUT, "%.10e %.10e %.10e ", BINEREAL2[k]/INBIN[k], BINEIMAG2[k]/INBIN[k], BINEABS2[k]/INBIN[k] );
		fprintf(POUT, "\n");
		fclose(POUT);
		
		sprintf(FNAME,"MagneticEnergyMeasuredModes.dat");
		POUT=fopen(FNAME,"a");	
		fprintf(POUT, "%d ", FILENUM);
		for (int k=0; k<BINNUM; k++) fprintf(POUT, "%.10e %.10e %.10e ", BINEREAL3[k]/INBIN[k], BINEIMAG3[k]/INBIN[k], BINEABS3[k]/INBIN[k] );
		fprintf(POUT, "\n");
		fclose(POUT);

		//fprintf(stderr, "\n\n DEBUG6 \n\n");
		
		delete[] INBIN;
		delete[] BINNAIVEK;
		delete[] BINREALK;

		delete[] BINEREAL;
		delete[] BINEIMAG;
		delete[] BINEABS;
		
		delete[] BINEREAL2;
		delete[] BINEIMAG2;
		delete[] BINEABS2;
		
		delete[] BINEREAL3;
		delete[] BINEIMAG3;
		delete[] BINEABS3;
		
		//fprintf(stderr, "\n\n DEBUG7 \n\n");
		
		fftw_free(ENERGY);
		fftw_free(ELENERGY);
		fftw_free(MAGENERGY);

		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(PPOS[DIM]);
		fftw_free(PPOS);
		
		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(MOM[DIM]);
		fftw_free(MOM);	
		
		
	}


    void SUN_CLASS_MEASURE_FIELD_MOMENTA(double* FIELDNOW, int FILENUM, Parameters* param)
	{
		FILE *POUT;

		//-----------------------------------------------------------------------
		//	I am grateful to Sabolcs Borsanyi for providing me with the
		//	3D counting scheme for the negative momenta 
		//-----------------------------------------------------------------------
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		int64_t MCNT[NDIM];	//COUNTERS FOR THE NEGATIVE MOMENTA
		int64_t PCNT;		//SINGLE COUNTER FOR THE POSITIVE MOMENTA
			
		int AXLENFFT[NDIM];	//AXES LENGTH FOR THE FFT PLANS
		double KVAL[NDIM];	//POSITIVE MOMENTA VALUES
		double MKVL[NDIM];	//NEGATIVE MOMENTA VALUES

		fftw_plan FFTA,FFTE;		//FFTW PLAN FOR THE FORWARD TRANSFORMATION
		fftw_plan IFFTA,IFFTE;		//FFTW PLAN FOR THE BACKWARD TRANSFORMATION
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		AXLENFFT[0]=CurParam->NX;
		AXLENFFT[1]=CurParam->NY;
		AXLENFFT[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		int64_t BINNUM=CurParam->SpecBins;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		int64_t IVOL=AXLEN[0]*AXLEN[1]*AXLEN[2];		
		
		//*********************************************************************************
		//  SETUP THE FOURIER SPACE MOMENTA, WHICH NEED TO BE SET SUCH THAT THE NYQUIST
		//  FREQUENCY IS CORRECTLY CAPTURED AT AXLEN/2
		//*********************************************************************************
		
		double **PPOS;
		PPOS=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) PPOS[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<=AXLEN[DIM]/2;PCNT++) PPOS[DIM][PCNT]=PCNT*((2.*PI)/(((double)AXLEN[DIM])*((double)SUNCLDX)));
			for(int PCNT=AXLEN[DIM]/2+1;PCNT<AXLEN[DIM];PCNT++) PPOS[DIM][PCNT]=-PPOS[DIM][AXLEN[DIM]-PCNT];
		}

		double **MOM;
		MOM=(double**)fftw_malloc(sizeof(double*)*NDIM);
		for(int DIM=0;DIM<NDIM;DIM++) MOM[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

		for(int DIM=0;DIM<NDIM;DIM++){
			for(int PCNT=0;PCNT<AXLEN[DIM];PCNT++) MOM[DIM][PCNT]=sqrt(2.5-(8./3.)*cos(2.*PCNT*PI/((double)AXLEN[DIM]))+(1./6.)*cos(4.*PCNT*PI/((double)AXLEN[DIM])))/SUNCLDX;
		}
		
		//*********************************************************************************
		// THE ARRAY FOR THE A FIELDS USED IN THE INITIALIZATION
		//*********************************************************************************
		
		fftw_complex *AFIELDS,*EFIELDS, *CURFIELDA,*CURFIELDE;
		AFIELDS=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*NDIM*SUNGENNUM*IVOL);
		EFIELDS=(fftw_complex*) fftw_malloc(sizeof(fftw_complex)*NDIM*SUNGENNUM*IVOL);
		
		double AASTARxy=0;
		double AASTARyz=0;
		double AASTARxz=0;
		double AASTARx=0;
		double AASTARy=0;
		double AASTARz=0;

		double EESTARxy=0;
		double EESTARyz=0;
		double EESTARxz=0;
		double EESTARx=0;
		double EESTARy=0;
		double EESTARz=0;

		int64_t* INBIN;
		double* BINNAIVEK, *BINREALK, *BINAA, *BINEE, *BINAAVAR, *BINEEVAR;
		
		INBIN=new int64_t [BINNUM];
		BINNAIVEK= new double [BINNUM];
		BINREALK= new double [BINNUM];
		BINAA= new double [BINNUM];
		BINEE= new double [BINNUM];

		BINAAVAR= new double [BINNUM];
		BINEEVAR= new double [BINNUM];

		for(int k=0;k<BINNUM;k++){
			
			INBIN[k]=0;
			BINNAIVEK[k]=0;
			BINREALK[k]=0;
			BINAA[k]=0;
			BINEE[k]=0;
			
			BINAAVAR[k]=0;
			BINEEVAR[k]=0;
			
		}
		//*********************************************************************************
		// EXTRACT FROM THE LINK VARIABLES THE A FIELD
		//
		// BE AWARE THAT ONCE THE FIELD VALUES ARE LARGER THAN UNITY, THE SIMPLE
		// FORMULA U=1+iagA IS NOT APPLICABLE ANYMORE TO PROJECT OUT A. INDEED ONE HAS
		// TO CALCULATE THE COMPLEX LOGARITHM OF THE LINK U!
		//
		//*********************************************************************************
		
		char dsjobl='N';
		char dsjobr='V';
		
		int size=SUN;
		
		int *piv;
		piv = new int [SUN];
		
		Complex *dsw;
		dsw = new Complex [SUN];
		
		Complex *dsvr;
		dsvr = new Complex [SUN*SUN];
		
		Complex *dsvl;
		dsvl = new Complex [SUN*SUN];
		
		double *dsrwork;
		dsrwork= new double [2*SUN];
		
		int dsinfo;
		
		int dslda=SUN;
		int dslvr=SUN;
		int dslvl=SUN;
		int dslw=2*SUN;
					
		Complex *dswork;
		dswork= new Complex [2*SUN*SUN];	
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		int64_t pos0=ACNT[0]*STRIDES[0];
		
		for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
			
			int64_t USTRIDE=2*SUN*SUN;
			int64_t ESTRIDE=2*SUN*SUN*NDIM;
			
			int64_t POS1=pos1 + ACNT[2]*STRIDES[2];
			
			int64_t POS2=POS1;
			POS2*=(ESTRIDE+SUNGENNUM*NDIM);
			
	// 		double *TMP;
	// 		TMP=new double [SUNGENNUM];
			
			double TMP[SUNGENNUM];
			
			for(int64_t mu=0;mu<NDIM;mu++){
				
				Matrix U,EigenVec;
				
				U=&FIELDNOW[POS2 + (mu)*USTRIDE];
	// 			U.print();
				
				SetTranspose(&U[0] ,&FIELDNOW[POS2 + (mu)*USTRIDE]);		
				
				zgeev_( &dsjobl , &dsjobr , &size, (_lapack_complex_double*) &(U[0]), &dslda, (_lapack_complex_double*)dsw, (_lapack_complex_double*)dsvl, &dslvl, (_lapack_complex_double*)dsvr, &dslvr, (_lapack_complex_double*)dswork, &dslw, dsrwork, &dsinfo );		
				
				SetTranspose(&EigenVec[0], (double*) &dsvr[0]);
		
				for(int i=0;i<SUN;i++){	
					
					double aval=sqrt(dsw[i].imag()*dsw[i].imag()+dsw[i].real()*dsw[i].real());
					double logabs=log(aval);
					double angle=atan2(dsw[i].imag(),dsw[i].real());
					
					dsw[i].real(logabs);
					dsw[i].imag(angle);

				}
				
				
				U=EigenVec;
				
				for(int i=0;i<SUN;i++){
				for(int j=0;j<SUN;j++){
				
					U(i,j)*=dsw[j];
				
				}
				}
				
				EigenVec.Dagger();
				U*=EigenVec;

				for(int k=0; k<SUNGENNUM; k++) TMP[k]=2.*U.ImTrGen(k)/(SUNCLCOUPL*SUNCLDX); 
				
				for(int64_t k=0; k<SUNGENNUM; k++){
					AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]=TMP[k];
					AFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][1]=0.;
					
					EFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][0]=FIELDNOW[POS2 + ESTRIDE + (mu)*SUNGENNUM+k];
					EFIELDS[IVOL*(mu*SUNGENNUM+k)+POS1][1]=0.;

				}
			
			}
			
		}}}
		
		delete[] piv;
		delete[] dsw;
		delete[] dsvr;
		delete[] dsvl;
		delete[] dsrwork;
		delete[] dswork;
		
		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				CURFIELDA=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				CURFIELDE=&EFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				FFTA=fftw_plan_dft(3, AXLENFFT, CURFIELDA, CURFIELDA,-1, FFTW_ESTIMATE);
				IFFTA=fftw_plan_dft(3, AXLENFFT, CURFIELDA, CURFIELDA,+1, FFTW_ESTIMATE);
				
				FFTE=fftw_plan_dft(3, AXLENFFT, CURFIELDE, CURFIELDE,-1, FFTW_ESTIMATE);
				IFFTE=fftw_plan_dft(3, AXLENFFT, CURFIELDE, CURFIELDE,+1, FFTW_ESTIMATE);
				
				fftw_execute(FFTA);
				fftw_execute(FFTE);

				//-----------------------------------------------------------------------
				//	MCNT, the counter for the negative frequencies is always 
				//	incremented at the end of each loop step since,
				//	the zero position does not have a counterpart (the NYQUIST
				//	FREQUENCY takes this role) and +K <-> -K sorting starts from
				//	step 1 NOT 0.
				//-----------------------------------------------------------------------
				
				for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				  for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
						
					register double KNAIVSQ=0.;
					for(int DIM=0; DIM<NDIM; DIM++) KNAIVSQ+=pow(PPOS[DIM][ACNT[DIM]],2);
					double KNAIV=sqrt(KNAIVSQ);      
					
					double MAXMOM=PI/SUNCLDX;
					
						if( KNAIV < MAXMOM ){

							// USE A LOGARITHMIC SPACING FOR THE MOMENTA TO BETTER ACCOMMODATE THE LARGE
							// FLUCTUATIONS AT THE LOWER MOMENTA
								
							const int NBIN=(int)floor( BINNUM* ((log(KNAIVSQ/(2.*MAXMOM*2.*MAXMOM)) +log(AXLEN[0]*AXLEN[0]))/log(AXLEN[0]*AXLEN[0])) );

							if(NBIN<BINNUM&&NBIN>=0){			//ZERO MODE IS JUST THE AVERAGE SO WE DO NOT
													//INCLUDE IT. ALSO NUMBERS BECOME VERY LARGE
													//PROBLEMS WITH OVERFLOW MIGHT OCCUR
													
								double AASTAR=(1./(SUNGENNUM*NDIM))*(CURFIELDA[PCNT][0]*CURFIELDA[PCNT][0]+CURFIELDA[PCNT][1]*CURFIELDA[PCNT][1]);
								double EESTAR=(1./(SUNGENNUM*NDIM))*(CURFIELDE[PCNT][0]*CURFIELDE[PCNT][0]+CURFIELDE[PCNT][1]*CURFIELDE[PCNT][1]);
								
								double KREALSQ=0.;					
								for(int DIM=0; DIM<NDIM; DIM++) KREALSQ+=pow(MOM[DIM][ACNT[DIM]],2);
								
								AASTARx+=(AASTAR*pow(MOM[0][ACNT[0]],2));
								AASTARy+=(AASTAR*pow(MOM[1][ACNT[1]],2));
								AASTARz+=(AASTAR*pow(MOM[2][ACNT[2]],2));

								AASTARxy+=(AASTAR*(pow(MOM[0][ACNT[0]],2)+pow(MOM[1][ACNT[1]],2)));
								AASTARyz+=(AASTAR*(pow(MOM[1][ACNT[1]],2)+pow(MOM[2][ACNT[2]],2)));
								AASTARxz+=(AASTAR*(pow(MOM[2][ACNT[2]],2)+pow(MOM[0][ACNT[0]],2)));

								EESTARx+=(EESTAR*pow(MOM[0][ACNT[0]],2));
								EESTARy+=(EESTAR*pow(MOM[1][ACNT[1]],2));
								EESTARz+=(EESTAR*pow(MOM[2][ACNT[2]],2));

								EESTARxy+=(EESTAR*(pow(MOM[0][ACNT[0]],2)+pow(MOM[1][ACNT[1]],2)));
								EESTARyz+=(EESTAR*(pow(MOM[1][ACNT[1]],2)+pow(MOM[2][ACNT[2]],2)));
								EESTARxz+=(EESTAR*(pow(MOM[2][ACNT[2]],2)+pow(MOM[0][ACNT[0]],2)));

								INBIN[NBIN]++;
								BINNAIVEK[NBIN]+=sqrt(KNAIVSQ);		//FIND THE AVERAGE MOMENTUM THAT CORRESPONDS TO THIS BIN
								BINREALK[NBIN]+=sqrt(KREALSQ);		//
								BINAA[NBIN]+=AASTAR;		//
								BINEE[NBIN]+=EESTAR;		//
							
							}
						}
						
					}
				    }
				}
				
				
				fftw_execute(IFFTA);
				fftw_execute(IFFTE);

				fftw_destroy_plan(FFTA);
				fftw_destroy_plan(FFTE);

				fftw_destroy_plan(IFFTA);
				fftw_destroy_plan(IFFTE);
				
			}
		}
		
		
		for(int64_t SPINDEX=0; SPINDEX<NDIM; SPINDEX++){
			for(int64_t COLINDEX=0; COLINDEX<SUNGENNUM; COLINDEX++){
				
				CURFIELDA=&AFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				CURFIELDE=&EFIELDS[IVOL*(SPINDEX*SUNGENNUM+COLINDEX)];
				
				//*********************************************************************************
				//  SETUP THE FOURIER TRANSFORM PLANS FOR THIS CURRENT FIELD
				//*********************************************************************************
				
				FFTA=fftw_plan_dft(3, AXLENFFT, CURFIELDA, CURFIELDA,-1, FFTW_ESTIMATE);
				IFFTA=fftw_plan_dft(3, AXLENFFT, CURFIELDA, CURFIELDA,+1, FFTW_ESTIMATE);
				
				FFTE=fftw_plan_dft(3, AXLENFFT, CURFIELDE, CURFIELDE,-1, FFTW_ESTIMATE);
				IFFTE=fftw_plan_dft(3, AXLENFFT, CURFIELDE, CURFIELDE,+1, FFTW_ESTIMATE);
				
				fftw_execute(FFTA);
				fftw_execute(FFTE);

				//-----------------------------------------------------------------------
				//	MCNT, the counter for the negative frequencies is always 
				//	incremented at the end of each loop step since,
				//	the zero position does not have a counterpart (the NYQUIST
				//	FREQUENCY takes this role) and +K <-> -K sorting starts from
				//	step 1 NOT 0.
				//-----------------------------------------------------------------------
				
				for (ACNT[0]=0, MCNT[0]=0, PCNT=0; ACNT[0]<AXLEN[0]; ACNT[0]++, MCNT[0]=STRIDES[0]*(AXLEN[0]-ACNT[0]) ){
				  for (ACNT[1]=0, MCNT[1]=MCNT[0]; ACNT[1]<AXLEN[1]; ACNT[1]++, MCNT[1]=MCNT[0]+STRIDES[1]*(AXLEN[1]-ACNT[1])){
				      for (ACNT[2]=0, MCNT[2]=MCNT[1]; ACNT[2]<AXLEN[2]; ACNT[2]++, MCNT[2]=MCNT[1]+STRIDES[2]*(AXLEN[2]-ACNT[2]), PCNT++){
						
					register double KNAIVSQ=0.;
					for(int DIM=0; DIM<NDIM; DIM++) KNAIVSQ+=pow(PPOS[DIM][ACNT[DIM]],2);
					double KNAIV=sqrt(KNAIVSQ);      
					
					double MAXMOM=PI/SUNCLDX;
					
					if( KNAIV < MAXMOM ){

						// USE A LOGARITHMIC SPACING FOR THE MOMENTA TO BETTER ACCOMMODATE THE LARGE
						// FLUCTUATIONS AT THE LOWER MOMENTA
							
						const int NBIN=(int)floor( BINNUM* ((log(KNAIVSQ/(2.*MAXMOM*2.*MAXMOM)) +log(AXLEN[0]*AXLEN[0]))/log(AXLEN[0]*AXLEN[0])) );

						if(NBIN<BINNUM&&NBIN>=0){			//ZERO MODE IS JUST THE AVERAGE SO WE DO NOT
												//INCLUDE IT. ALSO NUMBERS BECOME VERY LARGE
												//PROBLEMS WITH OVERFLOW MIGHT OCCUR
						
						CURFIELDA[PCNT][0]/=VOL;
						CURFIELDA[PCNT][1]/=VOL;
						
						CURFIELDE[PCNT][0]/=VOL;
						CURFIELDE[PCNT][1]/=VOL;
						
						double AASTAR=(1./(SUNGENNUM*NDIM))*(CURFIELDA[PCNT][0]*CURFIELDA[PCNT][0]+CURFIELDA[PCNT][1]*CURFIELDA[PCNT][1]);
						double EESTAR=(1./(SUNGENNUM*NDIM))*(CURFIELDE[PCNT][0]*CURFIELDE[PCNT][0]+CURFIELDE[PCNT][1]*CURFIELDE[PCNT][1]);
						
						BINAAVAR[NBIN]+=(AASTAR-BINAA[NBIN]/INBIN[NBIN])*(AASTAR-BINAA[NBIN]/INBIN[NBIN]);		//
						BINEEVAR[NBIN]+=(EESTAR-BINEE[NBIN]/INBIN[NBIN])*(EESTAR-BINEE[NBIN]/INBIN[NBIN]);		//
						}
						
					}
						
				      }
				  }
				}
				
			fftw_destroy_plan(FFTA);
			fftw_destroy_plan(FFTE);

			fftw_destroy_plan(IFFTA);
			fftw_destroy_plan(IFFTE);
				
				
			}
		}
		
		
		//*********************************************************************************
		//  DEBUGGING THE CORRECT NOISE STRUCTURE IN AVERAGING OVER ALL COLOR AND
		//  SPATIAL DIRECTIONS
		//*********************************************************************************
		
		char FNAME[256];
			
		sprintf(FNAME,"MeasuredModes.dat");
		POUT=fopen(FNAME,"a");	
		fprintf(POUT, "%d ", FILENUM);
		for (int k=0; k<BINNUM; k++) fprintf(POUT, "%.10e %.10e ", BINAA[k]/INBIN[k], sqrt(BINAAVAR[k]/(INBIN[k]*(INBIN[k]-1))) );
		for (int k=0; k<BINNUM; k++) fprintf(POUT, "%.10e %.10e ", BINEE[k]/INBIN[k], sqrt(BINEEVAR[k]/(INBIN[k]*(INBIN[k]-1))) );
		fprintf(POUT, "\n");
		fclose(POUT);
		
		sprintf(FNAME,"Spectrum.%d",FILENUM);
		POUT=fopen(FNAME,"w");
		for (int k=0; k<BINNUM; k++) if(INBIN[k]!=0) fprintf(POUT, "%.10e %.10e %.10e %.10e %.10e %.10e\n",BINNAIVEK[k]/INBIN[k],BINREALK[k]/INBIN[k], BINAA[k]/INBIN[k], sqrt(BINAAVAR[k]/(INBIN[k]*(INBIN[k]-1))), BINEE[k]/INBIN[k], sqrt(BINEEVAR[k]/(INBIN[k]*(INBIN[k]-1))) );	
		fclose(POUT);

		sprintf(FNAME,"MeasuredAAnisotropy.dat");
		POUT=fopen(FNAME,"a");	
		fprintf(POUT, "%d %e %e %e %e %e %e %e %e %e\n", FILENUM, (1./log(10.))*log(0.5*AASTARxy/AASTARz), (1./log(10.))*log(0.5*AASTARyz/AASTARx), (1./log(10.))*log(0.5*AASTARxz/AASTARy), AASTARx,AASTARy,AASTARz,AASTARxy,AASTARyz,AASTARxz);  
		fclose(POUT);	
		
		sprintf(FNAME,"MeasuredEAnisotropy.dat");
		POUT=fopen(FNAME,"a");
		fprintf(POUT, "%d %e %e %e %e %e %e %e %e %e\n", FILENUM, (1./log(10.))*log(0.5*EESTARxy/EESTARz), (1./log(10.))*log(0.5*EESTARyz/EESTARx), (1./log(10.))*log(0.5*EESTARxz/EESTARy), EESTARx,EESTARy,EESTARz,EESTARxy,EESTARyz,EESTARxz);  	
		fclose(POUT);
			
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "*** AAnisotropy in X: %+e                    ***\n",(1./log(10.))*log(0.5*AASTARyz/AASTARx));
		fprintf(stderr, "*** AAnisotropy in Y: %+e                    ***\n",(1./log(10.))*log(0.5*AASTARxz/AASTARy));
		fprintf(stderr, "*** AAnisotropy in Z: %+e                    ***\n",(1./log(10.))*log(0.5*AASTARxy/AASTARz));
		fprintf(stderr, "***                                                    ***\n");

		delete[] INBIN;
		delete[] BINNAIVEK;
		delete[] BINREALK;
		delete[] BINAA;
		delete[] BINEE;
		delete[] BINAAVAR;
		delete[] BINEEVAR;	
		
		fftw_free(AFIELDS);
		fftw_free(EFIELDS);

		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(PPOS[DIM]);
		fftw_free(PPOS);
		
		for(int DIM=0;DIM<NDIM;DIM++) fftw_free(MOM[DIM]);
		fftw_free(MOM);	
		
		
	}

    void SUN_CLASS_SPATIALPLAQ(double* FIELDNOW, Complex* SpatialPlaqTr, void* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		double SUNCLDT=CurParam->dt;
		double SUNCLCOUPL=CurParam->g;
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
			
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		SpatialPlaqTr[0]=0;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
							
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);			

				for(int64_t mu=0; mu<NDIM; mu++)			
					for(int64_t nu=mu+1; nu<NDIM; nu++){
										
					
					int64_t POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					int64_t POS3=0;
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					Matrix P1,TMP;
				
					P1 =&FIELDNOW[POS + (mu)*USTRIDE];
					P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
					TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
					P1*=TMP;
					TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
					P1*=TMP;

					SpatialPlaqTr[0]+=P1.Trace();
					
					
				}
			}
		    }
		}
		
		SpatialPlaqTr[0]/=VOL*SUN*NDIM;	
		
	}


    void SUN_CLASS_SPTIALWL22(double *FIELDNOW, Complex* WL22,Parameters* param)
	{
		
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		WL22[0]=0;

		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				for(int64_t mu=1; mu<NDIM; mu++){
				  for(int64_t nu=0; nu<mu; nu++){
				
					Matrix TMP,P1;
					
					int64_t POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					int64_t POS3=0;	
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==mu)?2:0) ) %AXLEN[k] )*STRIDES[k];
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					P1 = &FIELDNOW[POS  + (mu)*USTRIDE];
					P1*= &FIELDNOW[POS2 + (mu)*USTRIDE];
					P1*= &FIELDNOW[POS3 + (nu)*USTRIDE];
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?2:0) + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;

					P1*= &FIELDNOW[POS2 + (nu)*USTRIDE];
					
					POS3=0;	
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?2:0) + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
					P1*=TMP;
					
					POS2=0;	
					for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==nu)?2:0) ) %AXLEN[k] )*STRIDES[k];
					POS2*=ESTRIDE+SUNGENNUM*NDIM;
				
					POS3=0;	
					for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];
					POS3*=ESTRIDE+SUNGENNUM*NDIM;
					
					TMP.Dagger(&FIELDNOW[POS2 + (mu)*USTRIDE]);
					P1*=TMP;
					TMP.Dagger(&FIELDNOW[POS3 + (nu)*USTRIDE]);
					P1*=TMP;
					TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
					P1*=TMP;

					WL22[0]+=P1.Trace()/(SUN*NDIM*VOL);
						
				}}   
			}
		    }
		}	
	}

    void SUN_CLASS_TMPWLOOP(double* InitUConf, double* CurrentUConf, int RMIN, int RMAX , Complex* WLOOP, Parameters* param)
	{
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		for(int k=0;k<(RMAX-RMIN+1);k++) WLOOP[k]=0.;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);

				for (int64_t PLANE=0;PLANE<NDIM;PLANE++){

					Matrix CurLoop, TMP;
					
					for(int dx=RMIN; dx<=RMAX; dx++){					
						
						//-----------------------------------------------------------------------
						//	CONSTRUCT THE "TEMPORAL" PATH
						//-----------------------------------------------------------------------

						CurLoop.Unit();

						//-----------------------------------------------------------------------
						//	CONSTRUCT THE "RIGHT" PATH
						//-----------------------------------------------------------------------

						if(dx!=0) {
							TMP=&InitUConf[ POS+(PLANE)*USTRIDE];
						}else{
							TMP.Unit();
						}
						
						for(int dxx=1;dxx<dx;dxx++){

							int64_t POS2=0;
							for(int i=0;i<NDIM;i++) POS2+= ( (ACNT[i]+ ((PLANE==i)?dxx:0))%AXLEN[i] )*STRIDES[i];
							POS2*=(ESTRIDE+SUNGENNUM*NDIM);
							
							TMP*=&InitUConf[ POS2+(PLANE)*USTRIDE];

						}
						
						CurLoop*=TMP;
						
						//-----------------------------------------------------------------------
						//	CONSTRUCT THE "SECOND TEMPORAL PATH" PATH: SINCE A^0=0 JUST 1
						//-----------------------------------------------------------------------

						//-----------------------------------------------------------------------
						//	CONSTRUCT THE "SPATIALLEFT" PATH
						//-----------------------------------------------------------------------
						
						if(dx!=0){
							TMP=&CurrentUConf[ POS+(PLANE)*USTRIDE];
						}else{
							TMP.Unit();
						}
						
						for(int dxx=1;dxx<dx;dxx++){

							int64_t POS2=0;
							for(int i=0;i<NDIM;i++) POS2+= ( (ACNT[i]+ ((PLANE==i)?dxx:0))%AXLEN[i] )*STRIDES[i];
							POS2*=(ESTRIDE+SUNGENNUM*NDIM);
							
							TMP*=&CurrentUConf[ POS2+(PLANE)*USTRIDE];

						}
						
						TMP.Dagger();
						
						CurLoop*=TMP;
						
						WLOOP[ (dx-RMIN) ] += CurLoop.Trace();
											
					}
					
				}			
				
				
			}	
		    }
		}
		
		
		for(int dx=RMIN; dx<=RMAX; dx++) WLOOP[ (dx-RMIN) ]/=SUN*NDIM*VOL;
	}


    void SUN_CLASS_GAUSS_CONSTR(double* FIELD, double* DeviationLocal, double* DeviationGlobal, void* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLCOUPL=CurParam->g;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		DeviationLocal[0]=0.;
		DeviationGlobal[0]=0.;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){
							
				Matrix E1,E2,TMP;
				Matrix U1,U2;
				
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				double CFac=16.*SUNCLDT*sqrt(SUN)/(SUNCLCOUPL);

				Complex *CurDeV;
				CurDeV= new Complex [SUNGENNUM];
				for(int i=0;i<SUNGENNUM;i++) CurDeV[i]=0.;
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				for(int64_t DIM=0; DIM<NDIM; DIM++){
				
					int64_t POS2=0;
					for (int k=0;k<NDIM;k++) POS2+= ( (ACNT[k] - ((k==DIM)?1:0) + AXLEN[k] )%AXLEN[k] )*STRIDES[k];			
					POS2*=(ESTRIDE+SUNGENNUM*NDIM);
				
					E1.FROM_REAL_SUN_GENERATORS( &FIELD[POS +ESTRIDE+SUNGENNUM*DIM] );
					E2.FROM_REAL_SUN_GENERATORS( &FIELD[POS2+ESTRIDE+SUNGENNUM*DIM] );
					
					U2= &FIELD[POS2+USTRIDE*DIM];

					TMP.Dagger(U2);
					TMP*=E2;
					TMP*=U2;
					
					TMP-=E1;
					
					for(int i=0;i<SUNGENNUM;i++){
						CurDeV[i]+=TMP.TrGen(i);
					}		
						
				}
				
				for(int i=0;i<SUNGENNUM;i++) DeviationLocal[0] += CFac*(fabs(CurDeV[i].real()) + fabs(CurDeV[i].imag()));
				for(int i=0;i<SUNGENNUM;i++) DeviationGlobal[0] += CFac*(CurDeV[i].real() + CurDeV[i].imag());
				
				delete[] CurDeV;
				
				
			}
		    }
		}
		
		DeviationLocal[0]/=VOL;
		DeviationGlobal[0]/=VOL;
		
		
	}
		
    void SUN_CLASS_ENERGY(double* FIELDNOW, double* CENRG, double* CMNRG, Parameters* param)
	{
		
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		double SUNCLCOUPL=CurParam->g;
		double SUNCLDX=CurParam->dx;
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;			
		
		CMNRG[0]=0.;
		CENRG[0]=0.;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]++){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]++){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]++){

				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;

				double MAGFAC=2./(SUNCLCOUPL*SUNCLCOUPL*SUNCLDX);
				double EFAC=2.*SUN*SUNCLDX/(SUNCLCOUPL*SUNCLCOUPL);
				
				int64_t POS=pos1 + ACNT[2]*STRIDES[2];
				POS*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				//-----------------------------------------------------------------------
				//	MAGNETIC ENERGY PART                        
				//-----------------------------------------------------------------------

				Matrix P1;
				Matrix Unit;
				
					for(int64_t mu=0; mu<NDIM-1; mu++){
					for(int64_t nu=mu+1; nu<NDIM; nu++){

						Matrix TMP;
		
						int64_t POS2=0;	
						for (int k=0;k<NDIM;k++) POS2+= ( ( ACNT[k] + ((k==mu)?1:0) ) %AXLEN[k] )*STRIDES[k];
						POS2*=ESTRIDE+SUNGENNUM*NDIM;
					
						int64_t POS3=0;
						for (int k=0;k<NDIM;k++) POS3+= ( ( ACNT[k] + ((k==nu)?1:0) ) %AXLEN[k] )*STRIDES[k];					
						POS3*=ESTRIDE+SUNGENNUM*NDIM;
						
						P1= &FIELDNOW[POS  + (mu)*USTRIDE];
						P1*=&FIELDNOW[POS2 + (nu)*USTRIDE];
						TMP.Dagger(&FIELDNOW[POS3 + (mu)*USTRIDE]);
						P1*=TMP;
						TMP.Dagger(&FIELDNOW[POS + (nu)*USTRIDE]);
						P1*=TMP;
						
						CMNRG[0]+=MAGFAC*( (SUN)-P1.RealTrace())/(VOL);
							
				}}
				
				//-----------------------------------------------------------------------
				//	ELECTRIC ENERGY PART                        
				//-----------------------------------------------------------------------
				
				for(int64_t mu=0; mu<NDIM; mu++){
				  for(int64_t gen=0; gen<SUNGENNUM; gen++){
					  
					CENRG[0]+=EFAC*( FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen]*FIELDNOW[POS+ESTRIDE+SUNGENNUM*mu+gen] )/(VOL);
					
				  }
				}
				
			}
		    }
		}	
	}

    void SUN_CLASS_ENERGY_FMUNU(double* FIELDNOW, double *FIELDNEXT, double* FIELDBEF, double* EnergyE, double* EnergyM, Parameters* param)
	{
	  
	//CHECKED ON MAY 21st 2015 ==> OK	A. ROTHKOPF (BUT DO NOT UNDERSTAND WHY DIFFERENT FROM NAIVE ENERGY DEF)
	  
		Parameters *CurParam;
		CurParam=(Parameters*) param;

		int64_t AXLEN[NDIM];
		int64_t STRIDES[NDIM];
		int64_t ACNT[NDIM];
		
		AXLEN[0]=CurParam->NX;
		AXLEN[1]=CurParam->NY;
		AXLEN[2]=CurParam->NZ;
		
		double SUNCLDT=CurParam->dt;
		double SUNCLDX=CurParam->dx;
		double SUNCLCOUPL=CurParam->g;
		
		int *mucomp=CurParam->mucomp;
		int *nucomp=CurParam->nucomp;
		int *sgcomp=CurParam->sgcomp;
		int *rocomp=CurParam->rocomp;
		
		STRIDES[0]=AXLEN[2]*AXLEN[1];
		STRIDES[1]=AXLEN[2];
		STRIDES[2]=1;
		
		int STEP=1;
		
		double SUNFAC=SUN;
		
		double VOL=AXLEN[0]*AXLEN[1]*AXLEN[2];
		
		EnergyE[0]=0.;
		EnergyM[0]=0.;

		Matrix** Fmunu;
		
		Fmunu = new Matrix* [4];
		for (int k=0;k<4;k++) Fmunu[k]= new Matrix [4];
		
		//-----------------------------------------------------------------------
		//	WE DIRECTLY CALCULATE F_\mu\nu USING 
		//	THE FIELD TENSOR FROM THE ANTISYMMETRIZED PLAQUETTE
		//	SIN(X)~X
		//-----------------------------------------------------------------------

		Matrix PSumSTART;
			
		for  (ACNT[0]=0; ACNT[0]<AXLEN[0]; ACNT[0]+=STEP){
			
		    int64_t pos0=ACNT[0]*STRIDES[0];
		    
		    for (ACNT[1]=0; ACNT[1]<AXLEN[1]; ACNT[1]+=STEP){
			    
			int64_t pos1=pos0 + ACNT[1]*STRIDES[1];
			
			for (ACNT[2]=0; ACNT[2]<AXLEN[2]; ACNT[2]+=STEP){
							
				int64_t USTRIDE=2*SUN*SUN;
				int64_t ESTRIDE=2*SUN*SUN*NDIM;
				
				int64_t POSUSTART=pos1 + ACNT[2]*STRIDES[2];
				POSUSTART*=(ESTRIDE+SUNGENNUM*NDIM);			
				
				//-----------------------------------------------------------------------
				//	CALCULATE THE SYMMETRIZED PLAQUETTE
				//-----------------------------------------------------------------------
				
				FMUNU_SPATIAL( Fmunu, FIELDNOW, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL);
				FMUNU_TMPORAL( Fmunu, FIELDNOW, FIELDNEXT, FIELDBEF, AXLEN, ACNT, STRIDES, SUNCLDX, SUNCLDT, SUNCLCOUPL, 0);
				
				for(int k=0;k<NDIM;k++){
				  for( int l=0;l<NDIM;l++){
				    
					Matrix TMP;
					
					TMP=Fmunu[k][l];
					TMP*=Fmunu[k][l];
				    
					EnergyM[0]+=2.*TMP.RealTrace();
					
					//fprintf(stderr, "%d %d - %e \n", k,l, TMP.RealTrace());
				  }}
				  
				  
				for(int k=NDIM;k<NDIM+1;k++){
				  for( int l=0;l<NDIM;l++){
				    
					Matrix TMP;
					
					TMP=Fmunu[k][l];
					TMP*=Fmunu[k][l];
				    
					EnergyE[0]+=2.*TMP.RealTrace();
					//fprintf(stderr, "%d %d - %e \n", k,l, TMP.RealTrace());
				  }}
				  
				for(int k=0;k<NDIM;k++){
				  for( int l=NDIM;l<NDIM+1;l++){
				    
					Matrix TMP;
					
					TMP=Fmunu[k][l];
					TMP*=Fmunu[k][l];
				    
					EnergyE[0]+=2.*TMP.RealTrace();
					//fprintf(stderr, "%d %d - %e \n", k,l, TMP.RealTrace());
				  }}
				
				//fprintf(stderr, "DEBUG\n"); exit(-1);
				
			}
		    }
		}
		
		
		EnergyE[0]/=4.*VOL;
		EnergyM[0]/=4.*VOL;
		
		for (int k=0;k<4;k++) delete[] Fmunu[k];
		delete[] Fmunu;
	}

int main(int argc,char *argv[])
{

    //-----------------------------------------------------------------------
    //	Simulation Parameters
    //-----------------------------------------------------------------------

	//GEOMETRY 
  
// 	int NX;
// 	int NY;
// 	int NZ;
// 	
// 	int NT;
// 	
// 	double dx;
// 	double dt;
	
	//Physics Parameters
	
	// for Yang-Mills
// 	double g;
// 	double BetaL;
// 	
// 	// for Chiral EFT 
// 	double ColorConductivity;
// 	double DiffusionConst;
// 	double ConstantChemicalPotential;
// 	int NFermions;
	
	// Cooling Parameters
	
// 	int WHICHCOOLPRESCR;
// 	int COARSENING;
// 
// 	int COOLDEPTH;
// 	int COOLEVERY;
// 	
// 	double dtau;
// 
// 	double ConditioningCooling;
	
	// Initial Condition Parameters
	
// 	int INITCOND;
// 	double InitCondN5;

	// For Thermal Equilibrium
	
// 	int METRPRECONSTEPS;
// 	int MOOREMIXINGSTPS;
// 	int MOOREPROJMIXSTP;

	// For free Gaussian Fields
	
// 	double SigmaX,SigmaY,SigmaZ,InitGGFStrength;
	
	//Measurement Parameters

// 	int MEASEACH;
// 	
// 	int WloopRMax;
// 	int WloopRMin;
// 	
// 	int SpecBins;
	
// 	int WHICHNCSDEF
  
	// Time Evolution Parameters
	
// 	int WHICHREALTIME;
// 
	double SEED;
	
	int SAVEEACH;
	
	int VERBOSITY;

	double N5SWITCHONAT;
	
	char CONFIGFILENAME[512];
	
    //-----------------------------------------------------------------------
    //	Auxiliary DEFINITIONS
    //-----------------------------------------------------------------------
	
	int STARTINGTIME=0;
	
	//LeviCivita Implementation
	
	int *mucomp, *nucomp, *sgcomp, *rocomp;
	double *LCval;
	
    //-----------------------------------------------------------------------
    //	Getting Parameters from Command Line
    //-----------------------------------------------------------------------

	Parameters MyParam;
	
if(argc!=39){
	fprintf(stderr, "CORRECT NUMBER OF ARGUMENTS IS: %3d GIVEN :%3d\n", 39, argc);
	exit(-1);
}

	MyParam.NX=atoi(argv[11]);
	MyParam.NY=atoi(argv[37]);
	MyParam.NZ=atoi(argv[38]);
	MyParam.NT=atoi(argv[12]);

	MyParam.dt=atof(argv[2]);
	MyParam.dx=atof(argv[5]);
		
	MyParam.BetaL=atof(argv[1]);
	MyParam.g=atof(argv[4]);

	MyParam.ColorConductivity=atof(argv[22]);
	MyParam.DiffusionConst=atof(argv[28]);
	MyParam.ConstantChemicalPotential=atof(argv[29]);
	MyParam.NFermions=atoi(argv[30]);

	MyParam.WHICHCOOLPRESCR=atoi(argv[19]);
	MyParam.COOLEVERY=atoi(argv[8]);
	MyParam.COOLDEPTH=atoi(argv[9]);
	MyParam.COARSENING=atoi(argv[25]);
	MyParam.dtau=atof(argv[3]);
	MyParam.ConditioningCooling=atof(argv[23]);
	
	MyParam.INITCOND=atoi(argv[10]); // INITCOND=1 Anisotropic Gaussian A's,INITCOND=3 Thermal, |INITCOND|=2 Unit U,  INITCOND>0 Zero E, INITCOND<0 Random E
	MyParam.InitCondN5=atof(argv[26]);
	
	MyParam.METRPRECONSTEPS=atoi(argv[31]);
	MyParam.MOOREMIXINGSTPS=atoi(argv[32])*(0.075/MyParam.dt);
	MyParam.MOOREPROJMIXSTP=atoi(argv[33]);
	
	MyParam.SigmaX=atof(argv[13]);
	MyParam.SigmaY=atof(argv[14]);
	MyParam.SigmaZ=atof(argv[15]);
	MyParam.InitGGFStrength=atof(argv[16]);
	
	MyParam.MEASEACH=atoi(argv[6]);
	MyParam.WloopRMax=atoi(argv[7]);
	MyParam.WloopRMin=atoi(argv[34]);
	MyParam.SpecBins=atoi(argv[27]);
	MyParam.WHICHNCSDEF=atoi(argv[35]);
	
	MyParam.WHICHREALTIME=atoi(argv[17]);
	SEED=atof(argv[18]);
	MyParam.VERBOSITY=atoi(argv[20]);
	VERBOSITY=MyParam.VERBOSITY;
	MyParam.SAVEEACH=atoi(argv[24]);
	
	sprintf(CONFIGFILENAME,argv[36]);
	
	MyParam.DispersionSwitch=1;
	
	MyParam.N5SWITCHONAT=atof(argv[21]);
	N5SWITCHONAT=MyParam.N5SWITCHONAT;
	
	int DYNAMICALN5=0;
	if(MyParam.WHICHREALTIME==2||MyParam.WHICHREALTIME==4||MyParam.WHICHREALTIME==7||MyParam.WHICHREALTIME==8) DYNAMICALN5=1;
	if(MyParam.WHICHREALTIME==5) DYNAMICALN5=2;
	MyParam.DYNAMICALN5=DYNAMICALN5;
	
	fprintf(stderr, "**********************************************************\n");
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  AKAMATSU, ROTHKOPF, YAMAMOTO                      ***\n");
	fprintf(stderr, "***  CHIRAL INSTABILITIES IN A CLASSICAL ANOMALOUS     ***\n");
	fprintf(stderr, "***  EFFECTIVE FIELD THEORY OF SU(%3d) GAUGE FIELDS    ***\n",SUN);
	fprintf(stderr, "***  WITH Nf=%3d SPECIES OF CHIRAL FERMIONS            ***\n",MyParam.NFermions);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  SEED: %.20e                  ***\n",SEED);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  NX:%3d  NY:%3d  NZ:%3d      NT:%6d             ***\n",MyParam.NX,MyParam.NY,MyParam.NZ,MyParam.NT);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  dx:%e  dt:%e                  ***\n",MyParam.dx, MyParam.dt);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  Coupling: g:%e                          ***\n",MyParam.g);
	fprintf(stderr, "***        BetaL:%e                          ***\n",MyParam.BetaL);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  Cool every %3d dt steps by %3d dtau steps         ***\n",MyParam.COOLEVERY,MyParam.COOLDEPTH);
	fprintf(stderr, "***  with dtau: %e                           ***\n",MyParam.dtau);
	fprintf(stderr, "***  using prescription: %3d                           ***\n",MyParam.WHICHCOOLPRESCR);
	if(MyParam.COARSENING!=0) fprintf(stderr, "***  as well as coarsening after %4d steps.           ***\n",MyParam.COARSENING);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  Measure Observables at every %3d steps            ***\n",MyParam.MEASEACH);
	if(MyParam.WloopRMax>0) fprintf(stderr, "***  Measure the Wilson Loop between R:%3d and %3d     ***\n",MyParam.WloopRMin,MyParam.WloopRMax);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  TIME EVOLUTION PROCEEDS AS:                       ***\n");
	
if(MyParam.WHICHREALTIME==0){
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  HAMILTONIAN DYNAMICS ONLY                         ***\n");
	fprintf(stderr, "***                                                    ***\n");
	
}else if (MyParam.WHICHREALTIME==1) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  LANGEVIN DYNAMICS WIHTOUT N5                      ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***                                                    ***\n");  
	
}else if (MyParam.WHICHREALTIME==2) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  LANGEVIN DYNAMICS WITH DYNAMICAL N5 (SPAT DEP)    ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***       DiffusionConst: %e                 ***\n",MyParam.DiffusionConst);
	fprintf(stderr, "***       Numbr Fermions: %3d                         ***\n",MyParam.NFermions);
	fprintf(stderr, "***       Init Cond n5  : %e                 ***\n",MyParam.InitCondN5);
 	fprintf(stderr, "***                                                    ***\n");  
 	if(MyParam.COARSENING!=0) { fprintf(stderr, "COARSENING FOR NCS AND SPATIAL N5 EVOLUTION NOT COMPATIBLE\n"); exit(-1);}
 
}else if (MyParam.WHICHREALTIME==3) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  CHIRAL EFT WITHOUT DYNAMICAL N5                   ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***                                                    ***\n");  
  
}else if (MyParam.WHICHREALTIME==4) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  CHIRAL EFT WITH DYNAMICAL N5 (NO SPATIAL DEP)     ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***       Numbr Fermions: %3d                         ***\n",MyParam.NFermions);
	fprintf(stderr, "***       Init Cond n5  : %e                 ***\n",MyParam.InitCondN5);
	fprintf(stderr, "***                                                    ***\n");  
  
}else if (MyParam.WHICHREALTIME==5) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  CHIRAL EFT WITH DYNAMICAL N5 (WITH SPATIAL DEP)   ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***       DiffusionConst: %e                 ***\n",MyParam.DiffusionConst);
	fprintf(stderr, "***       Numbr Fermions: %3d                         ***\n",MyParam.NFermions);
	fprintf(stderr, "***       Init Cond n5  : %e                 ***\n",MyParam.InitCondN5);
	fprintf(stderr, "***                                                    ***\n");  
  
	if(MyParam.COARSENING!=0) { fprintf(stderr, "COARSENING FOR NCS AND SPATIAL N5 EVOLUTION NOT COMPATIBLE\n"); exit(-1);}
	
}else if (MyParam.WHICHREALTIME==6) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  TASLLER LANGEVIN WITHOUT N5                       ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***                                                    ***\n");  
  
}else if (MyParam.WHICHREALTIME==7) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  TASLLER LANGVN WITH DYNAMICAL N5 (NO SPATIAL DEP) ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***       Numbr Fermions: %3d                         ***\n",MyParam.NFermions);
	fprintf(stderr, "***       Init Cond n5  : %e                 ***\n",MyParam.InitCondN5);
	fprintf(stderr, "***                                                    ***\n");   
	
}else if (MyParam.WHICHREALTIME==8) {
  
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***  CHIRAL EFT WITHOUT DYNAMICAL N5 NO DISPERSION      ***\n");
	fprintf(stderr, "***         ColorConduct: %e                 ***\n",MyParam.ColorConductivity);
	fprintf(stderr, "***                                                    ***\n");  
	
} else {
	fprintf(stderr, "PLEASE CHOOSE ONE TIME EVOLUTION BETWEEN 0 AND 8\n"); exit(-1);
}
	fprintf(stderr, "**********************************************************\n\n");
	
    //-----------------------------------------------------------------------
    //	SHow conventions used for Generators
    //-----------------------------------------------------------------------
	
	fprintf(stderr, "**********************************************************\n");
	fprintf(stderr, "***  CONVENTIONS USED FOR THE GENERATORS               ***\n");
	fprintf(stderr, "**********************************************************\n\n");
	
#if (SUN==3)
	double CONVGM[8];
	Matrix TmpGM;
	
	for(int k=0; k<8;k++){
	
		for(int j=0; j<8;j++)CONVGM[j]=0.;
		CONVGM[k]=1.;
		
		TmpGM.FROM_REAL_SUN_GENERATORS(CONVGM);
		
		TmpGM.print();
		
	}
#elif (SUN==2)
	double CONVGM[3];
	Matrix TmpGM;
	
	for(int k=0; k<3;k++){
	
		for(int j=0; j<3;j++)CONVGM[j]=0.;
		CONVGM[k]=1.;
		
		TmpGM.FROM_REAL_SUN_GENERATORS(CONVGM);
		
		TmpGM.print();
		
	}
#endif

	//-----------------------------------------------------------------------
	//	Initialize the LeviCivita Summation for Chern Simons
	//-----------------------------------------------------------------------
	
	fprintf(stderr, "**********************************************************\n");
	fprintf(stderr, "***  THE LEVI CIVITA SYMBOL FOR Fmunu tilde(Fmunu)     ***\n");
	fprintf(stderr, "***                                                    ***\n");  

	SetLeviCivita(&MyParam);

	for(int k=0;k<24;k++){
		fprintf(stderr, "*** e_abcd!=0 a=%3d b=%3d c=%d d=%d   val:%+e  ***\n",			
		MyParam.mucomp[k],
		MyParam.nucomp[k],
		MyParam.sgcomp[k],
		MyParam.rocomp[k],
		MyParam.LCval[k]);
	  
	}
	
	fprintf(stderr, "**********************************************************\n\n");


	//-----------------------------------------------------------------------
	//	Setup momenta for the fourier space routines
	//-----------------------------------------------------------------------

	int64_t ACNT[3];
	
	int AXLEN[3];
	
	AXLEN[0]=MyParam.NX;
	AXLEN[1]=MyParam.NY;
	AXLEN[2]=MyParam.NZ;

	int64_t STRIDES[3];
	STRIDES[0]=AXLEN[2]*AXLEN[1];
	STRIDES[1]=AXLEN[2];
	STRIDES[2]=1;
	
	double SUNCLDX=MyParam.dx;
	
	double **PPOS;
	PPOS=(double**)fftw_malloc(sizeof(double*)*NDIM);
	for(int DIM=0;DIM<NDIM;DIM++) PPOS[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

	for(int DIM=0;DIM<NDIM;DIM++){
		for(int PCNT=0;PCNT<=AXLEN[DIM]/2;PCNT++) PPOS[DIM][PCNT]=PCNT*((2.*PI)/(((double)AXLEN[DIM])*((double)SUNCLDX)));
		for(int PCNT=AXLEN[DIM]/2+1;PCNT<AXLEN[DIM];PCNT++) PPOS[DIM][PCNT]=-PPOS[DIM][AXLEN[DIM]-PCNT];
	}

	double **MOM;
	MOM=(double**)fftw_malloc(sizeof(double*)*NDIM);
	for(int DIM=0;DIM<NDIM;DIM++) MOM[DIM]=(double*)fftw_malloc(sizeof(double)*AXLEN[DIM]);

	for(int DIM=0;DIM<NDIM;DIM++){
		for(int PCNT=0;PCNT<AXLEN[DIM];PCNT++) MOM[DIM][PCNT]=sqrt(2.5-(8./3.)*cos(2.*PCNT*PI/((double)AXLEN[DIM]))+(1./6.)*cos(4.*PCNT*PI/((double)AXLEN[DIM])))/SUNCLDX;
	}


	//-----------------------------------------------------------------------
	//	Initialize the Fields
	//-----------------------------------------------------------------------
	
	double *FIELDNOW, *FIELDNXT, *FIELDBEF, *FIELDCOOLNOW, *FIELDCOOLNXT, *FIELDCOOLBEF, *FIELDCOOLTMP, *N5NOW, *N5NEXT, *N5SCRATCH;
	
	int64_t VVAL=MyParam.NX*MyParam.NY*MyParam.NZ*(SUNGENNUM*NDIM+2*SUN*SUN*NDIM);
	int64_t SVOL=MyParam.NX*MyParam.NY*MyParam.NZ;
	
	FIELDNOW= new double [VVAL];
	FIELDNXT= new double [VVAL];
	FIELDBEF= new double [VVAL];
	
	FIELDCOOLNOW= new double [VVAL];
	FIELDCOOLNXT= new double [VVAL];
	FIELDCOOLBEF= new double [VVAL];
	FIELDCOOLTMP= new double [VVAL];

	for(int64_t k=0; k< VVAL; k++){
		FIELDNOW[k]=0.;
		FIELDNXT[k]=0.;
		FIELDBEF[k]=0.;
		FIELDCOOLNOW[k]=0.;
		FIELDCOOLNXT[k]=0.;
		FIELDCOOLBEF[k]=0.;
	}

	if(DYNAMICALN5==1){
	  
	  	N5NOW= new double [1];
		N5NEXT= new double [1];	
		N5SCRATCH= new double [1];

		N5NOW[0]=0;
		N5NEXT[0]=0;
		N5SCRATCH[0]=0;
		
		
	}else if (DYNAMICALN5==2) {
	  
		N5NOW= new double [MyParam.NX*MyParam.NY*MyParam.NZ];
		N5NEXT= new double [MyParam.NX*MyParam.NY*MyParam.NZ];	
		N5SCRATCH= new double [MyParam.NX*MyParam.NY*MyParam.NZ];	  
	  
		for(int64_t k=0; k< SVOL; k++){
			N5NOW[k]=0.;
			N5NEXT[k]=0.;
			N5SCRATCH[k]=0.;
		}
	}
		
	// FOR THE TEMPORAL SPATIAL WILSON LOOP WE MIGHT NEED A COPY OF THE INITIAL CONFIGURATION	
		
	double* INITCONF;
	
	//-----------------------------------------------------------------------
	//	Initialize the Random Number Generator
	//-----------------------------------------------------------------------
	
	int MYRANK=0;
	
	TLP_ALGORITHM **GSLRAN;
	
	GSLRAN = new TLP_ALGORITHM* [THREADS];
	for(int k=0;k<THREADS;k++) GSLRAN[k] = (TLP_ALGORITHM *) malloc(sizeof(TLP_ALGORITHM));
	for(int k=0;k<THREADS;k++) TLP_ALGORITHM_SET (GSLRAN[k], 2);
	
	TLP_RANDGEN *MPIRNGSEED;
	MPIRNGSEED = TLP_RANDGEN_INITIALIZE( GSLRAN[0], SEED );
	for(int k=0;k<(MYRANK+1)*120;k++)  SEED=TLP_RAND_GEN_INT_UNIFORM (MPIRNGSEED,4294967295);
	
	TLP_RANDGEN **Rng;
	Rng = new TLP_RANDGEN* [THREADS];
	Rng[0] = TLP_RANDGEN_INITIALIZE( GSLRAN[0], SEED );
	for(int k=0;k<150;k++) double tmp =TLP_RAND_GEN_INT_UNIFORM (Rng[0],4294967295);
	for(int k=1;k<THREADS;k++) Rng[k] = TLP_RANDGEN_INITIALIZE( GSLRAN[k], TLP_RAND_GEN_INT_UNIFORM (Rng[0],4294967295) );	

	//-----------------------------------------------------------------------
	//	Initialize the CHIRAL CHEMICAL POTENTIAL IF SET
	//-----------------------------------------------------------------------

	if(MyParam.InitCondN5!=0.){
	      if(DYNAMICALN5==1){
		
		      N5NOW[0]=MyParam.InitCondN5;
		      N5NEXT[0]=0.;
		      N5SCRATCH[0]=0.;
		      
		      
	      }else if (DYNAMICALN5==2) {
		
		      for(int64_t k=0; k< SVOL; k++){
			      N5NOW[k]=MyParam.InitCondN5;
			      N5NEXT[k]=0.;
			      N5SCRATCH[k]=0.;
		      }
	      }
	}
	
	//-----------------------------------------------------------------------
	//	Initialize the GAUGE FIELDS AND MOMENTA
	//-----------------------------------------------------------------------
	
	fprintf(stderr, "**********************************************************\n");
	
	if(MyParam.INITCOND==1){
	
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  GAUSSIAN MOMENTUM SPACE INITIALIZATION            ***\n");
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***    x-Dir : g:%e                          ***\n",MyParam.SigmaX);
		fprintf(stderr, "***    y-Dir : g:%e                          ***\n",MyParam.SigmaY);
		fprintf(stderr, "***    z-Dir : g:%e                          ***\n",MyParam.SigmaZ);
		fprintf(stderr, "***  Amplitde: g:%e                          ***\n",MyParam.InitGGFStrength);
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CLASS_INITIALIZE_GAUSSIAN_MOMENTA(FIELDNOW, Rng, &MyParam);
	
	}else if (MyParam.INITCOND==2){
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  BOTH E AND U SET TO RANDOM VALUES                 ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CLASS_INITIALIZE_BOTH(FIELDNOW, Rng[0], &MyParam);

	}else if (MyParam.INITCOND==3){
	  
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  U SET TO RANDOM VALUES E SET TO ZERO              ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CLASS_INITIALIZE_BOTH(FIELDNOW, Rng[0], &MyParam);
		
	}else if (MyParam.INITCOND==4){
	  
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  U SET TO UNITY E SET TO ZERO                      ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CLASS_INITIALIZE_BOTH(FIELDNOW, Rng[0], &MyParam);
		
	}else if (MyParam.INITCOND==5){
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  MOORE's THERMALIZATION PRESCRIPTION               ***\n");
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***    PRECONDITION U FOR %4d STEPS                   ***\n",MyParam.METRPRECONSTEPS);
		fprintf(stderr, "***    CARRY OUT %4d DRAW E + MIXING STEPS            ***\n",MyParam.MOOREPROJMIXSTP);
		fprintf(stderr, "***    MIXING STEP PROGRESSES EOM FOR %4d STEPS       ***\n",MyParam.MOOREMIXINGSTPS);
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CLASS_INITIALIZE_BOTH(FIELDNOW, Rng[0], &MyParam);
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  PRETHERMALIZATION OF THE GAUGE LINKS (METROPOLIS) ***\n");
		fprintf(stderr, "***       BETA: %+.5e                           ***\n",MyParam.BetaL);
		fprintf(stderr, "***                                                    ***\n");
		
		SUN_CL_3D_METROPOLIS_PRECONDITION(FIELDNOW, Rng, &MyParam);		

		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "***  THERMALIZATION ACCORDING TO MOORE WITH GC. PROJ   ***\n");
		fprintf(stderr, "***                 %3d SWEEPS                         ***\n",MyParam.MOOREPROJMIXSTP);

		for(int THRMSTEPS=0; THRMSTEPS<MyParam.MOOREPROJMIXSTP; THRMSTEPS++){
			
			if(VERBOSITY==1){
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "***  STARTING WITH GAUSSIAN E AND MIXING STEP %3d      ***\n",THRMSTEPS);
			fprintf(stderr, "***                                                    ***\n");
			}
			
			SUN_CLASS_DRAW_E_THERMAL(FIELDNOW, Rng, &MyParam);
			
			SUN_CLASS_PROJECT_E_GC(FIELDNOW,FIELDNXT, THRMSTEPS, &MyParam);

			if(VERBOSITY==1){
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "*** FOR LEAPFROG BRING E TO t=DT/2                     ***\n");
			fprintf(stderr, "***                                                    ***\n");
			}
			
			SUN_CLASS_EVOLVE_EFIELD_FIRSTSTEP(FIELDNOW, &MyParam);	  
			
			for(int MIXCNT=0; MIXCNT<MyParam.MOOREMIXINGSTPS; MIXCNT++){

				if(VERBOSITY==1){
				fprintf(stderr, "***  THM PHASE [%3d] MIXING STEP [%3d]                 ***\n",THRMSTEPS,MIXCNT);
				fprintf(stderr, "***                                                    ***\n");
				}
				
				SUN_CLASS_EVOLVE_EFIELD(FIELDNOW, FIELDNXT, &MyParam);
				SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);

				double* tmp=FIELDNOW;		
				FIELDNOW=FIELDNXT;
				FIELDNXT=tmp;
				
			}
			
			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE INITIAL ENERGY DENSITY
			//---------------------------------------------------------------------------------------------
			
			double ClassENrg, ClassMNrg;
			
			SUN_CLASS_ENERGY(FIELDNOW, &ClassENrg, &ClassMNrg, &MyParam);
			
			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE INITIAL GAUSS CONSTRAINT (LOCAL AND GLOBAL)
			//---------------------------------------------------------------------------------------------
			
			double GaussDevLoc,GaussDevGlob;
			
			SUN_CLASS_GAUSS_CONSTR(FIELDNOW, &GaussDevLoc, &GaussDevGlob, &MyParam);

			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE INITIAL PLAQUETTE
			//---------------------------------------------------------------------------------------------
			
			Complex SpPlaq;
			
			SUN_CLASS_SPATIALPLAQ(FIELDNOW, &SpPlaq, &MyParam);	
			
			if(VERBOSITY==1){
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "***        GCLOC:%+.5e     GCGLB:%+.5e  ***\n",GaussDevLoc,GaussDevGlob);
			fprintf(stderr, "***         NRG: (%.8e - %.8e)     ***\n",ClassENrg,ClassMNrg);
			fprintf(stderr, "***        PLAQ: (%+.8e - %+.8e)   ***\n",SpPlaq.real(), SpPlaq.imag() );
			fprintf(stderr, "***                                                    ***\n");
			}
			
		}	
		
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "*** THERMALIZATION DONE! OUTPUT TO HD                  ***\n");
		fprintf(stderr, "***                                                    ***\n");
		
		char OLDFILE[256];
		std::fstream TMPOUT;
			
		sprintf(OLDFILE,"INITIAL_CONFIG.dat");
			
		TMPOUT.open(OLDFILE, std::ios::out|std::ios::binary);

		TMPOUT.write ((char*)(&STARTINGTIME), sizeof(int));
		TMPOUT.write ((char*)(&MyParam.WHICHREALTIME), sizeof(int));
		
		TMPOUT.write ((char*)FIELDNOW, sizeof(double)*VVAL);
		TMPOUT.write ((char*)FIELDBEF, sizeof(double)*VVAL);
		TMPOUT.write ((char*)FIELDCOOLNOW, sizeof(double)*VVAL);
		TMPOUT.write ((char*)FIELDCOOLBEF, sizeof(double)*VVAL);
		
		if(DYNAMICALN5==1){
		    TMPOUT.write ((char*)N5NOW, sizeof(double)*1);
		    TMPOUT.write ((char*)N5NEXT, sizeof(double)*1);
		}else if (DYNAMICALN5==2){
		    TMPOUT.write ((char*)N5NOW , sizeof(double)*SVOL);
		    TMPOUT.write ((char*)N5NEXT, sizeof(double)*SVOL);  
		}
		
		for(int k=0;k<THREADS;k++) TMPOUT.write ((char*)&(GSLRAN[k]->STATE_SIZE), sizeof(int));		
		for(int k=0;k<THREADS;k++) TMPOUT.write ((char*)(Rng[k]->CURSTATE), GSLRAN[k]->STATE_SIZE);
		
		TMPOUT.close();
		
	}else if (MyParam.INITCOND==6||MyParam.INITCOND==7){
		
		std::fstream TMPOUT;
			
		TMPOUT.open(CONFIGFILENAME, std::ios::in|std::ios::binary);
		
		if(TMPOUT.is_open()==1){
					
			TMPOUT.read ((char*)(&STARTINGTIME), sizeof(int));
		
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "*** PREVIOUS INITIALIZED CONFIGURATION                 ***\n");
			fprintf(stderr, "*** FROM TIME STEP: %8d FOUND                     ***\n",STARTINGTIME);
			fprintf(stderr, "***                                                    ***\n");
			
			int GENERATEDWHICHREALTIME;
			TMPOUT.read ((char*)(&GENERATEDWHICHREALTIME), sizeof(int));

			fprintf(stderr, "*** GENERATED WITH REALTIME: %3d                       ***\n",GENERATEDWHICHREALTIME);

			TMPOUT.read ((char*)FIELDNOW, sizeof(double)*VVAL);
			TMPOUT.read ((char*)FIELDBEF, sizeof(double)*VVAL);

			TMPOUT.read ((char*)FIELDCOOLNOW, sizeof(double)*VVAL);
			TMPOUT.read ((char*)FIELDCOOLBEF, sizeof(double)*VVAL);

			int GENERATEDDYNAMICALN5=0;
			if(GENERATEDWHICHREALTIME==2||GENERATEDWHICHREALTIME==4||GENERATEDWHICHREALTIME==7||GENERATEDWHICHREALTIME==8) GENERATEDDYNAMICALN5=1;
			if(GENERATEDWHICHREALTIME==5) GENERATEDDYNAMICALN5=2;
			
			if (GENERATEDDYNAMICALN5==1||GENERATEDDYNAMICALN5==2){
			  
			    fprintf(stderr, "*** CONFIG CONTAINS THE FINITE MU FIELDS               ***\n");
			  
			    if(DYNAMICALN5==0){
			      
				 fprintf(stderr, "***    USING A CONFIG GENERATED WITH MU FOR NON MU REALT**\n");
			      
				 if(GENERATEDDYNAMICALN5==1){
				    N5NOW= new double [1];
				    N5NEXT= new double [1];	
				    N5SCRATCH= new double [1];
				 }else if (GENERATEDDYNAMICALN5==2){			   
				    N5NOW= new double [MyParam.NX*MyParam.NY*MyParam.NZ];
				    N5NEXT= new double [MyParam.NX*MyParam.NY*MyParam.NZ];	
				    N5SCRATCH= new double [MyParam.NX*MyParam.NY*MyParam.NZ];
				  }				 
				  
			    }
			    
			    if(DYNAMICALN5==1){
				TMPOUT.read ((char*)N5NOW , sizeof(double)*1);
				TMPOUT.read ((char*)N5NEXT, sizeof(double)*1);
			    }else if (DYNAMICALN5==2){			   
				TMPOUT.read ((char*)N5NOW , sizeof(double)*SVOL);
				TMPOUT.read ((char*)N5NEXT, sizeof(double)*SVOL);
			    }				 
			  
			}
			
			int CurSTATESIZE;
			
			for(int k=0;k<THREADS;k++){
				TMPOUT.read ((char*)&(CurSTATESIZE), sizeof(int));	
				if( GSLRAN[k]->STATE_SIZE !=CurSTATESIZE) { fprintf(stderr, "PROBLEM WITH STATE SIZE OF RNG %d - %d %d\n",k,GSLRAN[k]->STATE_SIZE,CurSTATESIZE); exit(-1);}
			}
				
			for(int k=0;k<THREADS;k++) TMPOUT.read ((char*)(Rng[k]->CURSTATE), GSLRAN[k]->STATE_SIZE);
			
			TMPOUT.close();

			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "*** PREVIOUSLY SAVED     CONFIGURATION READ            ***\n");
			fprintf(stderr, "***                                                    ***\n");
		} else{
			fprintf(stderr, "NO PREVIOUS CONFIGURATION: %s FOUND\n", CONFIGFILENAME); exit(-1);
		}
		
	}else if (MyParam.INITCOND==8||MyParam.INITCOND==9){
		
		std::fstream TMPOUT;
			
		TMPOUT.open(CONFIGFILENAME, std::ios::in|std::ios::binary);
		
		if(TMPOUT.is_open()==1){
					
			TMPOUT.read ((char*)(&STARTINGTIME), sizeof(int));
		
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "*** OLD      INITIALIZED CONFIGURATION                 ***\n");
			fprintf(stderr, "*** FROM TIME STEP: %8d FOUND                     ***\n",STARTINGTIME);
			fprintf(stderr, "***                                                    ***\n");
			
			int GENERATEDWHICHREALTIME;
			TMPOUT.read ((char*)(&GENERATEDWHICHREALTIME), sizeof(int));

			fprintf(stderr, "*** GENERATED WITH PREVIOUS REALTIME OPTION %3d        ***\n",GENERATEDWHICHREALTIME);

			TMPOUT.read ((char*)FIELDNOW, sizeof(double)*VVAL);
			TMPOUT.read ((char*)FIELDBEF, sizeof(double)*VVAL);

			TMPOUT.read ((char*)FIELDCOOLNOW, sizeof(double)*VVAL);
			TMPOUT.read ((char*)FIELDCOOLBEF, sizeof(double)*VVAL);
			
			fprintf(stderr, "*** CONFIG CONTAINS THE FINITE MU FIELDS               ***\n");
			
			double *N5NOWTMP;
			double *N5NEXTTMP;		    
			
			N5NOWTMP= new double [MyParam.NX*MyParam.NY*MyParam.NZ];
			N5NEXTTMP= new double [MyParam.NX*MyParam.NY*MyParam.NZ];
			  
			TMPOUT.read ((char*)N5NOWTMP , sizeof(double)*SVOL);
			TMPOUT.read ((char*)N5NEXTTMP, sizeof(double)*SVOL);
				
			
			if(DYNAMICALN5==1){
			    N5NOW[0]=N5NOWTMP[0];
			    N5NEXT[0]=N5NEXTTMP[0];
			}else if (DYNAMICALN5==2){			  
			    for (int k=0; k< SVOL; k++) N5NOW[k]=N5NOWTMP[k];
			    for (int k=0; k< SVOL; k++) N5NEXT[k]=N5NEXTTMP[k];
			}				 
		      
			  
			delete[] N5NOWTMP;
			delete[] N5NEXTTMP;
			
			int CurSTATESIZE;
			
			for(int k=0;k<THREADS;k++){
				TMPOUT.read ((char*)&(CurSTATESIZE), sizeof(int));	
				if( GSLRAN[k]->STATE_SIZE !=CurSTATESIZE) { fprintf(stderr, "PROBLEM WITH STATE SIZE OF RNG %d - %d %d\n",k,GSLRAN[k]->STATE_SIZE,CurSTATESIZE); exit(-1);}
			}
				
			for(int k=0;k<THREADS;k++) TMPOUT.read ((char*)(Rng[k]->CURSTATE), GSLRAN[k]->STATE_SIZE);
			
			TMPOUT.close();

			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "*** PREVIOUSLY SAVED     CONFIGURATION READ            ***\n");
			fprintf(stderr, "***                                                    ***\n");
		
		} else{
			fprintf(stderr, "NO PREVIOUS CONFIGURATION: %s FOUND\n", CONFIGFILENAME); exit(-1);
		}
	}

	fprintf(stderr, "**********************************************************\n\n");
		
	//-----------------------------------------------------------------------
	//	IF DIFFERENT MU5 TO BE SET THAN IN THE SAVED STATE 
	//-----------------------------------------------------------------------
	
	if(DYNAMICALN5!=0&& ( MyParam.INITCOND==7 || MyParam.INITCOND==9 )  ){
	  
	      if(DYNAMICALN5==1){
		
		      N5NOW[0]=MyParam.InitCondN5;
		      N5NEXT[0]=0.;
		      N5SCRATCH[0]=0.;
		      
		      
	      }else if (DYNAMICALN5==2) {
		
		      for(int64_t k=0; k< SVOL; k++){
			      N5NOW[k]=MyParam.InitCondN5;
			      N5NEXT[k]=0.;
			      N5SCRATCH[k]=0.;
		      }
	      }
	  
	}	
	
	if(DYNAMICALN5!=0){
	  
	    double n5Avg=0;
	    
	    if( DYNAMICALN5==2 ){
		for  (int k=0;k<AXLEN[0]*AXLEN[1]*AXLEN[2];k++){
			n5Avg+=N5NOW[k]/SVOL;
		}
	    }else{
		n5Avg=N5NOW[0]; 
	    }
	    
	    fprintf(stderr, "***   <n5>:      %+.8e        Starting       ***\n", n5Avg);
	  	  
	}
	
	//-----------------------------------------------------------------------
	//	MEASURE THE PROPERTIES OF THE INITIAL GAUGE FIELDS
	//-----------------------------------------------------------------------

	fprintf(stderr, "***  MEASURING THE SPECTRA AND ENERGY MODES  ***\n");
	
	SUN_CLASS_MEASURE_FIELD_MOMENTA(FIELDNOW, 0, &MyParam);
	
	SUN_CLASS_MEASURE_ENERGY_FOURIER(FIELDNOW, 0, &MyParam,1);	
	
	//---------------------------------------------------------------------------------------------
	//	EVALUATE THE INITIAL ENERGY DENSITY
	//---------------------------------------------------------------------------------------------
	
	double ClassENrg, ClassMNrg,ClassENrgFmunu, ClassMNrgFmunu;
	
	SUN_CLASS_ENERGY(FIELDNOW, &ClassENrg, &ClassMNrg, &MyParam);
	
	//---------------------------------------------------------------------------------------------
	//	EVALUATE THE INITIAL GAUSS CONSTRAINT (LOCAL AND GLOBAL)
	//---------------------------------------------------------------------------------------------
	
	double GaussDevLoc,GaussDevGlob;
	
	SUN_CLASS_GAUSS_CONSTR(FIELDNOW, &GaussDevLoc, &GaussDevGlob, &MyParam);

	//---------------------------------------------------------------------------------------------
	//	EVALUATE THE INITIAL PLAQUETTE
	//---------------------------------------------------------------------------------------------
	
	Complex SpPlaq,WL22;
	
	SUN_CLASS_SPATIALPLAQ(FIELDNOW, &SpPlaq, &MyParam);	
	SUN_CLASS_SPTIALWL22(FIELDNOW, &WL22,&MyParam);	
	
	
	fprintf(stderr, "**********************************************************\n");
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***   START REALTIME EVOLUTION OF THE ENSEMBLE %4d    ***\n",0);
	fprintf(stderr, "***                                                    ***\n");
	fprintf(stderr, "***   %4d  GCLOC:%+.5e     GCGLB:%+.5e  ***\n",0,GaussDevLoc,GaussDevGlob);
	fprintf(stderr, "***         NRG: (%.8e - %.8e)     ***\n",ClassENrg,ClassMNrg);
	fprintf(stderr, "***        PLAQ: (%+.8e - %+.8e)   ***\n",SpPlaq.real(), SpPlaq.imag() );
	fprintf(stderr, "***       WL2x2: (%+.8e - %+.8e)   ***\n",WL22.real(), WL22.imag() );
	fprintf(stderr, "***                                                    ***\n");

	//---------------------------------------------------------------------------------------------
	//	ARRAYS FOR THE DETERMINATION OF THE TOPOLOGICAL CHARGE, I.E. THE CHERN SIMONS NUMBER
	//---------------------------------------------------------------------------------------------	
	
	double* dNCSNaive;
	dNCSNaive=new double [MyParam.NT];
	for(int i=0; i< MyParam.NT;i++)dNCSNaive[i]=0.;
	double DeltaNCSNaive=0.;
	
	double* dNCSCooled;
	dNCSCooled=new double [MyParam.NT];
	for(int i=0; i< MyParam.NT;i++)dNCSCooled[i]=0.;
	double DeltaNCSCooled=0.;

	FILE* NCSEVOLOUT;
	
	if(STARTINGTIME==0){
	
	      NCSEVOLOUT=fopen("NCS.dat","w");
	      	
	} else {
	  
	     NCSEVOLOUT=fopen("NCS.dat","a");
	  
	}
	
	//---------------------------------------------------------------------------------------------
	//	ARRAYS FOR THE DETERMINATION OF THE TEMPORAL REAL-TIME WILSON LOOP
	//---------------------------------------------------------------------------------------------	
	
	Complex* WLOOPREALTIME;
	FILE* WLOOPOUT;
	int WLOOPRLEN;
	
	if(MyParam.WloopRMax!=0){
	  
	      INITCONF = new double [VVAL];	
	      for(int k=0;k<VVAL;k++){
		      INITCONF[k]=FIELDNOW[k];
	      }
	      
	      WLOOPRLEN=MyParam.WloopRMax-MyParam.WloopRMin+1;
	      
	      WLOOPREALTIME = new Complex [WLOOPRLEN];
	      
	      for(int k=0; k<	WLOOPRLEN; k++) WLOOPREALTIME[k]=0.;
	      
	      if(STARTINGTIME==0){
	      
		  WLOOPOUT=fopen("WilsonLoopRealTime.dat","w");
		      
	      } else {
		
		  WLOOPOUT=fopen("WilsonLoopRealTime.dat","a");
		
	      }
	}
	
	//---------------------------------------------------------------------------------------------
	//	File Output for N5 and Energy
	//---------------------------------------------------------------------------------------------	
	
	FILE* N5OUT, *POUT;
	
	if(STARTINGTIME==0){
	
	    POUT=fopen("EnergyGaussWL22.dat","w");
	    if(DYNAMICALN5!=0) N5OUT=fopen("N5RealTime.dat","w");
		
	} else {
	  
	    POUT=fopen("EnergyGaussWL22.dat","a");
	    if(DYNAMICALN5!=0) N5OUT=fopen("N5RealTime.dat","a");
	  
	}
	
	//---------------------------------------------------------------------------------------------
	//	START THE TIME EVOLUTION
	//---------------------------------------------------------------------------------------------	

	if(STARTINGTIME==0){
	
		fprintf(stderr, "***   Current Step %3d    Hamilton Phase               ***\n",0);
		
		SUN_CLASS_EVOLVE_EFIELD(FIELDNOW, FIELDNXT, &MyParam);
		SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);
		
		double* tmp=FIELDBEF;
		FIELDBEF=FIELDNOW;
		FIELDNOW=FIELDNXT;
		FIELDNXT=tmp;	

		STARTINGTIME++;
	
	}
	
	SUN_CLASS_ENERGY(FIELDNOW, &ClassENrg, &ClassMNrg, &MyParam);
	SUN_CLASS_GAUSS_CONSTR(FIELDNOW, &GaussDevLoc, &GaussDevGlob, &MyParam);	
	
	for(int EVTIME=STARTINGTIME; EVTIME<MyParam.NT; EVTIME++){

		//---------------------------------------------------------------------------------------------
		//	ANNOUNCE THE NEXT TIME EVOLUTION STEP
		//---------------------------------------------------------------------------------------------
		
		if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");
		if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");
		if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP STARTED                             ***\n",EVTIME);
		
		//---------------------------------------------------------------------------------------------
		//	MEASURE THE TEMPORAL WILSON LOOP
		//---------------------------------------------------------------------------------------------

		if( (EVTIME)%MyParam.MEASEACH==0 && MyParam.WloopRMax!=0){	
			
		      SUN_CLASS_TMPWLOOP(INITCONF, FIELDNOW, MyParam.WloopRMin, MyParam.WloopRMax , WLOOPREALTIME, &MyParam);

		      if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");
		      if(VERBOSITY==1) fprintf(stderr, "***   Wilson Loops up to R%3d      Eval:  %3d          ***\n",WLOOPRLEN,EVTIME);
		      if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");			
		      for(int R=0; R<WLOOPRLEN; R++) fprintf(stderr, "***           %6e    %6e             ***\n",WLOOPREALTIME[R].real(), WLOOPREALTIME[R].imag());
		      if(VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");			
		      
		      
		      fprintf(WLOOPOUT, "%d %.10e %.10e %.10e %.10e ", EVTIME,ClassENrg, ClassMNrg,GaussDevLoc, GaussDevGlob );
		      for(int R=0; R<WLOOPRLEN; R++) fprintf(WLOOPOUT, "%.10e %.10e ", WLOOPREALTIME[R].real(), WLOOPREALTIME[R].imag());
		      fprintf(WLOOPOUT, "\n");
		      fflush(WLOOPOUT);
		
		}

		//---------------------------------------------------------------------------------------------
		//	EVOLVE ALONG THE CORRESPONDING REAL-TIME PRESCRIPTION
		//---------------------------------------------------------------------------------------------

		if(MyParam.WHICHREALTIME==0){
		
			if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:   Hamilton E, U Evolution          ***\n",EVTIME);
		
			SUN_CLASS_EVOLVE_EFIELD(FIELDNOW, FIELDNXT, &MyParam);
			SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);

		}else if (MyParam.WHICHREALTIME==1){
		  
			if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:   Langevin U    Evolution  (RK)    ***\n",EVTIME);
		  
			SUN_CLASS_EVOLVE_UFIELD_LANGEVIN_RUNGE_KUTTA(FIELDNOW, FIELDNXT, Rng, &MyParam);
			
		}else if (MyParam.WHICHREALTIME==2){
		  
			if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:   Langevin U,n5 Evolution (Naive)  ***\n",EVTIME);
		  
			SUN_CLASS_EVOLVE_UFIELD_LANGEVIN_FINITEMU(FIELDNOW,FIELDNXT,FIELDBEF, N5NOW, N5NEXT, N5SCRATCH, Rng, &MyParam );
			
		}else if (MyParam.WHICHREALTIME==3||MyParam.WHICHREALTIME==4||MyParam.WHICHREALTIME==5){
		  
			if(VERBOSITY==1){
			 if(MyParam.WHICHREALTIME==3) fprintf(stderr, "***  [%6d] STEP:  Chiral EFT E,U Evolution (no n5)  ***\n",EVTIME);
			 if(MyParam.WHICHREALTIME==4) fprintf(stderr, "***  [%6d] STEP:  Chiral EFT E,U Evolution (n5(t))  ***\n",EVTIME);
			 if(MyParam.WHICHREALTIME==5) fprintf(stderr, "***  [%6d] STEP:  Chiral EFT E,U Evolution (n5(t,x))***\n",EVTIME); 
			}
			
			if(EVTIME*MyParam.dt>=N5SWITCHONAT){
			    SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT(FIELDNOW, FIELDNXT,  N5NOW, Rng, &MyParam);
			    if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:  With N5 activated                 ***\n",EVTIME);
			}else{
			    SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT(FIELDNOW, FIELDNXT,  N5SCRATCH, Rng, &MyParam);			  
			    if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:  With N5 deactivated               ***\n",EVTIME);
			}
			
			SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);
			
		}else if (MyParam.WHICHREALTIME==6||MyParam.WHICHREALTIME==7){
		  
			if(VERBOSITY==1){
			 if(MyParam.WHICHREALTIME==6) fprintf(stderr, "***  [%6d] STEP: Tassler EFT E,U Evolution (no n5)  ***\n",EVTIME);
			 if(MyParam.WHICHREALTIME==7) fprintf(stderr, "***  [%6d] STEP: Tassler EFT E,U Evolution (n5(t))  ***\n",EVTIME);
			}
	
			if(EVTIME%(abs(MyParam.ColorConductivity))==0&&EVTIME>0){
			  
				SUN_CLASS_DRAW_E_THERMAL(FIELDNOW, Rng, &MyParam);
				
				if (VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");
				if (VERBOSITY==1) fprintf(stderr, "*** FOR LEAPFROG BRING E TO t=DT/2                     ***\n");
				if (VERBOSITY==1) fprintf(stderr, "***                                                    ***\n");
				
				SUN_CLASS_EVOLVE_EFIELD_FIRSTSTEP(FIELDNOW, &MyParam);	  
				
				
				if (VERBOSITY==1) fprintf(stderr, "***   Current Step %3d    EFIELD REFRESH               ***\n",EVTIME);
			} 

			if(EVTIME*MyParam.dt>=N5SWITCHONAT){
				SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT_TASSLER(FIELDNOW, FIELDNXT,  N5NOW,  &MyParam);
			}else{
				SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT_TASSLER(FIELDNOW, FIELDNXT,  N5SCRATCH,  &MyParam);
			}
			
			SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);
			
		}else if (MyParam.WHICHREALTIME==8){
		  
			if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:  Chiral EFT E,U Evolution (n5,ndis)***\n",EVTIME);
			
			MyParam.DispersionSwitch=0.;
			
			if(EVTIME*MyParam.dt>=N5SWITCHONAT){
			    SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT(FIELDNOW, FIELDNXT,  N5NOW, Rng, &MyParam);
			}else{
			    SUN_CLASS_EVOLVE_EFIELD_EXTERNAL_ANOMALOUSCURRENT(FIELDNOW, FIELDNXT,  N5SCRATCH, Rng, &MyParam);			  
			}
			
			SUN_CLASS_EVOLVE_UFIELD(FIELDNOW, FIELDNXT, &MyParam);
			
		}
		
		
		//---------------------------------------------------------------------------------------------
		//	MEASURE THE CHERN SIMONS NUMBER NAIVELY WITHOUT COOLING
		//---------------------------------------------------------------------------------------------
		
		if( MyParam.COOLEVERY!=0&&(EVTIME)%MyParam.COOLEVERY==0&&EVTIME>1){
		  
			if(MyParam.WHICHNCSDEF==0){
			    SUN_CLASS_MEAS_dNCS_ANTISYMONLY(FIELDNOW,FIELDNXT, FIELDBEF, &dNCSNaive[EVTIME],&MyParam);
			}else{
			    SUN_CLASS_MEAS_dNCS_ROTHKOPF(FIELDNOW,FIELDNXT, FIELDBEF, &dNCSNaive[EVTIME],&MyParam);			  
			}
			
			double IntFac=1.;
			
			if(EVTIME==2) IntFac=0.5;
			
			DeltaNCSNaive+=IntFac*dNCSNaive[EVTIME]*MyParam.dt*MyParam.COOLEVERY;
			
			if(VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Naive dNCS: %+.8e        ***\n",EVTIME, dNCSNaive[EVTIME]);
			if(VERBOSITY==1)  fprintf(stderr, "***                 Naive  NCS: %+.8e        ***\n",DeltaNCSNaive);
		    
		}
		
		//---------------------------------------------------------------------------------------------
		//	PERFORM COOLING
		//---------------------------------------------------------------------------------------------

		if(MyParam.WHICHCOOLPRESCR!=0&&(EVTIME)%MyParam.COOLEVERY==0&&EVTIME>1){
		
		      Parameters CoolParam;
		      CoolParam=MyParam;

		      if(MyParam.COOLEVERY>1||EVTIME==2){
			      
			      if (VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Cooling: Load all 3 Configs        ***\n",EVTIME);
			      
			      for(int k=0;k<VVAL;k++){
				      FIELDCOOLNOW[k]=FIELDNOW[k];
				      FIELDCOOLNXT[k]=FIELDNXT[k];
				      FIELDCOOLBEF[k]=FIELDBEF[k];
			      }	
			      
			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLBEF, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]- PLAQ: (%+.8e - %+.8e)   ***\n",0,SpPlaq.real(), SpPlaq.imag() );
			      
			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNOW, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]c PLAQ: (%+.8e - %+.8e)   ***\n",0,SpPlaq.real(), SpPlaq.imag() );
			      
			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNXT, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]+ PLAQ: (%+.8e - %+.8e)   ***\n",0,SpPlaq.real(), SpPlaq.imag() );
			      
		      }else{
			      if (VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Cooling: Load new Config           ***\n",EVTIME);
			      for(int k=0;k<VVAL;k++){
				      FIELDCOOLNXT[k]=FIELDNXT[k];
			      }	
			      
			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNXT, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]+ PLAQ: (%+.8e - %+.8e)   ***\n",0,SpPlaq.real(), SpPlaq.imag() );
			            
		      }		  
		  
		
		      if(MyParam.WHICHCOOLPRESCR<5){
		
			    for( int tau=0; tau<MyParam.COARSENING; tau++){
			      
			      if(MyParam.COOLEVERY>1||EVTIME==2){
				
					if(MyParam.WHICHCOOLPRESCR==1){
					    SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLNOW, FIELDCOOLTMP, &MyParam );
					    SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLBEF, FIELDCOOLTMP, &MyParam );
					}else if (MyParam.WHICHCOOLPRESCR==2){
					    SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNOW, FIELDCOOLTMP, &MyParam );
					    SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLBEF, FIELDCOOLTMP, &MyParam );
					}else if (MyParam.WHICHCOOLPRESCR==3){
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNOW, FIELDCOOLTMP, &MyParam );
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLBEF, &MyParam );
					}else if (MyParam.WHICHCOOLPRESCR==4){
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNOW, FIELDCOOLTMP, &MyParam );
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLBEF, FIELDCOOLTMP, &MyParam );
					} 
					
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLBEF, &SpPlaq, &MyParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]- PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNOW, &SpPlaq, &MyParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]c PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
					
			      }
	      
			      if(MyParam.WHICHCOOLPRESCR==1){
				  SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLTMP, &MyParam );
			      }else if (MyParam.WHICHCOOLPRESCR==2){
				  SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNXT, FIELDCOOLTMP, &MyParam );
			      }else if (MyParam.WHICHCOOLPRESCR==3){
				  SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLTMP, &MyParam );
			      }else if (MyParam.WHICHCOOLPRESCR==4){
				  SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNXT, FIELDCOOLTMP, &MyParam );
			      }

			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNXT, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]+ PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
			      
			    }
			    
			      
			    if(MyParam.COARSENING!=0){
			      
				  if(MyParam.COOLEVERY>1||EVTIME==2){
					  
					if (VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Cooling: Coarsen  3 Configs        ***\n",EVTIME);
					  
					  LOAD_UFIELD_COARSENDED(FIELDCOOLTMP,FIELDCOOLNOW,&MyParam);
					  double *tmp;
					  tmp=FIELDCOOLNOW;
					  FIELDCOOLNOW=FIELDCOOLTMP;
					  FIELDCOOLTMP=tmp;
					
					  LOAD_UFIELD_COARSENDED(FIELDCOOLTMP,FIELDCOOLBEF,&MyParam);
					  tmp=FIELDCOOLBEF;
					  FIELDCOOLBEF=FIELDCOOLTMP;
					  FIELDCOOLTMP=tmp;
					  
				  } else {
				  
					if (VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Cooling: Coarsen  1 Configs        ***\n",EVTIME);
				  
				  }
				  
				  LOAD_UFIELD_COARSENDED(FIELDCOOLTMP,FIELDCOOLNXT,&MyParam);
				  double *tmp;
				  tmp=FIELDCOOLNXT;
				  FIELDCOOLNXT=FIELDCOOLTMP;
				  FIELDCOOLTMP=tmp;
				  
			      
				  CoolParam.NX=MyParam.NX/2;
				  CoolParam.NY=MyParam.NY/2;
				  CoolParam.NZ=MyParam.NZ/2;
			    }
			    
			    
			    for( int tau=MyParam.COARSENING; tau<MyParam.COOLDEPTH; tau++){
			      
			      if(MyParam.COOLEVERY>1||EVTIME==2){
				
					if(MyParam.WHICHCOOLPRESCR==1){
					    SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLNOW, FIELDCOOLTMP, &CoolParam );
					    SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLBEF, FIELDCOOLTMP, &CoolParam );
					}else if (MyParam.WHICHCOOLPRESCR==2){
					    SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNOW, FIELDCOOLTMP, &CoolParam );
					    SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLBEF, FIELDCOOLTMP, &CoolParam );
					}else if (MyParam.WHICHCOOLPRESCR==3){
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNOW, FIELDCOOLTMP, &CoolParam );
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLBEF, &CoolParam );
					}else if (MyParam.WHICHCOOLPRESCR==4){
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNOW, FIELDCOOLTMP, &CoolParam );
					    SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLBEF, FIELDCOOLTMP, &CoolParam );
					} 
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLBEF, &SpPlaq, &CoolParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]- PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNOW, &SpPlaq, &CoolParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]c PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
					
			      }
	      
			      if(MyParam.WHICHCOOLPRESCR==1){
				  SUN_CLASS_UFIELD_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLTMP, &CoolParam );
			      }else if (MyParam.WHICHCOOLPRESCR==2){
				  SUN_CLASS_UFIELD_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNXT, FIELDCOOLTMP, &CoolParam );
			      }else if (MyParam.WHICHCOOLPRESCR==3){
				  SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE(FIELDCOOLNXT, FIELDCOOLTMP, &CoolParam );
			      }else if (MyParam.WHICHCOOLPRESCR==4){
				  SUN_CLASS_UFIELD_IMPROVED_COOLING_INPLACE_RUNGE_KUTTA(FIELDCOOLNXT, FIELDCOOLTMP, &CoolParam );
			      }

			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNXT, &SpPlaq, &CoolParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]+ PLAQ: (%+.8e - %+.8e)   ***\n",tau+1,SpPlaq.real(), SpPlaq.imag() );
			      
			    }
			    		  
		      } else{
			
			      if(MyParam.COOLEVERY>1||EVTIME==2){
				
					SUN_CLASS_FOURIER_COOL_INPLACE(FIELDCOOLNOW, PPOS[0][MyParam.COOLDEPTH], &MyParam);
					SUN_CLASS_FOURIER_COOL_INPLACE(FIELDCOOLBEF, PPOS[0][MyParam.COOLDEPTH], &MyParam);					
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLBEF, &SpPlaq, &MyParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]- PLAQ: (%+.8e - %+.8e)   ***\n",MyParam.COOLDEPTH,SpPlaq.real(), SpPlaq.imag() );
					
					if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNOW, &SpPlaq, &MyParam);	
					if (VERBOSITY==1) fprintf(stderr, "*** [%3d]c PLAQ: (%+.8e - %+.8e)   ***\n",MyParam.COOLDEPTH,SpPlaq.real(), SpPlaq.imag() );
					
			      }
	      
			      SUN_CLASS_FOURIER_COOL_INPLACE(FIELDCOOLNXT, PPOS[0][MyParam.COOLDEPTH], &MyParam);

			      if (VERBOSITY==1) SUN_CLASS_SPATIALPLAQ(FIELDCOOLNXT, &SpPlaq, &MyParam);	
			      if (VERBOSITY==1) fprintf(stderr, "*** [%3d]+ PLAQ: (%+.8e - %+.8e)   ***\n",MyParam.COOLDEPTH,SpPlaq.real(), SpPlaq.imag() );
			
		      }
		      
		      
		      if(MyParam.WHICHNCSDEF==0){
			  SUN_CLASS_MEAS_dNCS_ANTISYMONLY(FIELDCOOLNOW,FIELDCOOLNXT, FIELDCOOLBEF, &dNCSCooled[EVTIME],&CoolParam);
		      }else{
			  SUN_CLASS_MEAS_dNCS_ROTHKOPF(FIELDCOOLNOW,FIELDCOOLNXT, FIELDCOOLBEF, &dNCSCooled[EVTIME],&CoolParam);			  
		      }
		      
		      double IntFac=1.;
		      
		      if(EVTIME==2) IntFac=0.5;
		      
		      DeltaNCSCooled+=IntFac*dNCSCooled[EVTIME]*MyParam.dt*MyParam.COOLEVERY;
		      
		      if(VERBOSITY==1)  fprintf(stderr, "***  [%6d] STEP: Coold dNCS: %+.8e        ***\n",EVTIME, dNCSCooled[EVTIME]);
		      if(VERBOSITY==1)  fprintf(stderr, "***                 Coold  NCS: %+.8e        ***\n",DeltaNCSCooled);
		  
		}
	
		if((EVTIME)%MyParam.COOLEVERY==0&&EVTIME>1){
		  
		      fprintf(NCSEVOLOUT, "%d %.10e %.10e %.10e %.10e\n", EVTIME+1, dNCSNaive[EVTIME], DeltaNCSNaive, dNCSCooled[EVTIME],DeltaNCSCooled);	
		      fflush(NCSEVOLOUT);
		  
		  
		}
		
		
		if(DYNAMICALN5!=0){
		  
		    double n5Avg=0;
		    
		    if(EVTIME*MyParam.dt>=N5SWITCHONAT){
		      
			  if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:  N5 evolution switched on          ***\n",EVTIME);
		      
			  if( DYNAMICALN5==2 ){
		      
			      if(EVTIME*MyParam.dt>=N5SWITCHONAT) SUN_CLASS_EVOLVE_N5_SPATIAL( FIELDCOOLNOW, FIELDCOOLNXT, FIELDCOOLBEF, N5NOW, N5NEXT, N5SCRATCH, Rng, &MyParam );
			    
			      for  (int k=0;k<AXLEN[0]*AXLEN[1]*AXLEN[2];k++){
				      n5Avg+=N5NEXT[k]/SVOL;
			      }
			      
			  }else{
			    
			      double NCSTOUSE;
			    
			      if(MyParam.WHICHCOOLPRESCR>0) { NCSTOUSE=dNCSCooled[EVTIME]; }else{ NCSTOUSE=dNCSNaive[EVTIME]; }
			    
			      if(EVTIME*MyParam.dt>=N5SWITCHONAT) SUN_CLASS_EVOLVE_N5_NOSPATIAL(N5NOW, N5NEXT, NCSTOUSE , &MyParam);
			      
			      n5Avg=N5NEXT[0]; 
			  }
		    } else{
		      
			    if(VERBOSITY==1) fprintf(stderr, "***  [%6d] STEP:  N5 evolution switched off         ***\n",EVTIME);
		      
		    }
		    
		    if (VERBOSITY==1 ) fprintf(stderr, "***   <n5>:      %+.8e                       ***\n", n5Avg );
		    fprintf(N5OUT, "%d %.10e\n", EVTIME+1, n5Avg);
		    fflush(N5OUT);
			  
		}

		

		
		double* tmp=FIELDBEF;
		FIELDBEF=FIELDNOW;
		FIELDNOW=FIELDNXT;
		FIELDNXT=tmp;	
		
		if(DYNAMICALN5>0&&EVTIME*MyParam.dt>=N5SWITCHONAT){
		  
		  tmp=N5NOW;
		  N5NOW=N5NEXT;
		  N5NEXT=tmp;	
		  
		}
		
		tmp=FIELDCOOLBEF;
		FIELDCOOLBEF=FIELDCOOLNOW;
		FIELDCOOLNOW=FIELDCOOLNXT;
		FIELDCOOLNXT=tmp;	

		
		if( (EVTIME+1)%MyParam.MEASEACH==0){
						
			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE ENERGY DENSITY
			//---------------------------------------------------------------------------------------------
		
			SUN_CLASS_ENERGY(FIELDNOW, &ClassENrg, &ClassMNrg, &MyParam);
			SUN_CLASS_ENERGY_FMUNU(FIELDNOW,FIELDNXT,FIELDBEF, &ClassENrgFmunu, &ClassMNrgFmunu, &MyParam);
		
			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE GAUSS CONSTRAINT (LOCAL AND GLOBAL)
			//---------------------------------------------------------------------------------------------
			
			SUN_CLASS_GAUSS_CONSTR(FIELDNOW, &GaussDevLoc, &GaussDevGlob, &MyParam);
		
			//---------------------------------------------------------------------------------------------
			//	EVALUATE THE 2x2 WILSON LOOP
			//---------------------------------------------------------------------------------------------
			
			SUN_CLASS_SPATIALPLAQ(FIELDNOW, &SpPlaq, &MyParam);	
			SUN_CLASS_SPTIALWL22(FIELDNOW, &WL22, &MyParam);	

			if(VERBOSITY==1){
			fprintf(stderr, "***                                                    ***\n");
			fprintf(stderr, "***  Measurements for t=%4d                           ***\n",EVTIME+1);
			fprintf(stderr, "***           GCLOC:%+.5e  GCGLB:%+.5e   ***\n",GaussDevLoc,GaussDevGlob);
			fprintf(stderr, "***         NRG: (%.8e - %.8e)     ***\n",ClassENrg,ClassMNrg);
			fprintf(stderr, "***        PLAQ: (%+.8e - %+.8e)   ***\n",SpPlaq.real(), SpPlaq.imag() );
			fprintf(stderr, "***       WL2x2: (%+.8e - %+.8e)   ***\n",WL22.real(), WL22.imag() );
			fprintf(stderr, "***                                                    ***\n");
			}
						
			POUT=fopen("EnergyGaussWL22.dat","a");
			fprintf(POUT, "%d %.10e %.10e %.10e %.10e %.10e %.10e %.10e %.10e %.10e %.10e\n", EVTIME+1,GaussDevLoc,GaussDevGlob,ClassENrg,ClassMNrg,SpPlaq.real(), SpPlaq.imag(),WL22.real(), WL22.imag(),ClassENrgFmunu, ClassMNrgFmunu );
			fclose(POUT);			

		}
			
		if( (EVTIME+1)%MyParam.SAVEEACH==0 ){
			
				fprintf(stderr, "***                                                    ***\n");
				fprintf(stderr, "*** REACHED FIELD SAVINGS POINT: OUTPUT TO HD          ***\n");
				fprintf(stderr, "***                                                    ***\n");
				
				char OLDFILE[256];
				std::fstream TMPOUT;
					
				sprintf(OLDFILE,"FLDBACKUP.%d",EVTIME+1);
					
				TMPOUT.open(OLDFILE, std::ios::out|std::ios::binary);

				int OUTTIME=EVTIME+1;
				
				TMPOUT.write ((char*)(&OUTTIME), sizeof(int));
				TMPOUT.write ((char*)(&MyParam.WHICHREALTIME), sizeof(int));				
				
				TMPOUT.write ((char*)FIELDNOW, sizeof(double)*VVAL);
				TMPOUT.write ((char*)FIELDBEF, sizeof(double)*VVAL);				
				
				TMPOUT.write ((char*)FIELDCOOLNOW, sizeof(double)*VVAL);
				TMPOUT.write ((char*)FIELDCOOLBEF, sizeof(double)*VVAL);
	
				if(DYNAMICALN5==1){
				  TMPOUT.write ((char*)N5NOW , sizeof(double)*1);
				  TMPOUT.write ((char*)N5NEXT, sizeof(double)*1);				  
				}else if (DYNAMICALN5==2) {
				  TMPOUT.write ((char*)N5NOW , sizeof(double)*SVOL);
				  TMPOUT.write ((char*)N5NEXT, sizeof(double)*SVOL);				  
				}
				
				for(int k=0;k<THREADS;k++) TMPOUT.write ((char*)&(GSLRAN[k]->STATE_SIZE), sizeof(int));		
				for(int k=0;k<THREADS;k++) TMPOUT.write ((char*)(Rng[k]->CURSTATE), GSLRAN[k]->STATE_SIZE);
				
				TMPOUT.close();				
				
		}
		
	}
	
}



//---------------------------------------------------------------------------------------------
//	AUXILIARY ROUTINES FOR THE FAST SU(3) MATRIX EXPONENTIAL
//---------------------------------------------------------------------------------------------



#define SQR(x)      ((x)*(x))                        // x^2 
#define SQR_ABS(x)  (SQR(creal(x)) + SQR(cimag(x)))  // |x|^2


void zhetrd3(double complex A[3][3], double complex Q[3][3],
                    double d[3], double e[2])
// ----------------------------------------------------------------------------
// Reduces a hermitian 3x3 matrix to real tridiagonal form by applying
// (unitary) Householder transformations:
//            [ d[0]  e[0]       ]
//    A = Q . [ e[0]  d[1]  e[1] ] . Q^T
//            [       e[1]  d[2] ]
// The function accesses only the diagonal and upper triangular parts of
// A. The access is read-only.
// ---------------------------------------------------------------------------
{
  const int n = 3;
  double complex u[n], q[n];
  double complex omega, f;
  double K, h, g;
  
  // Initialize Q to the identitity matrix
#ifndef EVALS_ONLY
  for (int i=0; i < n; i++)
  {
    Q[i][i] = 1.0;
    for (int j=0; j < i; j++)
      Q[i][j] = Q[j][i] = 0.0;
  }
#endif

  // Bring first row and column to the desired form 
  h = SQR_ABS(A[0][1]) + SQR_ABS(A[0][2]);
  if (creal(A[0][1]) > 0)
    g = -sqrt(h);
  else
    g = sqrt(h);
  e[0] = g;
  f    = g * A[0][1];
  u[1] = conj(A[0][1]) - g;
  u[2] = conj(A[0][2]);
  
  omega = h - f;
  if (creal(omega) > 0.0)
  {
    omega = 0.5 * (1.0 + conj(omega)/omega) / creal(omega);
    K = 0.0;
    for (int i=1; i < n; i++)
    {
      f    = conj(A[1][i]) * u[1] + A[i][2] * u[2];
      q[i] = omega * f;                  // p
      K   += creal(conj(u[i]) * f);      // u* A u
    }
    K *= 0.5 * SQR_ABS(omega);

    for (int i=1; i < n; i++)
      q[i] = q[i] - K * u[i];
    
    d[0] = creal(A[0][0]);
    d[1] = creal(A[1][1]) - 2.0*creal(q[1]*conj(u[1]));
    d[2] = creal(A[2][2]) - 2.0*creal(q[2]*conj(u[2]));
    
    // Store inverse Householder transformation in Q
#ifndef EVALS_ONLY
    for (int j=1; j < n; j++)
    {
      f = omega * conj(u[j]);
      for (int i=1; i < n; i++)
        Q[i][j] = Q[i][j] - f*u[i];
    }
#endif

    // Calculate updated A[1][2] and store it in f
    f = A[1][2] - q[1]*conj(u[2]) - u[1]*conj(q[2]);
  }
  else
  {
    for (int i=0; i < n; i++)
      d[i] = creal(A[i][i]);
    f = A[1][2];
  }

  // Make (23) element real
  e[1] = cabs(f);
#ifndef EVALS_ONLY
  if (e[1] != 0.0)
  {
    f = conj(f) / e[1];
    for (int i=1; i < n; i++)
      Q[i][n-1] = Q[i][n-1] * f;
  }
#endif
}

// ----------------------------------------------------------------------------
void dsytrd3(double A[3][3], double Q[3][3], double d[3], double e[2])
// ----------------------------------------------------------------------------
// Reduces a symmetric 3x3 matrix to tridiagonal form by applying
// (unitary) Householder transformations:
//            [ d[0]  e[0]       ]
//    A = Q . [ e[0]  d[1]  e[1] ] . Q^T
//            [       e[1]  d[2] ]
// The function accesses only the diagonal and upper triangular parts of
// A. The access is read-only.
// ---------------------------------------------------------------------------
{
  const int n = 3;
  double u[n], q[n];
  double omega, f;
  double K, h, g;
  
  // Initialize Q to the identitity matrix
#ifndef EVALS_ONLY
  for (int i=0; i < n; i++)
  {
    Q[i][i] = 1.0;
    for (int j=0; j < i; j++)
      Q[i][j] = Q[j][i] = 0.0;
  }
#endif

  // Bring first row and column to the desired form 
  h = SQR(A[0][1]) + SQR(A[0][2]);
  if (A[0][1] > 0)
    g = -sqrt(h);
  else
    g = sqrt(h);
  e[0] = g;
  f    = g * A[0][1];
  u[1] = A[0][1] - g;
  u[2] = A[0][2];
  
  omega = h - f;
  if (omega > 0.0)
  {
    omega = 1.0 / omega;
    K     = 0.0;
    for (int i=1; i < n; i++)
    {
      f    = A[1][i] * u[1] + A[i][2] * u[2];
      q[i] = omega * f;                  // p
      K   += u[i] * f;                   // u* A u
    }
    K *= 0.5 * SQR(omega);

    for (int i=1; i < n; i++)
      q[i] = q[i] - K * u[i];
    
    d[0] = A[0][0];
    d[1] = A[1][1] - 2.0*q[1]*u[1];
    d[2] = A[2][2] - 2.0*q[2]*u[2];
    
    // Store inverse Householder transformation in Q
#ifndef EVALS_ONLY
    for (int j=1; j < n; j++)
    {
      f = omega * u[j];
      for (int i=1; i < n; i++)
        Q[i][j] = Q[i][j] - f*u[i];
    }
#endif

    // Calculate updated A[1][2] and store it in e[1]
    e[1] = A[1][2] - q[1]*u[2] - u[1]*q[2];
  }
  else
  {
    for (int i=0; i < n; i++)
      d[i] = A[i][i];
    e[1] = A[1][2];
  }
}

// ----------------------------------------------------------------------------
void dsyev2(double A, double B, double C, double *rt1, double *rt2,
                   double *cs, double *sn)
// ----------------------------------------------------------------------------
// Calculates the eigensystem of a real symmetric 2x2 matrix
//    [ A  B ]
//    [ B  C ]
// in the form
//    [ A  B ]  =  [ cs  -sn ] [ rt1   0  ] [  cs  sn ]
//    [ B  C ]     [ sn   cs ] [  0   rt2 ] [ -sn  cs ]
// where rt1 >= rt2. Note that this convention is different from the one used
// in the LAPACK routine DLAEV2, where |rt1| >= |rt2|.
// ----------------------------------------------------------------------------
{
  double sm = A + C;
  double df = A - C;
  double rt = sqrt(SQR(df) + 4.0*B*B);
  double t;

  if (sm > 0.0)
  {
    *rt1 = 0.5 * (sm + rt);
    t = 1.0/(*rt1);
    *rt2 = (A*t)*C - (B*t)*B;
  }
  else if (sm < 0.0)
  {
    *rt2 = 0.5 * (sm - rt);
    t = 1.0/(*rt2);
    *rt1 = (A*t)*C - (B*t)*B;
  }
  else       // This case needs to be treated separately to avoid div by 0
  {
    *rt1 = 0.5 * rt;
    *rt2 = -0.5 * rt;
  }

  // Calculate eigenvectors
  if (df > 0.0)
    *cs = df + rt;
  else
    *cs = df - rt;

  if (fabs(*cs) > 2.0*fabs(B))
  {
    t   = -2.0 * B / *cs;
    *sn = 1.0 / sqrt(1.0 + SQR(t));
    *cs = t * (*sn);
  }
  else if (fabs(B) == 0.0)
  {
    *cs = 1.0;
    *sn = 0.0;
  }
  else
  {
    t   = -0.5 * (*cs) / B;
    *cs = 1.0 / sqrt(1.0 + SQR(t));
    *sn = t * (*cs);
  }

  if (df > 0.0)
  {
    t   = *cs;
    *cs = -(*sn);
    *sn = t;
  }
}


#define M_sqrt3    1.73205080756887729352744634151   // sqrt(3)




